@@ -4,7 +4,7 @@
  * Project: GGZ Chess game module
  * Date: 03/01/01
  * Desc: Game main functions
- * $Id: game.c,v 1.28 2003/02/15 03:34:42 jdorje Exp $
+ * $Id: game.c 7748 2006-01-07 22:02:13Z jdorje $
  *
  * Copyright (C) 2000 Ismael Orenstein.
  *
@@ -32,27 +32,76 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <ggz.h>
+#include <stdarg.h>
+#include <stdio.h>
+#include <time.h>
+#include <string.h>
 
 #include "chess.h"
 #include "game.h"
+#include "ai.h"
 
 /* Game from cgc */
 game_t *game;
 /* Game info from ggz */
 struct chess_info game_info;
-/* Cronometer
+/* Chronometer
  * Use to store the time of the current play */
-struct timeval cronometer;
+struct timeval chronometer;
+/* Savegame file */
+static FILE *savegame;
+/* Moves table */
+/*static char moves[1000][6];
+static int currentmove;*/
+
+static void game_restart_chronometer(void);
+static int game_read_chronometer(void);
+
+/* All the important stuff happens here */
+static int game_update(int event_id, const void *data);
+
+/* Send MSG_SEAT to the player */
+static void game_send_seat(int seat);
+
+/* Send MSG_PLAYERS to all the players */
+static void game_send_players(void);
+
+/* Send REQ_TIME */
+static void game_request_time(int seat);
+
+/* Send RSP_TIME */
+static void game_send_time(int seat);
+
+/* Send MSG_START to everyone */
+static void game_send_start(void);
+
+/* Send RSP_TIME to players */
+static void game_send_update(void);
+
+/* Send MSG_MOVE to everyone */
+static void game_send_move(char *move, int time);
+
+/* Send MSG_GAMEOVER to everyone */
+static void game_send_gameover(char code);
+
+/* Send REQ_DRAW */
+static void game_send_draw(int seat);
+
+/* PGN savegame function */
+static void game_save(char *fmt, ...);
+
 
 /* Lazy way of doing things */
 #define OUT_OF_TIME(a) (game_info.seconds[a] <= 0)
 
 /* Translates a GGZ state event into a chess game event. */
-void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data) {
-  GGZdModState old_state = *(GGZdModState*)data;
+void game_handle_ggz_state(GGZdMod *ggz,
+			   GGZdModEvent event, const void *data)
+{
+  const GGZdModState *old_state = data;
   GGZdModState new_state = ggzdmod_get_state(ggz);
 
-  if (old_state == GGZDMOD_STATE_CREATED)
+  if (*old_state == GGZDMOD_STATE_CREATED)
     game_update(CHESS_EVENT_LAUNCH, NULL);
 
   if (new_state == GGZDMOD_STATE_DONE);
@@ -60,21 +109,26 @@ void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data) {
 }
 
 /* Translates a GGZ join event into a chess game event. */
-void game_handle_ggz_join(GGZdMod *ggz, GGZdModEvent event, void *data) {
-  int player = ((GGZSeat*)data)->num;
-  game_update(CHESS_EVENT_JOIN, &player);
+void game_handle_ggz_join(GGZdMod *ggz,
+			  GGZdModEvent event, const void *data) {
+  const GGZSeat *old_seat = data;
+
+  game_update(CHESS_EVENT_JOIN, &old_seat->num);
 }
 
 /* Translates a GGZ leave event into a chess game event. */
-void game_handle_ggz_leave(GGZdMod *ggz, GGZdModEvent event, void *data) {
-  int player = ((GGZSeat*)data)->num;
-  game_update(CHESS_EVENT_LEAVE, &player);
+void game_handle_ggz_leave(GGZdMod *ggz,
+			   GGZdModEvent event, const void *data) {
+  const GGZSeat *old_seat = data;
+
+  game_update(CHESS_EVENT_LEAVE, &old_seat->num);
 }
 
 static int seats_full(void)
 {
-	return ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_OPEN)
-		+ ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_RESERVED) == 0;
+	return ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_OPEN) == 0
+	  && ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_RESERVED) == 0
+	  && ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_ABANDONED) == 0;
 }
 
 /* game_update(int event_id, void *data) 
@@ -93,9 +147,14 @@ static int seats_full(void)
  *  CHESS_EVENT_FLAG: (int)time until now for move
  *  CHESS_EVENT_DRAW: (int)What to add to the draw char
  *  CHESS_EVENT_START: NULL */
-int game_update(int event_id, void *data) {
+static int game_update(int event_id, const void *data)
+{
   int time, st;
-  struct timeval now;
+  int from, to;
+  int ret;
+  char botmove[6];
+  int color, seatnum;
+
   switch (event_id) {
     case CHESS_EVENT_LAUNCH:
       /* Check for current state */
@@ -105,6 +164,14 @@ int game_update(int event_id, void *data) {
       game = cgc_create_game();
       /* Go to wait state */
       game_info.state = CHESS_STATE_WAIT;
+
+      if (ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_BOT) == 1) {
+        seatnum = ((ggzdmod_get_seat(game_info.ggz, 0).type == GGZ_SEAT_BOT) ? 0 : 1);
+        color = ((seatnum == 0) ? WHITE : BLACK);
+        if (cgc_join_game(game, color) < 0)
+          return -1;
+        chess_ai_init((color == BLACK ? C_BLACK : C_WHITE), 2);
+      }
       break;
     case CHESS_EVENT_JOIN:
       /* Check for current state */
@@ -126,12 +193,12 @@ int game_update(int event_id, void *data) {
         /* Congratulations: You are the host! */
         game_info.host = *(int *)data;
         game_request_time(*(int *)data);
-      } else if (game_info.clock_type >= 0) {
+      } else if (game_info.clock_type != CHESS_CLOCK_UNSET) {
         /* We already have set the time */
         game_send_time(*(int *)data);
       }
       /* Should we start the game? */
-      if (game_info.clock_type >= 0 && seats_full())
+      if (game_info.clock_type != CHESS_CLOCK_UNSET && seats_full())
         game_update(CHESS_EVENT_START, NULL);
       break;
     case CHESS_EVENT_TIME:
@@ -153,7 +220,7 @@ int game_update(int event_id, void *data) {
       if (ggzdmod_get_seat(game_info.ggz, 1).type == GGZ_SEAT_PLAYER)
         game_send_time(1);
       /* Should we start the game? */
-      if (game_info.clock_type >= 0 && seats_full())
+      if (game_info.clock_type != CHESS_CLOCK_UNSET && seats_full())
         game_update(CHESS_EVENT_START, NULL);
       break;
     case CHESS_EVENT_START:
@@ -163,24 +230,49 @@ int game_update(int event_id, void *data) {
       ggzdmod_set_state(game_info.ggz, GGZDMOD_STATE_PLAYING);
       /* Send start message! */
       game_send_start();
-      /* Stop the cronometer if it's the case */
+      /* Stop the chronometer if it's the case */
       if (game_info.clock_type == CHESS_CLOCK_SERVERLAG ||
           game_info.clock_type == CHESS_CLOCK_SERVER)
-        game_stop_cronometer();
+        game_restart_chronometer();
       /* Update state */
       game_info.state = CHESS_STATE_PLAYING;
+      /* If bot is white (seat 1), start to move */
+      seatnum = ((ggzdmod_get_seat(game_info.ggz, 0).type == GGZ_SEAT_BOT) ? 0 : 1);
+      if ((ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_BOT) == 1)
+      && (game_info.turn % 2 == seatnum)) {
+        ret = chess_ai_find((seatnum == 0 ? C_WHITE : C_BLACK), &from, &to);
+        botmove[0] = from % 8 + 'A';
+        botmove[1] = from / 8 + '1';
+        botmove[2] = to % 8 + 'A';
+        botmove[3] = to / 8 + '1';
+        botmove[4] = 0;
+        botmove[5] = 0;
+        data = &botmove;
+        game_update(CHESS_EVENT_MOVE, data);
+      }
       break;
     case CHESS_EVENT_MOVE:
+      ggz_debug(DEBUG_GAME, "** got move!\n");
       /* Check for state */
       if (game_info.state != CHESS_STATE_PLAYING)
         break;
+      ggz_debug(DEBUG_GAME, "** handle move!\n");
       ggzdmod_log(game_info.ggz, "Move: %s", (char *)data);
       /* Try to make the move via cgc */
       if ( !data || (st = cgc_make_move(game, (char *)data)) < 0) {
+        ggz_debug(DEBUG_GAME, "** cgc: not valid (%p): %d!\n", data, st);
         ggzdmod_log(game_info.ggz, "CGC status: %d", st);
         game_send_move(NULL, 0);
         break;
       }
+      ggz_debug(DEBUG_GAME, "** move is valid!\n");
+      from = ((char*)data)[0] - 65 + (((char*)data)[1] - 49) * 8;
+      to = ((char*)data)[2] - 65 + (((char*)data)[3] - 49) * 8;
+      ggz_debug(DEBUG_GAME, "** ai: execute %i->%i\n", from, to);
+      ret = chess_ai_move(from, to, 1);
+      ggz_debug(DEBUG_GAME, "** ai: executed with result %i\n", ret);
+      chess_ai_output();
+
       /* Move was valid */
       if (game_info.clock_type == CHESS_CLOCK_NOCLOCK)
           game_send_move((char *)data, 0);
@@ -192,9 +284,9 @@ int game_update(int event_id, void *data) {
           game_info.seconds[game_info.turn % 2] -= time;
         else
           game_info.seconds[game_info.turn % 2] = 0;
-        /* If server, stop the cronometer */
+        /* If server, stop the chronometer */
         if (game_info.clock_type != CHESS_CLOCK_CLIENT)
-          game_stop_cronometer();
+          game_restart_chronometer();
         /* Send the move */
         game_send_move((char *)data, time);
       }
@@ -235,6 +327,29 @@ int game_update(int event_id, void *data) {
         game_update(CHESS_EVENT_DRAW, &st);
       }
       game_info.turn++;
+      ggz_debug(DEBUG_GAME, "** move was: %i/%i %i/%i\n",
+        ((char*)data)[0] - 65, ((char*)data)[1] - 49,
+        ((char*)data)[2] - 65, ((char*)data)[3] - 49);
+      ggz_debug(DEBUG_GAME, "** next turn!\n");
+
+      seatnum = ((ggzdmod_get_seat(game_info.ggz, 0).type == GGZ_SEAT_BOT) ? 0 : 1);
+      if ((ggzdmod_count_seats(game_info.ggz, GGZ_SEAT_BOT) == 1)
+      && (game_info.turn % 2 == seatnum)) {
+        ggz_debug(DEBUG_GAME, "** now move chess bot!\n");
+        ret = chess_ai_find((seatnum == 0 ? C_WHITE : C_BLACK), &from, &to);
+        ggz_debug(DEBUG_GAME, "** would move from %i to %i! (valid: %i)\n", from, to, ret);
+        botmove[0] = from % 8 + 'A';
+        botmove[1] = from / 8 + '1';
+        botmove[2] = to % 8 + 'A';
+        botmove[3] = to / 8 + '1';
+        botmove[4] = 0;
+        botmove[5] = 0;
+        data = &botmove;
+        ggz_debug(DEBUG_GAME, "** that would be: %i/%i %i/%i\n",
+          ((char*)data)[0] - 65, ((char*)data)[1] - 49,
+          ((char*)data)[2] - 65, ((char*)data)[3] - 49);
+        game_update(CHESS_EVENT_MOVE, data);
+      }
       break;
     case CHESS_EVENT_UPDATE_TIME:
       if (game_info.state != CHESS_STATE_PLAYING)
@@ -253,9 +368,8 @@ int game_update(int event_id, void *data) {
         break;
       /* If we have server clock, then update the time */
       if (game_info.clock_type == CHESS_CLOCK_SERVER) {
-        gettimeofday(&now, NULL);
-        time = now.tv_sec - cronometer.tv_sec;
-        game_stop_cronometer();
+	time = game_read_chronometer();
+        game_restart_chronometer();
         game_info.seconds[game_info.turn%2] -= time;
       }
       if (OUT_OF_TIME(0))
@@ -336,12 +450,13 @@ int game_update(int event_id, void *data) {
  *  CHESS_REQ_FLAG   -> Check for the right clock, then trigger EVENT_FLAG
  *  CHESS_REQ_DRAW   -> Check if he hasn't done it yet, then trigger EVENT_DRAW
  *  */
-void game_handle_player_data(GGZdMod *ggz, GGZdModEvent id, void *seat_data) {
-  int *seat = (int*)seat_data;
+void game_handle_player_data(GGZdMod *ggz, GGZdModEvent id,
+			     const void *seat_data)
+{
+  const int *seat = seat_data;
   int fd, time;
   char op;
-  void *data;
-  struct timeval now;
+  void *data = NULL;
 
   /* Is the seat valid? */
   fd = ggzdmod_get_seat(ggz, *seat).fd;
@@ -411,8 +526,7 @@ void game_handle_player_data(GGZdMod *ggz, GGZdModEvent id, void *seat_data) {
           /* Now put it on the data buffer */
         } else {
           /* How much type did that player take ? */
-          gettimeofday(&now, NULL);
-          time = now.tv_sec - cronometer.tv_sec;
+          time = game_read_chronometer();
         }
         /* If first turn, there is no time! */
         if (game_info.turn == 0 || game_info.turn == 1)
@@ -449,8 +563,7 @@ void game_handle_player_data(GGZdMod *ggz, GGZdModEvent id, void *seat_data) {
       /* If server clock, get the time until now*/
       if (game_info.clock_type == CHESS_CLOCK_SERVER ||
           game_info.clock_type == CHESS_CLOCK_SERVERLAG) {
-        gettimeofday(&now, NULL);
-        time = now.tv_sec - cronometer.tv_sec;
+        time = game_read_chronometer();
       }
       /* Check if this player can claim */
       if (game_info.turn % 2 == *seat && game_info.seconds[*seat] - time <= 0)
@@ -467,10 +580,12 @@ void game_handle_player_data(GGZdMod *ggz, GGZdModEvent id, void *seat_data) {
       game_update(CHESS_EVENT_DRAW, &id);
       break;
   }
-  return;
+  if (data)
+    free(data);
 }
 
-void game_send_seat(int seat) {
+static void game_send_seat(int seat)
+{
   int fd = ggzdmod_get_seat(game_info.ggz, seat).fd;
   if (fd < 0)
     return;
@@ -480,7 +595,8 @@ void game_send_seat(int seat) {
     return;
 }
 
-void game_send_players(void) {
+static void game_send_players(void)
+{
 	int i, j, fd;
 	
 	for (j = 0; j < 2; j++) {
@@ -504,11 +620,21 @@ void game_send_players(void) {
 	}
 }
 
-void game_stop_cronometer(void) {
-  gettimeofday(&cronometer, NULL);
+static void game_restart_chronometer(void)
+{
+  gettimeofday(&chronometer, NULL);
 }
 
-void game_request_time(int seat) {
+static int game_read_chronometer(void)
+{
+  struct timeval now;
+
+  gettimeofday(&now, NULL);
+  return now.tv_sec - chronometer.tv_sec;
+}
+
+static void game_request_time(int seat)
+{
   int fd = ggzdmod_get_seat(game_info.ggz, seat).fd;
 
   if (fd < 0)
@@ -517,19 +643,22 @@ void game_request_time(int seat) {
   ggz_write_char(fd, CHESS_REQ_TIME);
 }
 
-void game_send_time(int seat) {
+static void game_send_time(int seat)
+{
   int fd = ggzdmod_get_seat(game_info.ggz, seat).fd;
+  int clock_type = game_info.clock_type;
+  int sec = game_info.seconds[0];
 
   if (fd < 0)
     return;
 
-  if (ggz_write_char(fd, CHESS_RSP_TIME) < 0 ||
-      ggz_write_int(fd, (game_info.clock_type<<24)+(game_info.seconds[0]&0xFFFFFF)))
+  if (ggz_write_char(fd, CHESS_RSP_TIME) < 0
+      || ggz_write_int(fd, (clock_type << 24) + (sec & 0xFFFFFF)))
     return;
-
 }
 
-void game_send_start(void) {
+static void game_send_start(void)
+{
   int fd, a;
 
   for (a = 0; a < 2; a++) {
@@ -542,7 +671,8 @@ void game_send_start(void) {
   }
 }
 
-void game_send_move(char *move, int time) {
+static void game_send_move(char *move, int time)
+{
   int fd, a;
 
   for (a = 0; a < 2; a++) {
@@ -570,13 +700,11 @@ void game_send_move(char *move, int time) {
     /* Send time, if not error */
     if (game_info.clock_type && move)
       ggz_write_int(fd, time);
-
   }
-
-
 }
 
-void game_send_update(void) {
+static void game_send_update(void)
+{
   int fd, a;
   for (a = 0; a < 2; a++) {
     fd = ggzdmod_get_seat(game_info.ggz, a).fd;
@@ -590,16 +718,24 @@ void game_send_update(void) {
   }
 }
 
-void game_send_gameover(char code) {
+static void game_send_gameover(char code)
+{
   int fd, a;
   GGZGameResult results[2];
   int winner = (code == CHESS_GAMEOVER_WIN_1_MATE) ? 0 : 1;
+  char *event, *site, *date, *round, *white, *black, *result;
+  time_t t;
+  move_t move, move2;
+  move_t *walker;
+  int i;
+  char movestr[8], movestr2[8];
 
   /* Report results to GGZ. */
   results[winner] = GGZ_GAME_WIN;
   results[1 - winner] = GGZ_GAME_LOSS;
-  ggzdmod_report_game(game_info.ggz, NULL, results);
+  ggzdmod_report_game(game_info.ggz, NULL, results, NULL);
 
+  /* Notify players */
   for (a = 0; a < 2; a++) {
     fd = ggzdmod_get_seat(game_info.ggz, a).fd;
 
@@ -609,9 +745,64 @@ void game_send_gameover(char code) {
     ggz_write_char(fd, CHESS_MSG_GAMEOVER);
     ggz_write_char(fd, code);
   }
+
+  /* Write PGN savegame headers */
+  event = "some game...";
+  site = "GGZ";
+  date = ggz_malloc(32);
+  strftime(date, 32, "%Y.%m.%d", localtime(&t));
+  round = "1";
+  white = ggz_strdup(ggzdmod_get_seat(game_info.ggz, 0).name);
+  black = ggz_strdup(ggzdmod_get_seat(game_info.ggz, 1).name);
+  result = ggz_malloc(8);
+  if(winner == 0) strncpy(result, "1-0", sizeof(result));
+  else strncpy(result, "0-1", sizeof(result));
+  /* '1-0' '1/2-1/2' '0-1' '*' */
+
+  game_save("[Event \"%s\"]", event);
+  game_save("[Site \"%s\"]", site);
+  game_save("[Date \"%s\"]", date);
+  game_save("[Round \"%s\"]", round);
+  game_save("[White \"%s\"]", white);
+  game_save("[Black \"%s\"]", black);
+  game_save("[Result \"%s\"]", result);
+  game_save("");
+
+  ggz_free(date);
+  ggz_free(result);
+  ggz_free(black);
+  ggz_free(white);
+
+  /* Write PGN savegame moves */
+  /* TODO: keep track of moves */
+  if(game->movelist && game->movelist->next) {
+    walker = game->movelist->next;
+	i = 1;
+    while(walker) {
+      move = *walker;
+      walker = walker->next;
+	  if(walker) {
+        move2 = *walker;
+        walker = walker->next;
+      } else {
+        move2.fs = -1;
+      }
+      snprintf(movestr, sizeof(movestr), "%c%c", move.rs + 'a', move.rd + '0');
+      snprintf(movestr2, sizeof(movestr2), "%c%c", move2.rs + 'a', move2.rd + '0');
+      game_save("%i. %s %s", i, movestr, movestr2);
+	  i += 1;
+    }
+  }
+
+  /* Finish savegame */
+  if (savegame) {
+    fclose(savegame);
+    savegame = NULL;
+  }
 }
 
-void game_send_draw(int seat) {
+static void game_send_draw(int seat)
+{
   int fd = ggzdmod_get_seat(game_info.ggz, seat).fd;
 
   if (fd < 0)
@@ -619,3 +810,28 @@ void game_send_draw(int seat) {
 
   ggz_write_char(fd, CHESS_REQ_DRAW);
 }
+
+static void game_save(char *fmt, ...)
+{
+  int fd;
+  char *savegamepath;
+  char buffer[1024];
+  va_list ap;
+
+  if(!savegame) {
+    savegamepath = ggz_strdup(DATADIR "/gamedata/Chess/savegame.pgn.XXXXXX");
+    fd = mkstemp(savegamepath);
+    ggz_free(savegamepath);
+    if(fd < 0) return;
+    savegame = fdopen(fd, "w");
+    if(!savegame) return;
+  }
+
+  va_start(ap, fmt);
+  vsnprintf(buffer, sizeof(buffer), fmt, ap);
+  va_end(ap);
+
+  fprintf(savegame, "%s\n", buffer);
+  fflush(savegame);
+}
+
