@@ -4,7 +4,7 @@
  * Project: GGZ ConnectX game module
  * Date: 27th June 2001
  * Desc: Game functions
- * $Id: game.c,v 1.2 2002/12/20 10:10:29 oojah Exp $
+ * $Id: game.c 7705 2005-12-29 11:53:14Z josef $
  *
  * Copyright (C) 2000 Brent Hendricks.
  *
@@ -33,13 +33,8 @@
 
 #include "game.h"
 
-#define GGZSTATISTICS
 #define GGZSPECTATORS
 
-#ifdef GGZSTATISTICS
-#include "ggz_stats.h"
-#endif
-
 /* Data structure for ConnectX game */
 struct connectx_game_t {
 	GGZdMod *ggz; /* GGZ data object */
@@ -52,7 +47,7 @@ struct connectx_game_t {
 	int optionsseat;
 };
 
-/* Escape protocol */
+/* ConnectX protocol */
 /* Messages from server */
 #define CONNECTX_MSG_SEAT     0
 #define CONNECTX_MSG_PLAYERS  1
@@ -65,18 +60,12 @@ struct connectx_game_t {
 #define CONNECTX_REQ_OPTIONS  8
 #define CONNECTX_MSG_CHAT     9
 #define CONNECTX_RSP_CHAT    10
-#ifdef GGZSTATISTICS
-#define CONNECTX_SND_STATS    11
-#endif
 
 /* Messages from client */
 #define CONNECTX_SND_MOVE     0
 #define CONNECTX_REQ_SYNC     1
 #define CONNECTX_SND_OPTIONS  2
 #define CONNECTX_REQ_NEWGAME  3
-#ifdef GGZSTATISTICS
-#define CONNECTX_REQ_STATS    4
-#endif
 
 /* Move errors */
 #define CONNECTX_ERR_STATE   -1
@@ -84,7 +73,7 @@ struct connectx_game_t {
 #define CONNECTX_ERR_BOUND   -3
 #define CONNECTX_ERR_FULL    -4
 
-/* Escape game states */
+/* ConnectX game states */
 #define CONNECTX_STATE_INIT        0 /* Table has been started, no players joined */
 #define CONNECTX_STATE_OPTIONS     1 /* Options have been requested */
 #define CONNECTX_STATE_WAITING     2 /* Waiting for all players to join, options have been received */
@@ -99,9 +88,6 @@ static int game_send_seat(int seat);
 static int game_send_players(void);
 static int game_send_move(int num, int move);
 static int game_send_sync(int fd);
-#ifdef GGZSTATISTICS
-static int game_send_stats(int num);
-#endif
 static int game_send_gameover(int winner);
 static int game_read_move(int num, int* move);
 
@@ -133,7 +119,7 @@ void game_init(GGZdMod *ggz)
 	/* Board is set up after the options have been collected */
 }
 
-void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	switch(ggzdmod_get_state(ggz)) {
 	case GGZDMOD_STATE_PLAYING:
@@ -146,9 +132,11 @@ void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data)
 	}
 }
 
-static int seats_full(void){
+static int seats_full(void)
+{
 	return ggzdmod_count_seats(connectx_game.ggz, GGZ_SEAT_OPEN) == 0
-		&& ggzdmod_count_seats(connectx_game.ggz, GGZ_SEAT_RESERVED) == 0;
+	  && ggzdmod_count_seats(connectx_game.ggz, GGZ_SEAT_RESERVED) == 0
+	  && ggzdmod_count_seats(connectx_game.ggz, GGZ_SEAT_ABANDONED) == 0;
 }
 
 static int seats_empty(void){
@@ -157,10 +145,10 @@ static int seats_empty(void){
 }
 
 #ifdef GGZSPECTATORS
-void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	int i, fd;
-	GGZSpectator *old_spectator = data;
+	const GGZSpectator *old_spectator = data;
 	GGZSpectator spectator;
 	GGZSeat seat;
 
@@ -181,7 +169,7 @@ void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, void *data
 	}
 }
 
-void game_handle_ggz_spectator_leave(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator_leave(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	if(seats_empty())
 		ggzdmod_set_state(connectx_game.ggz, GGZDMOD_STATE_DONE);
@@ -189,10 +177,10 @@ void game_handle_ggz_spectator_leave(GGZdMod *ggz, GGZdModEvent event, void *dat
 #endif
 
 /* Callback for ggzdmod JOIN, LEAVE and SEAT events */
-void game_handle_ggz_seat(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_seat(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	GGZdModState new_state;
-	GGZSeat *old_seat = data;
+	const GGZSeat *old_seat = data;
 	GGZSeat new_seat = ggzdmod_get_seat(ggz, old_seat->num);
 
 	if(seats_full()){
@@ -228,7 +216,7 @@ void game_handle_ggz_seat(GGZdMod *ggz, GGZdModEvent event, void *data)
 }
 
 /* Handle message from player */
-void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	int num = *(int*)data;
 	int fd, op, move;
@@ -259,11 +247,6 @@ void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, void *data)
 //		game_handle_newgame(num);
 		break;
 
-#ifdef GGZSTATISTICS
-	case CONNECTX_REQ_STATS:
-		game_send_stats(num);
-		break;
-#endif
 	default:
 		ggzdmod_log(ggz, "Unrecognised player opcode %d.", op);
 	}
@@ -271,7 +254,7 @@ void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, void *data)
 
 #ifdef GGZSPECTATORS
 /* handle message from spectator */
-void game_handle_ggz_spectator(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	int num = *(int*)data;
 	int fd, op;
@@ -354,31 +337,6 @@ static int game_send_move(int num, int column)
 	return 0;
 }
 
-#ifdef GGZSTATISTICS
-/* Send game statistics */
-static int game_send_stats(int num)
-{
-	int wins, losses, ties;
-	GGZStats *stats;
-	GGZSeat seat;
-
-	seat = ggzdmod_get_seat(connectx_game.ggz, num);
-
-	ggzdmod_log(connectx_game.ggz, "Handling statistics for player %d", num);
-
-	stats = ggzstats_new(connectx_game.ggz);
-	ggzstats_get_record(stats, num, &wins, &losses, &ties);
-	ggzstats_free(stats);
-
-	if(ggz_write_int(seat.fd, CONNECTX_SND_STATS) < 0
-		|| ggz_write_int(seat.fd, wins) < 0
-		|| ggz_write_int(seat.fd, losses) < 0
-	   || ggz_write_int(seat.fd, ties) < 0)
-		return -1;
-	return 0;
-}
-#endif
-
 /* Send out board layout */
 static int game_send_sync(int fd)
 {
@@ -409,22 +367,7 @@ static int game_send_gameover(int winner)
 #ifdef GGZSPECTATORS
 	GGZSpectator spectator;
 #endif
-
-#ifdef GGZSTATISTICS
-	GGZStats *stats;
-
-	stats = ggzstats_new(connectx_game.ggz);
-	if(winner < 2) {
-		ggzstats_set_game_winner(stats, winner, 1.0);
-		ggzstats_set_game_winner(stats, (winner + 1) % 2, 0.0);
-	} else{
-		/* Draw */
-		ggzstats_set_game_winner(stats, 0, 0.5);
-		ggzstats_set_game_winner(stats, 1, 0.5);
-	}
-	ggzstats_recalculate_ratings(stats);
-	ggzstats_free(stats);
-#endif
+	GGZGameResult results[2];
 
 	for(i = 0; i < 2; i++){
 		seat = ggzdmod_get_seat(connectx_game.ggz, i);
@@ -447,6 +390,17 @@ static int game_send_gameover(int winner)
 	}
 #endif
 
+	/* Report the results to GGZ. */
+	if (winner < 2) {
+		/* One player won. */
+		results[winner] = GGZ_GAME_WIN;
+		results[1 - winner] = GGZ_GAME_LOSS;
+	} else {
+		/* Draw. */
+		results[0] = results[1] = GGZ_GAME_TIE;
+	}
+	ggzdmod_report_game(connectx_game.ggz, NULL, results, NULL);
+
 	return 0;
 }
 
