@@ -4,7 +4,7 @@
  * Project: GGZ Connect the Dots game module
  * Date: 04/27/2000
  * Desc: Game functions
- * $Id: game.c,v 1.23 2003/02/15 03:45:28 jdorje Exp $
+ * $Id: game.c 7748 2006-01-07 22:02:13Z jdorje $
  *
  * Copyright (C) 2000 Brent Hendricks.
  *
@@ -28,6 +28,9 @@
 #endif
 
 #include <stdlib.h>
+#include <stdio.h>
+#include <stdarg.h>
+#include <string.h>
 #include <ggz.h>
 
 #include "game.h"
@@ -37,10 +40,12 @@
 struct dots_game_t dots_game;
 int s_x[2], s_y[2];
 int score;
+static FILE *savegame = NULL;
 
 /* Private functions */
 static int game_get_options(int);
 static int game_handle_newgame(int);
+static void game_save(char *fmt, ...);
 
 
 /* Setup game state and board */
@@ -61,29 +66,36 @@ void game_init(GGZdMod *ggz)
 			dots_game.owners_board[i][j] = -1;
 	dots_game.score[0] = dots_game.score[1] = 0;
 	dots_game.ggz = ggz;
+
+	game_save(NULL);
 }
 
 
 /* Handle message from GGZ server */
-void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
-	GGZdModState old_state = *(GGZdModState*)data;
+	const GGZdModState *old_state = data;
 	
-	if (old_state == GGZDMOD_STATE_CREATED)
+	if (*old_state == GGZDMOD_STATE_CREATED)
 		game_update(DOTS_EVENT_LAUNCH, NULL, NULL);
 }
 
 
-void game_handle_ggz_join(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_join(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
-	int player = ((GGZSeat*)data)->num;
-	game_update(DOTS_EVENT_JOIN, &player, NULL);
+	const GGZSeat *old_seat = data;
+	GGZSeat new_seat = ggzdmod_get_seat(ggz, old_seat->num);
+
+	game_update(DOTS_EVENT_JOIN, &old_seat->num, NULL);
+
+	game_save("player%i join %s", new_seat.num + 1, new_seat.name);
 }
 
-void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator_join(GGZdMod *ggz,
+				    GGZdModEvent event, const void *data)
 {
-	int spectator = *(int*)data;
-	int fd = ggzdmod_get_spectator(dots_game.ggz, spectator).fd;
+	const int *spectator = data;
+	int fd = ggzdmod_get_spectator(dots_game.ggz, *spectator).fd;
 
 	game_send_seat_spectators(fd);
 
@@ -96,17 +108,22 @@ void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, void *data
 	}
 }
 
-void game_handle_ggz_leave(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_leave(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
-	int player = ((GGZSeat*)data)->num;
-	game_update(DOTS_EVENT_LEAVE, &player, NULL);
+	const GGZSeat *old_seat = data;
+
+	game_update(DOTS_EVENT_LEAVE, &old_seat->num, NULL);
+
+	game_save("player%i leave %s", old_seat->num + 1, old_seat->name);
 }
 
 
 /* Handle message from player */
-void game_handle_player_data(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_player_data(GGZdMod *ggz,
+			     GGZdModEvent event, const void *data)
 {
-	int num = *(int*)data;
+	const int *num_ptr = data;
+	const int num = *num_ptr;
 	int fd, op, status;
 	unsigned char x, y;
 
@@ -143,7 +160,7 @@ void game_handle_player_data(GGZdMod *ggz, GGZdModEvent event, void *data)
 
 
 /* Handle message from spectator */
-void game_handle_spectator_data(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_spectator_data(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	int num = *(int*)data;
 	int fd, op, status;
@@ -177,10 +194,13 @@ static int game_get_options(int seat)
 {
 	int fd = ggzdmod_get_seat(dots_game.ggz, seat).fd;
 
-	if(ggz_read_char(fd, &dots_game.board_width) < 0
-	   || ggz_read_char(fd, &dots_game.board_height) < 0)
+	if(ggz_read_char(fd, (char*)&dots_game.board_width) < 0
+	   || ggz_read_char(fd, (char*)&dots_game.board_height) < 0)
 		return -1;
 
+	game_save("width %i", dots_game.board_width);
+	game_save("height %i", dots_game.board_height);
+
 	return game_update(DOTS_EVENT_OPTIONS, NULL, NULL);
 }
 
@@ -392,6 +412,12 @@ int game_send_gameover(char winner)
 	int i, fd;
 	GGZGameResult results[2];
 
+	if (winner == 2) {
+		game_save("tie game");
+	} else {
+		game_save("player%i winner", winner + 1);
+	}
+
 	/* Report game to GGZ. */
 	if (winner == 2) {
 		results[0] = results[1] = GGZ_GAME_TIE;
@@ -399,7 +425,7 @@ int game_send_gameover(char winner)
 		results[(int)winner] = GGZ_GAME_WIN;
 		results[1 - winner] = GGZ_GAME_LOSS;
 	}
-	ggzdmod_report_game(dots_game.ggz, NULL, results);
+	ggzdmod_report_game(dots_game.ggz, NULL, results, NULL);
 	
 	for(i=0; i<2; i++) {
 		if((fd = ggzdmod_get_seat(dots_game.ggz, i).fd) == -1)
@@ -428,6 +454,13 @@ int game_move(void)
 			game_update(DOTS_EVENT_MOVE_V, &x, &y);
 		else
 			game_update(DOTS_EVENT_MOVE_H, &x, &y);
+
+		/* Break out of recursion here */
+		if(dots_game.state == DOTS_STATE_INIT) return 0;
+
+		game_save("player%i move %u %u %s",
+			(dots_game.turn + 1) % 2 + 1, x, y,
+			(dir == 0 ? "vertical" : "horizontal"));
 	} else
 		game_req_move(num);
 
@@ -455,9 +488,9 @@ int game_handle_move(int num, int dir, unsigned char *x, unsigned char *y)
 	char status;
 	
 	ggzdmod_log(dots_game.ggz, "Handling move for player %d", num);
-	if(ggz_read_char(fd, x) < 0)
+	if(ggz_read_char(fd, (char*)x) < 0)
 		return -1;
-	if(ggz_read_char(fd, y) < 0)
+	if(ggz_read_char(fd, (char*)y) < 0)
 		return -1;
 
 	score = 0;
@@ -536,6 +569,10 @@ int game_handle_move(int num, int dir, unsigned char *x, unsigned char *y)
 
 	if(status < 0)
 		return 1;
+
+	game_save("player%i move %u %u %s",
+		dots_game.turn + 1, *x, *y,
+		(dir == 0 ? "vertical" : "horizontal"));
 	
 	/* We make a note who our opponent is, easier on the update func */
 	dots_game.opponent = (num + 1) % 2;
@@ -585,13 +622,14 @@ char game_check_win(void)
 
 static int seats_full(void)
 {
-	return ggzdmod_count_seats(dots_game.ggz, GGZ_SEAT_OPEN)
-		+ ggzdmod_count_seats(dots_game.ggz, GGZ_SEAT_RESERVED) == 0;
+	return ggzdmod_count_seats(dots_game.ggz, GGZ_SEAT_OPEN) == 0
+	  && ggzdmod_count_seats(dots_game.ggz, GGZ_SEAT_RESERVED) == 0
+	  && ggzdmod_count_seats(dots_game.ggz, GGZ_SEAT_ABANDONED) == 0;
 }
 
 
 /* Update game state */
-int game_update(int event, void *d1, void *d2)
+int game_update(int event, const void *d1, void *d2)
 {
 	int seat, fd, spectator;
 	char x, y;
@@ -728,3 +766,40 @@ static int game_handle_newgame(int seat)
 
 	return status;
 }
+
+
+static void game_save(char *fmt, ...)
+{
+	int fd;
+	char *savegamepath, *savegamename;
+	char buffer[1024];
+	va_list ap;
+
+	if(!fmt) {
+		if(savegame) {
+			fclose(savegame);
+			savegame = NULL;
+		}
+		return;
+	}
+
+	if(!savegame) {
+		savegamepath = ggz_strdup(DATADIR "/gamedata/Dots/savegame.XXXXXX");
+		fd = mkstemp(savegamepath);
+		if(fd < 0) return;
+		savegame = fdopen(fd, "w");
+		if(!savegame) return;
+
+		savegamename = savegamepath + strlen(savegamepath) - strlen("savegame.XXXXXX");
+		ggzdmod_report_savegame(dots_game.ggz, savegamename);
+		ggz_free(savegamepath);
+	}
+
+	va_start(ap, fmt);
+	vsnprintf(buffer, sizeof(buffer), fmt, ap);
+	va_end(ap);
+
+	fprintf(savegame, "%s\n", buffer);
+	fflush(savegame);
+}
+
