@@ -4,7 +4,7 @@
  * Project: GGZ Escape game module
  * Date: 27th June 2001
  * Desc: Game functions
- * $Id: game.c,v 1.41 2002/12/18 11:46:17 oojah Exp $
+ * $Id: game.c 7107 2005-04-15 17:54:31Z jdorje $
  *
  * Copyright (C) 2000 Brent Hendricks.
  *
@@ -33,13 +33,8 @@
 
 #include "game.h"
 
-//#define GGZSTATISTICS
 //#define GGZSPECTATORS
 
-#ifdef GGZSTATISTICS
-#include "ggz_stats.h"
-#endif
-
 /* Data structure for Escape game */
 struct escape_game_t {
 	GGZdMod *ggz; /* GGZ data object */
@@ -69,18 +64,12 @@ struct escape_game_t {
 #define ESCAPE_REQ_OPTIONS  8
 #define ESCAPE_MSG_CHAT     9
 #define ESCAPE_RSP_CHAT    10
-#ifdef GGZSTATISTICS
-#define ESCAPE_SND_STATS    11
-#endif
 
 /* Messages from client */
 #define ESCAPE_SND_MOVE     0
 #define ESCAPE_REQ_SYNC     1
 #define ESCAPE_SND_OPTIONS  2
 #define ESCAPE_REQ_NEWGAME  3
-#ifdef GGZSTATISTICS
-#define ESCAPE_REQ_STATS    4
-#endif
 
 /* Move errors */
 #define ESCAPE_ERR_STATE   -1
@@ -103,9 +92,6 @@ static int game_send_seat(int seat);
 static int game_send_players(void);
 static int game_send_move(int num, int move);
 static int game_send_sync(int fd);
-#ifdef GGZSTATISTICS
-static int game_send_stats(int num);
-#endif
 static int game_send_gameover(int winner);
 static int game_read_move(int num, int* move);
 
@@ -138,7 +124,7 @@ void game_init(GGZdMod *ggz)
 	/* Board is set up after the options have been collected */
 }
 
-void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	switch(ggzdmod_get_state(ggz)) {
 	case GGZDMOD_STATE_PLAYING:
@@ -151,9 +137,11 @@ void game_handle_ggz_state(GGZdMod *ggz, GGZdModEvent event, void *data)
 	}
 }
 
-static int seats_full(void){
+static int seats_full(void)
+{
 	return ggzdmod_count_seats(escape_game.ggz, GGZ_SEAT_OPEN) == 0
-		&& ggzdmod_count_seats(escape_game.ggz, GGZ_SEAT_RESERVED) == 0;
+	  && ggzdmod_count_seats(escape_game.ggz, GGZ_SEAT_RESERVED) == 0
+	  && ggzdmod_count_seats(escape_game.ggz, GGZ_SEAT_ABANDONED) == 0;
 }
 
 static int seats_empty(void){
@@ -162,10 +150,11 @@ static int seats_empty(void){
 }
 
 #ifdef GGZSPECTATORS
-void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator_join(GGZdMod *ggz,
+				    GGZdModEvent event, const void *data)
 {
 	int i, fd;
-	GGZSpectator *old_spectator = data;
+	const GGZSpectator *old_spectator = data;
 	GGZSpectator spectator;
 	GGZSeat seat;
 
@@ -186,7 +175,8 @@ void game_handle_ggz_spectator_join(GGZdMod *ggz, GGZdModEvent event, void *data
 	}
 }
 
-void game_handle_ggz_spectator_leave(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator_leave(GGZdMod *ggz,
+				     GGZdModEvent event, const void *data)
 {
 	if(seats_empty())
 		ggzdmod_set_state(escape_game.ggz, GGZDMOD_STATE_DONE);
@@ -194,10 +184,10 @@ void game_handle_ggz_spectator_leave(GGZdMod *ggz, GGZdModEvent event, void *dat
 #endif
 
 /* Callback for ggzdmod JOIN, LEAVE and SEAT events */
-void game_handle_ggz_seat(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_seat(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
 	GGZdModState new_state;
-	GGZSeat *old_seat = data;
+	const GGZSeat *old_seat = data;
 	GGZSeat new_seat = ggzdmod_get_seat(ggz, old_seat->num);
 
 	if(seats_full()){
@@ -233,7 +223,8 @@ void game_handle_ggz_seat(GGZdMod *ggz, GGZdModEvent event, void *data)
 }
 
 /* Handle message from player */
-void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_player(GGZdMod *ggz,
+			    GGZdModEvent event, const void *data)
 {
 	int num = *(int*)data;
 	int fd, op, move;
@@ -264,11 +255,6 @@ void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, void *data)
 //		game_handle_newgame(num);
 		break;
 
-#ifdef GGZSTATISTICS
-	case ESCAPE_REQ_STATS:
-		game_send_stats(num);
-		break;
-#endif
 	default:
 		ggzdmod_log(ggz, "Unrecognised player opcode %d.", op);
 	}
@@ -276,7 +262,8 @@ void game_handle_ggz_player(GGZdMod *ggz, GGZdModEvent event, void *data)
 
 #ifdef GGZSPECTATORS
 /* handle message from spectator */
-void game_handle_ggz_spectator(GGZdMod *ggz, GGZdModEvent event, void *data)
+void game_handle_ggz_spectator(GGZdMod *ggz,
+			       GGZdModEvent event, const void *data)
 {
 	int num = *(int*)data;
 	int fd, op;
@@ -359,31 +346,6 @@ static int game_send_move(int num, int direction)
 	return 0;
 }
 
-#ifdef GGZSTATISTICS
-/* Send game statistics */
-static int game_send_stats(int num)
-{
-	int wins, losses, ties;
-	GGZStats *stats;
-	GGZSeat seat;
-
-	seat = ggzdmod_get_seat(escape_game.ggz, num);
-
-	ggzdmod_log(escape_game.ggz, "Handling statistics for player %d", num);
-
-	stats = ggzstats_new(escape_game.ggz);
-	ggzstats_get_record(stats, num, &wins, &losses, &ties);
-	ggzstats_free(stats);
-
-	if(ggz_write_int(seat.fd, ESCAPE_SND_STATS) < 0
-		|| ggz_write_int(seat.fd, wins) < 0
-		|| ggz_write_int(seat.fd, losses) < 0
-	   || ggz_write_int(seat.fd, ties) < 0)
-		return -1;
-	return 0;
-}
-#endif
-
 /* Send out board layout */
 static int game_send_sync(int fd)
 {
@@ -417,26 +379,7 @@ static int game_send_gameover(int winner)
 #ifdef GGZSPECTATORS
 	GGZSpectator spectator;
 #endif
-
-#ifdef GGZSTATISTICS
-	GGZStats *stats;
-
-	stats = ggzstats_new(escape_game.ggz);
-	if(winner < 2) {
-		ggzstats_set_game_winner(stats, winner, 1.0);
-		ggzstats_set_game_winner(stats, (winner + 1) % 2, 0.0);
-	} else if(winner == 2){
-		/* Draw */
-		ggzstats_set_game_winner(stats, 0, 0.5);
-		ggzstats_set_game_winner(stats, 1, 0.5);
-	} else {
-		/* One player "commited suicide" */
-		ggzstats_set_game_winner(stats, (winner - 3), 0.5);
-		ggzstats_set_game_winner(stats, ((winner - 3) + 1) % 2, -0.5);
-	}
-	ggzstats_recalculate_ratings(stats);
-	ggzstats_free(stats);
-#endif
+	GGZGameResult results[2];
 
 	for(i = 0; i < 2; i++){
 		seat = ggzdmod_get_seat(escape_game.ggz, i);
@@ -459,6 +402,21 @@ static int game_send_gameover(int winner)
 	}
 #endif
 
+	/* Report the results to GGZ. */
+	if (winner < 2) {
+		/* One player won. */
+		results[winner] = GGZ_GAME_WIN;
+		results[1 - winner] = GGZ_GAME_LOSS;
+	} else if (winner == 2) {
+		/* Draw. */
+		results[0] = results[1] = GGZ_GAME_TIE;
+	} else {
+		/* One player forfeited. */
+		results[winner - 3] = GGZ_GAME_WIN;
+		results[4 - winner] = GGZ_GAME_FORFEIT;
+	}
+	ggzdmod_report_game(escape_game.ggz, NULL, results, NULL);
+
 	return 0;
 }
 
