@@ -4,7 +4,7 @@
  * Project: GGZCards Server
  * Date: 06/20/2001
  * Desc: Game-independent game functions
- * $Id: common.c,v 1.212 2002/10/27 12:49:52 jdorje Exp $
+ * $Id: common.c 7108 2005-04-15 18:31:59Z jdorje $
  *
  * This file contains code that controls the flow of a general
  * trick-taking game.  Game states, event handling, etc. are all
@@ -47,8 +47,6 @@
 #include "net.h"
 #include "options.h"
 
-#include "ggz_stats.h"
-
 
 /* Global game variables */
 game_t game = {
@@ -59,7 +57,8 @@ bool seats_full(void)
 {
 	/* This calculation is a bit inefficient, but that's OK */
 	return ggzdmod_count_seats(game.ggz, GGZ_SEAT_OPEN) == 0
-		&& ggzdmod_count_seats(game.ggz, GGZ_SEAT_RESERVED) == 0;
+	  && ggzdmod_count_seats(game.ggz, GGZ_SEAT_RESERVED) == 0
+	  && ggzdmod_count_seats(game.ggz, GGZ_SEAT_ABANDONED) == 0;
 }
 
 bool seats_empty(void)
@@ -113,18 +112,18 @@ void set_game_state(server_state_t state)
 
 /* Handle message from player.  */
 void handle_ggz_player_data_event(GGZdMod * ggz,
-				  GGZdModEvent event, void *data)
+				  GGZdModEvent event, const void *data)
 {
-	player_t p = *(int *) data;
-	handle_player_data_event(p);
+	const int *player = data;
+	handle_player_data_event(*player);
 }
 
 /* Handle message from spectator */
 void handle_ggz_spectator_data_event(GGZdMod *ggz,
-				     GGZdModEvent event, void *data)
+				     GGZdModEvent event, const void *data)
 {
-	int spectator = *(int *) data;
-	handle_player_data_event(SPECTATOR_TO_PLAYER(spectator));
+	const int *spectator = data;
+	handle_player_data_event(SPECTATOR_TO_PLAYER(*spectator));
 }
 
 /* Handle message from player/spectator. */
@@ -404,12 +403,13 @@ static void handle_done_event(void)
 }
 
 /* This handles a state change event, when the table changes state. */
-void handle_ggz_state_event(GGZdMod * ggz, GGZdModEvent event, void *data)
+void handle_ggz_state_event(GGZdMod * ggz,
+			    GGZdModEvent event, const void *data)
 {
-	GGZdModState old_state = *(GGZdModState*)data;
+	const GGZdModState *old_state = data;
 	GGZdModState new_state = ggzdmod_get_state(ggz);
 	
-	if (old_state == GGZDMOD_STATE_CREATED) {
+	if (*old_state == GGZDMOD_STATE_CREATED) {
 		assert(new_state == GGZDMOD_STATE_WAITING);
 		handle_launch_event();
 	}
@@ -436,20 +436,20 @@ void handle_ggz_state_event(GGZdMod * ggz, GGZdModEvent event, void *data)
    a player join or a player leave.  Eventually, other types of seat
    changes will be possible (although these are not yet tested).  A lot
    of the work is the same for any kind of seat change, though. */
-void handle_ggz_seat_event(GGZdMod *ggz, GGZdModEvent event, void *data)
+void handle_ggz_seat_event(GGZdMod *ggz, GGZdModEvent event, const void *data)
 {
-	GGZSeat old_seat = *(GGZSeat*)data;
-	player_t player = old_seat.num, p;
+	const GGZSeat *old_seat = data;
+	player_t player = old_seat->num, p;
 	/* seat_t seat = game.players[player].seat; */
 	GGZSeat new_seat = ggzdmod_get_seat(game.ggz, player);
 	bool is_join = (new_seat.type == GGZ_SEAT_PLAYER
 	                || new_seat.type == GGZ_SEAT_BOT)
-	               && (old_seat.type != new_seat.type
-	                   || strcmp(old_seat.name, new_seat.name));
-	bool is_leave = (old_seat.type == GGZ_SEAT_PLAYER
-			 || old_seat.type == GGZ_SEAT_BOT)
-	                && (new_seat.type != old_seat.type
-	                    || strcmp(old_seat.name, new_seat.name));
+	               && (old_seat->type != new_seat.type
+	                   || strcmp(old_seat->name, new_seat.name));
+	bool is_leave = (old_seat->type == GGZ_SEAT_PLAYER
+			 || old_seat->type == GGZ_SEAT_BOT)
+	                && (new_seat.type != old_seat->type
+	                    || strcmp(old_seat->name, new_seat.name));
 	GGZdModState new_state;
 
 	/* There's a big problem here since if the players join/leave before
@@ -460,7 +460,7 @@ void handle_ggz_seat_event(GGZdMod *ggz, GGZdModEvent event, void *data)
 	          player);
 
 	/* Start or stop any AI on the seat.  This needs to be done first. */
-	if (is_leave && old_seat.type == GGZ_SEAT_BOT)
+	if (is_leave && old_seat->type == GGZ_SEAT_BOT)
 		stop_ai(player);
 	if (is_join && new_seat.type == GGZ_SEAT_BOT)
 		start_ai(player, game.ai_module);
@@ -470,11 +470,7 @@ void handle_ggz_seat_event(GGZdMod *ggz, GGZdModEvent event, void *data)
 	   is also sent out in the sync, but there could be a better reason
 	   as well. */
 	net_broadcast_player_list();
-	
-	/* Update stats for this player */
-	if (game.stats)
-		ggzstats_reread(game.stats);
-	
+
 	if (is_join && new_seat.type == GGZ_SEAT_PLAYER) {
 		/* set the age of the player */
 		game.players[player].age = game.player_count;
@@ -516,14 +512,12 @@ void handle_ggz_seat_event(GGZdMod *ggz, GGZdModEvent event, void *data)
 	if (seats_full()) {
 		/* If all seats are full, start playing. */
 		new_state = GGZDMOD_STATE_PLAYING;
+	} else if (seats_empty()) {
+		/* If all seats are empty, the table is done. */
+		new_state = GGZDMOD_STATE_DONE;
 	} else {
-		if (seats_empty()) {
-			/* If all seats are empty, the table is done. */
-			new_state = GGZDMOD_STATE_DONE;
-		} else {
-			/* If some seats are empty, wait for them to fill. */
-			new_state = GGZDMOD_STATE_WAITING;
-		}
+		/* If some seats are empty, wait for them to fill. */
+		new_state = GGZDMOD_STATE_WAITING;
 	}
 	if (ggzdmod_set_state(game.ggz, new_state) < 0)
 		assert(FALSE);
@@ -550,7 +544,7 @@ void handle_ggz_seat_event(GGZdMod *ggz, GGZdModEvent event, void *data)
 }
 
 void handle_ggz_spectator_seat_event(GGZdMod *ggz,
-				     GGZdModEvent event, void *data)
+				     GGZdModEvent event, const void *data)
 {
 	GGZSpectator old = *(GGZSpectator*)data;
 	int spectator = old.num;
@@ -622,17 +616,6 @@ void handle_gameover_event(int winner_cnt, player_t * winners)
 
 	ggz_debug(DBG_MISC, "Handling gameover event.");
 
-	if (game.rated && game.stats) {
-		int i;
-		/* calculate new player ratings */
-		ggzstats_new_game(game.stats);
-		for (i = 0; i < winner_cnt; i++)
-			ggzstats_set_game_winner(game.stats, winners[i],
-					     1.0 / (double) winner_cnt);
-		if (ggzstats_recalculate_ratings(game.stats) < 0)
-			ggz_error_msg("ERROR: couldn't recalculate ratings.");
-	}
-
 	{
 		int teams[game.num_players];
 		GGZGameResult results[game.num_players];
@@ -646,7 +629,7 @@ void handle_gameover_event(int winner_cnt, player_t * winners)
 		for (i = 0; i < winner_cnt; i++)
 			results[winners[i]] = GGZ_GAME_WIN;
 
-		ggzdmod_report_game(game.ggz, teams, results);
+		ggzdmod_report_game(game.ggz, teams, results, NULL);
 	}
 
 	set_game_state(STATE_NOTPLAYING);
@@ -833,13 +816,6 @@ void init_game()
 		game.seats[s].hand.cards =
 			ggz_malloc(game.max_hand_length * sizeof(card_t));
 	}
-	
-	game.stats = ggzstats_new(game.ggz);
-	if (game.stats && game.num_teams > 0) {
-		ggzstats_set_num_teams(game.stats, game.num_teams);
-		for (p = 0; p < game.num_players; p++)
-			ggzstats_set_team(game.stats, p, game.players[p].team);
-	}
 
 	set_global_message("", "%s", "");
 
