@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 02.05.2002
  * Desc: Back-end functions for handling the postgresql style database
- * $Id: ggzdb_pgsql.c,v 1.5 2003/02/16 16:16:58 dr_maux Exp $
+ * $Id: ggzdb_pgsql.c 7900 2006-03-13 12:34:50Z josef $
  *
  * Copyright (C) 2000 Brent Hendricks.
  *
@@ -35,7 +35,11 @@
 #include <unistd.h>
 
 #include <ggz.h>
+#ifdef PGSQL_IN_PGSQLDIR
+#include <pgsql/libpq-fe.h>
+#else
 #include <postgresql/libpq-fe.h>
+#endif
 
 #include "err_func.h"
 #include "ggzd.h"
@@ -53,10 +57,10 @@ static int itercount;
 static pthread_mutex_t mutex;
 static GGZList *list;
 
-static char *dbhost;
-static char *dbname;
-static char *dbusername;
-static char *dbpassword;
+static const char *dbhost;
+static const char *dbname;
+static const char *dbusername;
+static const char *dbpassword;
 
 /* Entry type for the connection pool */
 typedef struct connection_t
@@ -89,6 +93,16 @@ static PGconn *claimconnection()
 			{
 				conn->used = 1;
 				pthread_mutex_unlock(&mutex);
+				if(PQstatus(conn->conn) == CONNECTION_BAD)
+				{
+					err_msg("Database connection lost, reconnecting.");
+					PQreset(conn->conn);
+					if((!conn->conn) || (PQstatus(conn->conn) == CONNECTION_BAD))
+					{
+						err_msg("Could not connect to database.");
+						return NULL;
+					}
+				}
 				return conn->conn;
 			}
 			entry = ggz_list_next(entry);
@@ -106,7 +120,7 @@ static PGconn *claimconnection()
 			conn->conn = PQconnectdb(conninfo);
 			if((!conn->conn) || (PQstatus(conn->conn) == CONNECTION_BAD))
 			{
-				err_sys("Could not connect to database.\n");
+				err_msg("Could not connect to database.");
 				pthread_mutex_unlock(&mutex);
 				return NULL;
 			}
@@ -120,7 +134,7 @@ static PGconn *claimconnection()
 		sleep(1);
 	}
 
-	err_sys("Number of database connections exceeded.\n");
+	err_msg("Number of database connections exceeded.");
 	return NULL;
 }
 
@@ -148,6 +162,82 @@ static void releaseconnection(PGconn *conn)
 	pthread_mutex_unlock(&mutex);
 }
 
+/* Helper function, might go into libggz*/
+char *_ggz_sql_escape(const char *str)
+{
+	char *new, *q;
+	const char *p;
+	size_t len = 0;
+
+	if(str == NULL)
+		return NULL;
+
+	len = strlen(str);
+
+	for(p = str; *p != '\0'; p++) {
+		if(*p == '\'') {
+			len += 1;
+		}
+	}
+
+	if(len == strlen(str))
+		return ggz_strdup(str);
+
+	q = new = ggz_malloc(len + 1);
+	for(p = str; *p != '\0'; p++) {
+		if(*p == '\'') {
+			*q++ = '\\';
+			*q = *p;
+			*q++;
+		} else {
+			*q = *p;
+			q++;
+		}
+	}
+	*q = '\0';
+
+	return new;
+}
+
+/* Helper function, might go into libggz*/
+/*char *_ggz_sql_unescape(const char *str)
+{
+	char *new, *q;
+	const char *p;
+	size_t len = 0;
+
+	if(str == NULL)
+		return NULL;
+
+	len = strlen(str);
+
+	for(p = str; *p != '\0'; p++) {
+		if(!strncmp(p, "\\'", 2)) {
+			p += 1;
+		}
+		len++;
+	}
+
+	if(len == strlen(str))
+		return ggz_strdup(str);
+
+	q = new = ggz_malloc(len + 1);
+	for(p = str; *p != '\0'; p++) {
+		if(!strncmp(p, "\\'", 2)) {
+			*q = '\'';
+			q++;
+			p += 1;
+		} else {
+			*q = *p;
+			q++;
+		}
+	}
+	*q = '\0';
+
+	return new;
+}*/
+
+
 /* Exported functions */
 
 /* Function to initialize the pgsql database system */
@@ -175,9 +265,25 @@ GGZReturn _ggzdb_init(ggzdbConnection connection, int set_standalone)
 
 	rc = GGZDB_NO_ERROR;
 
-	/* Hack: Fire-and-forget table initialization. It might already be
-	 * present, or the database user doesn't have the privileges. */
+	/* Check whether the database is ok */
 	init = 1;
+	res = PQexec(conn, "SELECT value FROM control WHERE key = 'version'");
+	if(PQresultStatus(res) == PGRES_TUPLES_OK)
+	{
+		if(PQntuples(res) == 1)
+		{
+			init = 0;
+			version = PQgetvalue(res, 0, 0);
+			if(strcmp(version, GGZDB_VERSION_ID))
+			{
+				err_msg_exit("Wrong database version: %s present, %s needed.\n", version, GGZDB_VERSION_ID);
+				rc = GGZDB_ERR_DB;
+			}
+		}
+		PQclear(res);
+	}
+
+	/* Initialize the database if needed */
 	if(init)
 	{
 		snprintf(query, sizeof(query), "CREATE TABLE users "
@@ -186,60 +292,56 @@ GGZReturn _ggzdb_init(ggzdbConnection connection, int set_standalone)
 
 		res = PQexec(conn, query);
 
-		/*if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;*/
+		if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;
 		PQclear(res);
 
-
 		snprintf(query, sizeof(query), "CREATE TABLE stats "
 			"(id serial, handle varchar(256), game varchar(256), wins int8, losses int8, ties int8, forfeits int8, "
 			"rating float, ranking int8, highscore int8)");
 
 		res = PQexec(conn, query);
 
-		/*if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;*/
+		if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;
 		PQclear(res);
 
+		snprintf(query, sizeof(query), "CREATE TABLE matches "
+			"(id serial, date int8, game varchar(256), winner varchar(256))");
+
+		res = PQexec(conn, query);
+
+		if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;
+		PQclear(res);
+
+		snprintf(query, sizeof(query), "CREATE TABLE matchplayers "
+			"(id serial, match int8, handle varchar(256), playertype varchar(256))");
+
+		res = PQexec(conn, query);
+
+		if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;
+		PQclear(res);
 
 		snprintf(query, sizeof(query), "CREATE TABLE control "
 			"(key varchar(256), value varchar(256))");
 
 		res = PQexec(conn, query);
 
-		/*if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;*/
+		if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;
 		PQclear(res);
-	}
 
-	/* Check database format version */
-	if(rc == GGZDB_NO_ERROR)
-	{
-		res = PQexec(conn, "SELECT value FROM control WHERE key = 'version'");
-		if(PQresultStatus(res) == PGRES_TUPLES_OK)
-		{
-			if(PQntuples(res) == 1)
-			{
-				version = PQgetvalue(res, 0, 0);
-				if(strcmp(version, GGZDB_VERSION_ID))
-				{
-					err_msg_exit("Wrong database version: %s present, %s needed.\n", version, GGZDB_VERSION_ID);
-					rc = GGZDB_ERR_DB;
-				}
-				PQclear(res);
-			}
-			else
-			{
-				PQclear(res);
-				snprintf(query, sizeof(query), "INSERT INTO control "
-					"(key, value) VALUES ('version', '%s')", GGZDB_VERSION_ID);
-				res = PQexec(conn, query);
-				PQclear(res);
-			}
-		}
-		else rc = GGZDB_ERR_DB;
+		snprintf(query, sizeof(query), "INSERT INTO control "
+			"(key, value) VALUES ('version', '%s')", GGZDB_VERSION_ID);
+
+		res = PQexec(conn, query);
+
+		if(PQresultStatus(res) != PGRES_COMMAND_OK) rc = GGZDB_ERR_DB;
+		PQclear(res);
+
+		if(rc == GGZDB_ERR_DB)
+			err_msg_exit("Could not initialize the database.\n");
 	}
 
 	releaseconnection(conn);
 
-	/* Hack is here ;) */
 	return rc;
 }
 
@@ -277,9 +379,9 @@ void _ggzdb_exit(void)
 
 
 /* Function to initialize the player table */
-GGZDBResult _ggzdb_init_player(char *datadir)
+GGZDBResult _ggzdb_init_player(const char *datadir)
 {
-	return 0;
+	return GGZDB_NO_ERROR;
 }
 
 
@@ -290,6 +392,10 @@ GGZDBResult _ggzdb_player_add(ggzdbPlayerEntry *pe)
 	PGconn *conn;
 	PGresult *res;
 	char query[4096];
+	char *handle_quoted;
+	char *password_quoted;
+	char *name_quoted;
+	char *email_quoted;
 
 	conn = claimconnection();
 	if (!conn) {
@@ -297,12 +403,37 @@ GGZDBResult _ggzdb_player_add(ggzdbPlayerEntry *pe)
 		return GGZDB_ERR_DB;
 	}
 
+	handle_quoted = _ggz_sql_escape(pe->handle);
+	password_quoted = _ggz_sql_escape(pe->password);
+	name_quoted = _ggz_sql_escape(pe->name);
+	email_quoted = _ggz_sql_escape(pe->email);
+
 	snprintf(query, sizeof(query), "INSERT INTO users "
 		 "(handle, password, name, email, lastlogin, permissions) "
 		 "VALUES ('%s', '%s', '%s', '%s', %li, %u)",
-		 pe->handle, pe->password, pe->name, pe->email,
+		 handle_quoted, password_quoted, name_quoted, email_quoted,
 		 pe->last_login, pe->perms);
 
+	if (handle_quoted)
+		ggz_free(handle_quoted);
+	if (password_quoted)
+		ggz_free(password_quoted);
+	if (name_quoted)
+		ggz_free(name_quoted);
+	if (email_quoted)
+		ggz_free(email_quoted);
+
+	res = PQexec(conn, query);
+
+	handle_quoted = _ggz_sql_escape(pe->handle);
+
+	snprintf(query, sizeof(query), "DELETE FROM stats "
+		 "WHERE lower(handle) = lower('%s')",
+		 handle_quoted);
+
+	if (handle_quoted)
+		ggz_free(handle_quoted);
+
 	res = PQexec(conn, query);
 
 	/* FIXME: is this correct?  We won't overwrite, or make a duplicate
@@ -324,6 +455,7 @@ GGZDBResult _ggzdb_player_get(ggzdbPlayerEntry *pe)
 	PGconn *conn;
 	PGresult *res;
 	char query[4096];
+	char *handle_quoted;
 
 	conn = claimconnection();
 	if (!conn) {
@@ -331,11 +463,15 @@ GGZDBResult _ggzdb_player_get(ggzdbPlayerEntry *pe)
 		return GGZDB_ERR_DB;
 	}
 
+	handle_quoted = _ggz_sql_escape(pe->handle);
+
 	snprintf(query, sizeof(query),
 		 "SELECT "
 		 "password, name, email, lastlogin, permissions "
-		 "FROM users WHERE handle = '%s'",
-		 pe->handle);
+		 "FROM users WHERE lower(handle) = lower('%s')",
+		 handle_quoted);
+
+	ggz_free(handle_quoted);
 
 	res = PQexec(conn, query);
 
@@ -353,7 +489,7 @@ GGZDBResult _ggzdb_player_get(ggzdbPlayerEntry *pe)
 			rc = GGZDB_ERR_NOTFOUND;
 		}
 	} else {
-		err_sys("Couldn't lookup player.");
+		err_msg("Couldn't lookup player.");
 		rc = GGZDB_ERR_DB;
 	}
 	PQclear(res);
@@ -371,6 +507,10 @@ GGZDBResult _ggzdb_player_update(ggzdbPlayerEntry *pe)
 	PGconn *conn;
 	PGresult *res;
 	char query[4096];
+	char *handle_quoted;
+	char *password_quoted;
+	char *name_quoted;
+	char *email_quoted;
 
 	conn = claimconnection();
 	if (!conn) {
@@ -378,17 +518,79 @@ GGZDBResult _ggzdb_player_update(ggzdbPlayerEntry *pe)
 		return GGZDB_ERR_DB;
 	}
 
+	handle_quoted = _ggz_sql_escape(pe->handle);
+	password_quoted = _ggz_sql_escape(pe->password);
+	name_quoted = _ggz_sql_escape(pe->name);
+	email_quoted = _ggz_sql_escape(pe->email);
+
 	snprintf(query, sizeof(query),
 		 "UPDATE users SET "
 		 "password = '%s', name = '%s', email = '%s', "
 		 "lastlogin = %li, permissions = %u WHERE "
-		 "handle = '%s'",
-		 pe->password, pe->name, pe->email, pe->last_login,
-		 pe->perms, pe->handle);
+		 "lower(handle) = lower('%s')",
+		 password_quoted, name_quoted, email_quoted,
+		 pe->last_login, pe->perms,
+		 handle_quoted);
+
+	if (handle_quoted)
+		ggz_free(handle_quoted);
+	if (password_quoted)
+		ggz_free(password_quoted);
+	if (name_quoted)
+		ggz_free(name_quoted);
+	if (email_quoted)
+		ggz_free(email_quoted);
 
 	res = PQexec(conn, query);
 	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
-		err_sys("Couldn't update player.");
+		err_msg("Couldn't update player.");
+		rc = GGZDB_ERR_DB;
+	}
+	PQclear(res);
+
+	releaseconnection(conn);
+
+	return rc;
+}
+
+
+/* Function to retrieve an extended player record */
+GGZDBResult _ggzdb_player_get_extended(ggzdbPlayerExtendedEntry *pe)
+{
+	GGZDBResult rc;
+	PGconn *conn;
+	PGresult *res;
+	char query[4096];
+	char *handle_quoted;
+
+	conn = claimconnection();
+	if (!conn) {
+		err_msg("_ggzdb_player_add: couldn't claim connection");
+		return GGZDB_ERR_DB;
+	}
+
+	handle_quoted = _ggz_sql_escape(pe->handle);
+
+	snprintf(query, sizeof(query),
+		 "SELECT "
+		 "id, photo "
+		 "FROM userinfo WHERE lower(handle) = lower('%s')",
+		 handle_quoted);
+
+	ggz_free(handle_quoted);
+
+	res = PQexec(conn, query);
+
+	if (PQresultStatus(res) == PGRES_TUPLES_OK) {
+		if(PQntuples(res) == 1)	{
+			pe->user_id = atol(PQgetvalue(res, 0, 0));
+			strncpy(pe->photo, PQgetvalue(res, 0, 1), sizeof(pe->photo));
+			rc = GGZDB_NO_ERROR;
+		} else	{
+			rc = GGZDB_ERR_NOTFOUND;
+		}
+	} else {
+		err_msg("Couldn't lookup player.");
 		rc = GGZDB_ERR_DB;
 	}
 	PQclear(res);
@@ -443,7 +645,7 @@ GGZDBResult _ggzdb_player_get_first(ggzdbPlayerEntry *pe)
 			rc = GGZDB_ERR_NOTFOUND;
 		}
 	} else {
-		err_sys("Couldn't lookup player.");
+		err_msg("Couldn't lookup player.");
 		PQclear(iterres);
 		iterres = NULL;
 		rc = GGZDB_ERR_DB;
@@ -460,7 +662,7 @@ GGZDBResult _ggzdb_player_get_first(ggzdbPlayerEntry *pe)
 GGZDBResult _ggzdb_player_get_next(ggzdbPlayerEntry *pe)
 {
 	if (!iterres) {
-		err_sys_exit("get_next called before get_first, dummy");
+		err_msg_exit("get_next called before get_first, dummy");
 	}
 
 	if (itercount < PQntuples(iterres) - 1) {
@@ -489,7 +691,7 @@ void _ggzdb_player_drop_cursor(void)
 		/* This isn't an error; since we clear the cursor at the end
 		   of _ggzdb_player_get_next we should expect to end up
 		   here.  --JDS */
-		/*err_sys_exit("drop_cursor called before get_first, dummy");*/
+		/*err_msg_exit("drop_cursor called before get_first, dummy");*/
 		return;
 	}
 
@@ -519,18 +721,23 @@ GGZDBResult _ggzdb_stats_lookup(ggzdbPlayerGameStats *stats)
 	PGresult *res;
 	char query[4096];
 	GGZDBResult rc = GGZDB_ERR_DB;
+	char *player_quoted;
 
 	conn = claimconnection();
 	if(!conn) {
-		err_msg("_ggzdb_player_add: couldn't claim connection");
+		err_msg("_ggzdb_stats_lookup: couldn't claim connection");
 		return rc;
 	}
 
+	player_quoted = _ggz_sql_escape(stats->player);
+
 	snprintf(query, sizeof(query),
 		"SELECT "
 		"wins, losses, ties, forfeits, rating, ranking, highscore "
-		"FROM stats WHERE handle = '%s' AND game = '%s'",
-		stats->player, stats->game);
+		"FROM stats WHERE lower(handle) = lower('%s') AND game = '%s'",
+		player_quoted, stats->game);
+
+	ggz_free(player_quoted);
 
 	res = PQexec(conn, query);
 	if(PQresultStatus(res) == PGRES_TUPLES_OK) {
@@ -544,7 +751,7 @@ GGZDBResult _ggzdb_stats_lookup(ggzdbPlayerGameStats *stats)
 			stats->ranking = atol(PQgetvalue(res, 0, 5));
 			stats->highest_score = atol(PQgetvalue(res, 0, 6));
 		} else {
-			err_sys("couldn't lookup player stats");
+			err_msg("couldn't lookup player stats");
 			rc = GGZDB_ERR_NOTFOUND;
 		}
 	}
@@ -561,37 +768,49 @@ GGZDBResult _ggzdb_stats_update(ggzdbPlayerGameStats *stats)
 	PGresult *res, *res2;
 	char query[4096];
 	int rc = GGZDB_ERR_DB;
+	char *player_quoted;
 
 	conn = claimconnection();
 	if (!conn) {
-		err_msg("_ggzdb_player_add: couldn't claim connection");
+		err_msg("_ggzdb_stats_update: couldn't claim connection");
 		return rc;
 	}
 
+	player_quoted = _ggz_sql_escape(stats->player);
+
 	snprintf(query, sizeof(query),
 		"UPDATE stats "
-		"SET wins = %i, losses = %i, ties = %i, forfeits = %i, rating = %f, ranking = %u, highscore = %li "
-		"WHERE handle = '%s' AND game = '%s'",
-		stats->wins, stats->losses, stats->ties, stats->forfeits, stats->rating, stats->ranking, stats->highest_score,
-		stats->player, stats->game);
+		"SET wins = %i, losses = %i, ties = %i, forfeits = %i, "
+		"rating = %f, ranking = %u, highscore = %li "
+		"WHERE lower(handle) = lower('%s') AND game = '%s'",
+		stats->wins, stats->losses, stats->ties, stats->forfeits,
+		stats->rating, stats->ranking, stats->highest_score,
+		player_quoted, stats->game);
+
+	ggz_free(player_quoted);
 
 	res = PQexec(conn, query);
 
 	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
-		err_sys("couldn't update stats");
+		err_msg("couldn't update stats");
 	} else {
 		if (!strcmp(PQcmdTuples(res), "0")) {
+			player_quoted = _ggz_sql_escape(stats->player);
+
 			snprintf(query, sizeof(query),
 				"INSERT INTO stats "
 				"(handle, game, wins, losses, ties, forfeits, rating, ranking, highscore) VALUES "
 				"('%s', '%s', %i, %i, %i, %i, %f, %u, %li)",
-				stats->player, stats->game, stats->wins, stats->losses, stats->ties, stats->forfeits,
+				player_quoted, stats->game,
+				stats->wins, stats->losses, stats->ties, stats->forfeits,
 				stats->rating, stats->ranking, stats->highest_score);
 
+			ggz_free(player_quoted);
+
 			res2 = PQexec(conn, query);
 
 			if (PQresultStatus(res) != PGRES_COMMAND_OK) {
-				err_sys("couldn't insert stats");
+				err_msg("couldn't insert stats");
 			}
 			else rc = GGZDB_NO_ERROR;
 			PQclear(res2);
@@ -602,6 +821,146 @@ GGZDBResult _ggzdb_stats_update(ggzdbPlayerGameStats *stats)
 
 	releaseconnection(conn);
 
+	_ggzdb_stats_match(stats);
+
+	return rc;
+}
+
+GGZDBResult _ggzdb_stats_match(ggzdbPlayerGameStats *stats)
+{
+	PGconn *conn;
+	PGresult *res;
+	char query[4096];
+	int rc = GGZDB_ERR_DB;
+	char *number, *playertype;
+	char *player_quoted;
+
+	conn = claimconnection();
+	if (!conn) {
+		err_msg("_ggzdb_stats_match: couldn't claim connection");
+		return rc;
+	}
+
+	snprintf(query, sizeof(query),
+		"SELECT MAX(id) FROM matches");
+
+	res = PQexec(conn, query);
+
+	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
+		err_msg("couldn't read match");
+		number = NULL;
+	}
+	else {
+		number = PQgetvalue(res, 0, 0);
+	}
+
+	playertype = "";
+	if(stats->player_type == GGZ_PLAYER_GUEST) playertype = "guest";
+	else if(stats->player_type == GGZ_PLAYER_NORMAL) playertype = "registered";
+	else if(stats->player_type == GGZ_PLAYER_BOT) playertype = "bot";
+
+	player_quoted = _ggz_sql_escape(stats->player);
+
+	snprintf(query, sizeof(query),
+		"INSERT INTO matchplayers "
+		"(match, handle, playertype) VALUES "
+		"(%s, '%s', '%s')",
+		number, player_quoted, playertype);
+
+	ggz_free(player_quoted);
+
+	PQclear(res);
+	res = PQexec(conn, query);
+
+	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
+		err_msg("couldn't insert matchplayer");
+	}
+	else rc = GGZDB_NO_ERROR;
+	PQclear(res);
+
+	releaseconnection(conn);
+
+	return rc;
+}
+
+GGZDBResult _ggzdb_stats_newmatch(const char *game, const char *winner, const char *savegame)
+{
+	PGconn *conn;
+	PGresult *res;
+	char query[4096];
+	int rc = GGZDB_ERR_DB;
+	char *winner_quoted;
+
+	conn = claimconnection();
+	if (!conn) {
+		err_msg("_ggzdb_stats_newmatch: couldn't claim connection");
+		return rc;
+	}
+
+	snprintf(query, sizeof(query),
+		"DELETE FROM savegames "
+		"WHERE game = '%s' AND savegame = '%s'",
+		game, savegame);
+
+	res = PQexec(conn, query);
+
+	winner_quoted = _ggz_sql_escape(winner);
+
+	snprintf(query, sizeof(query),
+		"INSERT INTO matches "
+		"(date, game, winner, savegame) VALUES "
+		"(%li, '%s', '%s', '%s')",
+		time(NULL), game, winner_quoted, savegame);
+
+	ggz_free(winner_quoted);
+
+	res = PQexec(conn, query);
+
+	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
+		err_msg("couldn't insert match");
+	}
+	else rc = GGZDB_NO_ERROR;
+	PQclear(res);
+
+	releaseconnection(conn);
+
+	return rc;
+}
+
+GGZDBResult _ggzdb_stats_savegame(const char *game, const char *owner, const char *savegame)
+{
+	PGconn *conn;
+	PGresult *res;
+	char query[4096];
+	int rc = GGZDB_ERR_DB;
+	char *owner_quoted;
+
+	conn = claimconnection();
+	if (!conn) {
+		err_msg("_ggzdb_stats_savegame: couldn't claim connection");
+		return rc;
+	}
+
+	owner_quoted = _ggz_sql_escape(owner);
+
+	snprintf(query, sizeof(query),
+		"INSERT INTO savegames"
+		"(date, game, owner, savegame) VALUES "
+		"(%li, '%s', '%s', '%s')",
+		time(NULL), game, owner, savegame);
+
+	ggz_free(owner_quoted);
+
+	res = PQexec(conn, query);
+
+	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
+		err_msg("couldn't insert savegame");
+	}
+	else rc = GGZDB_NO_ERROR;
+	PQclear(res);
+
+	releaseconnection(conn);
+
 	return rc;
 }
 
