@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 6/22/00
  * Desc: Functions for handling player logins
- * $Id: login.h,v 1.8 2002/10/20 09:05:31 jdorje Exp $
+ * $Id: login.h 7123 2005-04-23 11:31:46Z josef $
  *
  * Copyright (C) 2000 Brent Hendricks.
  *
@@ -36,7 +36,7 @@ typedef enum {
 } GGZLoginType;
 
 
-GGZPlayerHandlerStatus login_player(GGZLoginType type, GGZPlayer* player,
-                                   char *name, char *password);
+GGZPlayerHandlerStatus login_player(GGZLoginType type, GGZPlayer *player,
+				    char *name, const char *password, const char *email);
 
 #endif /* _GGZ_LOGIN_H */
