@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 9/22/01
  * Desc: Functions for handling network IO
- * $Id: net.c,v 1.105 2003/01/22 13:50:38 dr_maux Exp $
+ * $Id: net.c 7901 2006-03-13 12:35:44Z josef $
  * 
  * Code for parsing XML streamed from the server
  *
@@ -65,7 +65,7 @@ struct _GGZNetIO {
 	int fd;
 
 	/* Flag to indicate we're in a parse call */
-	char parsing;
+	bool parsing;
 
 	/* XML Parser */
 	XML_Parser parser;
@@ -81,7 +81,7 @@ struct _GGZNetIO {
 
 	/* Flag whether we've a tag on this round of parsing */
 	/* This element, coupled with the byte counter helps us prevent DoS */
-	char tag_seen;
+	bool tag_seen;
 };
 
 
@@ -95,8 +95,8 @@ typedef struct _GGZTableData {
 /* Seat data structure */
 typedef struct _GGZSeatData {
 	int index;
-	char *type; 
-	char *name;
+	const char *type; 
+	const char *name;
 } GGZSeatData;
 
 
@@ -104,14 +104,17 @@ typedef struct _GGZSeatData {
 typedef struct _GGZAuthData {
 	char *name;
 	char *password;
+	char *email;
 } GGZAuthData;
 
 
 /* Callbacks for XML parser */
-static void _net_parse_start_tag(void *, const char*, const char **);
+static void _net_parse_start_tag(void * data, const char *el,
+				 const char **attr);
 static void _net_parse_end_tag(void *data, const char *el);
 static void _net_parse_text(void *data, const char *text, int len);
-static GGZXMLElement* _net_new_element(char *tag, char **attrs);
+static GGZXMLElement* _net_new_element(const char *tag,
+				       const char * const * attrs);
 
 /* Handler functions for various tags */
 static void _net_handle_session(GGZNetIO *net, GGZXMLElement *session);
@@ -119,6 +122,7 @@ static void _net_handle_channel(GGZNetIO *net, GGZXMLElement *channel);
 static void _net_handle_login(GGZNetIO *net, GGZXMLElement *login);
 static void _net_handle_name(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_password(GGZNetIO *net, GGZXMLElement *element);
+static void _net_handle_email(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_update(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_list(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_enter(GGZNetIO *net, GGZXMLElement *element);
@@ -133,11 +137,12 @@ static void _net_handle_desc(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_motd(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_pong(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_ping(GGZNetIO *net, GGZXMLElement *element);
+static void _net_handle_info(GGZNetIO *net, GGZXMLElement *element);
 static void _net_handle_tls_start(GGZNetIO *net, GGZXMLElement *element);
 
 /* Utility functions */
 static int str_to_int(const char *str, int dflt);
-static int check_playerconn(GGZNetIO *net, const char *type);
+static bool check_playerconn(GGZNetIO *net, const char *type);
 static void _net_dump_data(struct _GGZNetIO *net, char *data, int size);
 static GGZReturn _net_send_result(GGZNetIO *net, const char *action,
 				  GGZClientReqError code);
@@ -147,7 +152,7 @@ static GGZReturn _net_send_login_anon_status(GGZNetIO *net,
 					     GGZClientReqError status);
 static GGZReturn _net_send_login_new_status(GGZNetIO *net,
 					    GGZClientReqError status,
-					    char *password);
+					    const char *password);
 static GGZReturn _net_send_table_status(GGZNetIO *net, GGZTable *table);
 static GGZReturn _net_send_table_seat(GGZNetIO *net, GGZTable *table,
 				      GGZTableSeat *seat);
@@ -158,6 +163,9 @@ static GGZReturn _net_send_seat(GGZNetIO *net, GGZTableSeat *seat);
 static GGZReturn _net_send_spectator(GGZNetIO *net,
 				     GGZTableSpectator *spectator);
 static GGZReturn _net_send_pong(GGZNetIO *net, const char *id);
+
+static GGZReturn _net_send_room_player_count(GGZNetIO *net,
+					     int index, int player_count);
 static GGZReturn _net_send_line(GGZNetIO *net, char *line, ...)
 				ggz__attribute((format(printf, 2, 3)));
 
@@ -165,8 +173,9 @@ static GGZAuthData* _net_authdata_new(void);
 static void _net_authdata_free(GGZAuthData *data);
 static void _net_auth_set_name(GGZXMLElement *login, char *name);
 static void _net_auth_set_password(GGZXMLElement *login, char *password);
+static void _net_auth_set_email(GGZXMLElement *login, char *email);
 static void _net_table_add_seat(GGZXMLElement*, GGZSeatData*);
-static void _net_table_set_desc(GGZXMLElement*, char*);
+static void _net_table_set_desc(GGZXMLElement*, const char*);
 static GGZTableData* _net_tabledata_new(void);
 static void _net_tabledata_free(GGZTableData*);
 static void* _net_seat_copy(void *data);
@@ -186,9 +195,9 @@ GGZNetIO* net_new(int fd, GGZClient *client)
 	net->fd = fd;
 	net->client = client;
 	net->dump_file = -1;
-	net->parsing = 0;
+	net->parsing = false;
 	net->byte_count = 0;
-	net->tag_seen = 0;
+	net->tag_seen = false;
 
         /* Init parser */
         if (!(net->parser = XML_ParserCreate(NULL)))
@@ -278,18 +287,22 @@ void net_free(GGZNetIO *net)
 
 /* net_send_XXX() functions for sending messages to the client */
 
-GGZReturn net_send_serverid(GGZNetIO *net, char *srv_name, int use_tls)
+GGZReturn net_send_serverid(GGZNetIO *net, const char *srv_name, bool use_tls)
 {
-	char *xml_srv_name;
+	char *xml_srv_name = ggz_xml_escape(srv_name);
 	GGZReturn status;
 
-	xml_srv_name = ggz_xml_escape(srv_name);
+	/* This could go into a separate function, net_send_header, along
+	   with the DTD. */
+	_net_send_line(net, "<?xml version='1.0' encoding='UTF-8'?>");
 
 	_net_send_line(net, "<SESSION>");
-	_net_send_line(net, "\t<SERVER ID='GGZ-%s' NAME='%s' VERSION='%d' STATUS='%s' TLS_SUPPORT='%s'>",
-		VERSION, xml_srv_name, GGZ_CS_PROTO_VERSION, "ok", ((use_tls && ggz_tls_support_query()) ? "yes" : "no"));
-	_net_send_line(net, "\t\t<OPTIONS CHATLEN='%d'/>", MAX_CHAT_LEN);
-	status = _net_send_line(net, "\t</SERVER>");
+	_net_send_line(net, "<SERVER ID='GGZ-%s' NAME='%s' VERSION='%d' "
+		       "STATUS='ok' TLS_SUPPORT='%s'>",
+		       VERSION, xml_srv_name, GGZ_CS_PROTO_VERSION,
+		       ((use_tls && ggz_tls_support_query()) ? "yes" : "no"));
+	_net_send_line(net, "<OPTIONS CHATLEN='%d'/>", MAX_CHAT_LEN);
+	status = _net_send_line(net, "</SERVER>");
 
 	ggz_free(xml_srv_name);
 
@@ -297,7 +310,7 @@ GGZReturn net_send_serverid(GGZNetIO *net, char *srv_name, int use_tls)
 }
  
 
-GGZReturn net_send_server_full(GGZNetIO *net, char *srv_name)
+GGZReturn net_send_server_full(GGZNetIO *net, const char *srv_name)
 {
 	char *xml_srv_name;
 	GGZReturn status;
@@ -305,7 +318,9 @@ GGZReturn net_send_server_full(GGZNetIO *net, char *srv_name)
 	xml_srv_name = ggz_xml_escape(srv_name);
 
 	_net_send_line(net, "<SESSION>");
-	status = _net_send_line(net, "\t<SERVER ID='GGZ-%s' NAME='%s' VERSION='%d' STATUS='%s'/>", VERSION, xml_srv_name, GGZ_CS_PROTO_VERSION, "full");
+	status = _net_send_line(net, "<SERVER ID='GGZ-%s' NAME='%s' "
+				"VERSION='%d' STATUS='full'/>",
+				VERSION, xml_srv_name, GGZ_CS_PROTO_VERSION);
 
 	ggz_free(xml_srv_name);
 
@@ -314,7 +329,7 @@ GGZReturn net_send_server_full(GGZNetIO *net, char *srv_name)
 
 
 GGZReturn net_send_login(GGZNetIO *net, GGZLoginType type,
-			 GGZClientReqError status, char *password)
+			 GGZClientReqError status, const char *password)
 {
 	switch (type) {
 	case GGZ_LOGIN:
@@ -337,7 +352,8 @@ GGZReturn net_send_motd(GGZNetIO *net)
 	int i, num;
 	char *line;
 	
-	_net_send_line(net, "<MOTD PRIORITY='normal'><![CDATA[");
+	_net_send_line(net, "<MOTD PRIORITY='normal' URL='%s'><![CDATA[",
+		(opt.motd_web ? opt.motd_web : ""));
 
 	num = motd_get_num_lines();
 		
@@ -372,10 +388,11 @@ GGZReturn net_send_room_list_count(GGZNetIO *net, int count)
 
 
 GGZReturn net_send_room(GGZNetIO *net, int index,
-			RoomStruct *room, char verbose)
+			RoomStruct *room, bool verbose)
 {
-	_net_send_line(net, "<ROOM ID='%d' NAME='%s' GAME='%d'>",
-		       index, room->name, room->game_type);
+	_net_send_line(net, "<ROOM ID='%d' NAME='%s' GAME='%d' PLAYERS='%d'>",
+		       index, room->name, room->game_type,
+		       room->player_count);
 	if (verbose && room->description)
 		_net_send_line(net, "<DESC>%s</DESC>", room->description);
 	return _net_send_line(net, "</ROOM>");
@@ -403,21 +420,32 @@ GGZReturn net_send_type_list_count(GGZNetIO *net, int count)
 }
 
 
-GGZReturn net_send_type(GGZNetIO *net, int index, GameInfo *type, char verbose)
+GGZReturn net_send_type(GGZNetIO *net, int index,
+			GameInfo *type, bool verbose)
 {
+	int i;
 	char *players = ggz_numberlist_write(&type->player_allow_list);
 	char *bots = ggz_numberlist_write(&type->bot_allow_list);
 	char *spectators = bool_to_str(type->allow_spectators);
+	char *peers = bool_to_str(type->allow_peers);
 
 	_net_send_line(net, "<GAME ID='%d' NAME='%s' VERSION='%s'>",
 		       index, type->name, type->version);
 	_net_send_line(net, "<PROTOCOL ENGINE='%s' VERSION='%s'/>",
 		       type->p_engine, type->p_version);
-	_net_send_line(net, "<ALLOW PLAYERS='%s' BOTS='%s' SPECTATORS='%s'/>",
-		       players, bots, spectators);
+	_net_send_line(net, "<ALLOW PLAYERS='%s' BOTS='%s' SPECTATORS='%s' PEERS='%s'/>",
+		       players, bots, spectators, peers);
 
 	ggz_free(players);
 	ggz_free(bots);
+
+	if (type->named_bots) {
+		for (i = 0; type->named_bots[i]; i++) {
+			_net_send_line(net, "<BOT NAME='%s' CLASS='%s'/>",
+				type->named_bots[i][0],
+				type->named_bots[i][1]);
+		}
+	}
 	
 	if (verbose) {
 		_net_send_line(net, "<ABOUT AUTHOR='%s' URL='%s'/>",
@@ -440,22 +468,19 @@ GGZReturn net_send_player_list_error(GGZNetIO *net, GGZClientReqError status)
 }
 
 
-GGZReturn net_send_player_list_count(GGZNetIO *net, int count)
+GGZReturn net_send_player_list_count(GGZNetIO *net, int room_id, int count)
 {
-	int room;
-
-	room = player_get_room(net->client->data);
-	
 	_net_send_line(net, "<RESULT ACTION='list' CODE='%s'>",
 		       ggz_error_to_string(E_OK));
-	return _net_send_line(net, "<LIST TYPE='player' ROOM='%d'>", room);
+	return _net_send_line(net, "<LIST TYPE='player' ROOM='%d'>",
+			      room_id);
 }
 
 
 static void _net_get_player_stats_string(GGZPlayer *player,
 					 char *buf, size_t bufsz)
 {
-	char record[256] = "", rating[64] = "";
+	char record[256] = "", rating[64] = "", highscore[128] = "";
 
 	/* The caller should ensure that these values are safe to access... */
 	/* FIXME: do this more elegantly */
@@ -471,61 +496,81 @@ static void _net_get_player_stats_string(GGZPlayer *player,
 		snprintf(rating, sizeof(rating),
 			 " RATING='%d'", player->rating);
 
-	snprintf(buf, bufsz, "%s%s", record, rating);
+#if 0
+	if (player->have_ranking)
+		snprintf(ranking, sizeof(ranking),
+			 " RANKING='%d", player->ranking);
+#endif
+
+	if (player->have_highscore) {
+		snprintf(highscore, sizeof(highscore),
+			 " HIGHSCORE='%d'", player->highscore);
+	}
+
+	snprintf(buf, bufsz, "%s%s%s", record, rating, highscore);
 }
 
 
 GGZReturn net_send_player(GGZNetIO *net, GGZPlayer *player)
 {
-	GGZPlayerType type;
-	char *type_desc = "**none**";
+	GGZPlayerType type = player_get_type(player);
+	const char *type_desc = ggz_playertype_to_string(type);
 	char stats[512];
-	
-	type = player_get_type(player);
-
-	switch (type) {
-	case GGZ_PLAYER_NORMAL:
-		type_desc = "normal";
-		break;
-	case GGZ_PLAYER_GUEST:
-		type_desc = "guest";
-		break;
-	case GGZ_PLAYER_ADMIN:
-		type_desc = "admin";
-		break;
-	case GGZ_PLAYER_NONE:
-		break;
-	}
+	char *player_name_quoted;
+	GGZReturn ret;
 
 	_net_get_player_stats_string(player, stats, sizeof(stats));
 
+	player_name_quoted = ggz_xml_escape(player->name);
+
 	/* The caller should ensure that these values are safe to access... */
-	return _net_send_line(net, 
+	ret = _net_send_line(net, 
 			      "<PLAYER ID='%s' TYPE='%s' TABLE='%d' "
 			      "LAG='%d'%s/>",
-			      player->name, type_desc, player->table,
+			      player_name_quoted, type_desc, player->table,
 			      player->lag_class, stats);
+
+	ggz_free(player_name_quoted);
+
+	return ret;
 }
 
 
 static GGZReturn _net_send_player_lag(GGZNetIO *net, GGZPlayer *player)
 {
+	char *player_name_quoted;
+	GGZReturn ret;
+
+	player_name_quoted = ggz_xml_escape(player->name);
+
 	/* The caller should ensure that these values are safe to access... */
-	return _net_send_line(net, 
+	ret = _net_send_line(net, 
 			      "<PLAYER ID='%s' LAG='%d'/>",
-			      player->name, player->lag_class);
+			      player_name_quoted, player->lag_class);
+
+	ggz_free(player_name_quoted);
+
+	return ret;
 }
 
 
 static GGZReturn _net_send_player_stats(GGZNetIO *net, GGZPlayer *player)
 {
 	char stats[512];
+	char *player_name_quoted;
+	GGZReturn ret;
 
 	_net_get_player_stats_string(player, stats, sizeof(stats));
 
-	return _net_send_line(net,
+	player_name_quoted = ggz_xml_escape(player->name);
+
+	ret = _net_send_line(net,
 			      "<PLAYER ID='%s'%s/>",
-			      player->name, stats);
+			      player_name_quoted, stats);
+
+	ggz_free(player_name_quoted);
+
+	return ret;
 }
 
 
@@ -542,37 +587,38 @@ GGZReturn net_send_table_list_error(GGZNetIO *net, GGZClientReqError status)
 }
 
 
-GGZReturn net_send_table_list_count(GGZNetIO *net, int count)
+GGZReturn net_send_table_list_count(GGZNetIO *net, int room_id, int count)
 {
-	int room;
-
-	room = player_get_room(net->client->data);
-
 	_net_send_line(net, "<RESULT ACTION='list' CODE='%s'>",
 		       ggz_error_to_string(E_OK));
-	return _net_send_line(net, "<LIST TYPE='table' ROOM='%d'>", room);
+	return _net_send_line(net, "<LIST TYPE='table' ROOM='%d'>", room_id);
 }
 
 
 GGZReturn net_send_table(GGZNetIO *net, GGZTable *table)
 {
 	int i;
+	char *description_quoted;
 
 	_net_send_line(net, "<TABLE ID='%d' GAME='%d' STATUS='%d' SEATS='%d'>",
 		       table->index, table->type, table->state, 
 		       seats_num(table));
 
-	_net_send_line(net, "<DESC>%s</DESC>", table->desc);
+	description_quoted = ggz_xml_escape(table->desc);
+
+	_net_send_line(net, "<DESC>%s</DESC>", description_quoted);
+
+	ggz_free(description_quoted);
 	
 	for (i = 0; i < seats_num(table); i++) {
-		GGZTableSeat seat = {index: i,
-				     type:  table->seat_types[i]};
+		GGZTableSeat seat = {.index = i,
+				     .type =  table->seat_types[i]};
 		strncpy(seat.name, table->seat_names[i], MAX_USER_NAME_LEN+1);
 		_net_send_seat(net, &seat);
 	}
 
 	for (i = 0; i < spectator_seats_num(table); i++) {
-		GGZTableSpectator seat = {index: i};
+		GGZTableSpectator seat = {.index = i};
 		if (table->spectators[i][0] == '\0') continue;
 		strcpy(seat.name, table->spectators[i]);
 		_net_send_spectator(net, &seat);
@@ -599,6 +645,8 @@ GGZReturn net_send_chat(GGZNetIO *net, GGZChatType type,
 			const char *sender, const char *msg)
 {
 	const char *type_str = ggz_chattype_to_string(type);
+	char *sender_quoted;
+	GGZReturn ret;
 
 	if (type == GGZ_CHAT_BEEP) {
 		/* A beep chat can't have a message. */
@@ -614,13 +662,19 @@ GGZReturn net_send_chat(GGZNetIO *net, GGZChatType type,
 		}
 	}
 
+	sender_quoted = ggz_xml_escape(sender);
+
 	if (msg) {
-		return _net_send_line(net, "<CHAT TYPE='%s' FROM='%s'>"
+		ret = _net_send_line(net, "<CHAT TYPE='%s' FROM='%s'>"
 				      "<![CDATA[%s]]></CHAT>", 
-				      type_str, sender, msg);
+				      type_str, sender_quoted, msg);
 	} else 
-		return _net_send_line(net, "<CHAT TYPE='%s' FROM='%s'/>",
-				      type_str, sender);
+		ret = _net_send_line(net, "<CHAT TYPE='%s' FROM='%s'/>",
+				      type_str, sender_quoted);
+
+	ggz_free(sender_quoted);
+
+	return ret;
 }
 
 
@@ -637,7 +691,7 @@ GGZReturn net_send_table_launch(GGZNetIO *net, GGZClientReqError status)
 
 
 GGZReturn net_send_table_join(GGZNetIO *net,
-			      int is_spectator,
+			      bool is_spectator,
 			      const unsigned int table_index)
 {
 	return _net_send_line(net,
@@ -680,19 +734,22 @@ GGZReturn net_send_reseat_result(GGZNetIO *net, GGZClientReqError status)
 }
 
 
-GGZReturn net_send_player_update(GGZNetIO *net,
-				 GGZPlayerUpdateType opcode, const char *name)
+GGZReturn net_send_player_update(GGZNetIO *net, GGZPlayerUpdateType opcode,
+				 const char *name,
+				 int room_id, int other_room_id)
 {
 	GGZPlayer *player;
 	GGZPlayer p2;
-	int room;
-
-	room = player_get_room(net->client->data);
+	char *name_quoted;
 	
 	switch (opcode) {
 	case GGZ_PLAYER_UPDATE_DELETE:
-		_net_send_line(net, "<UPDATE TYPE='player' ACTION='delete' ROOM='%d'>", room);
-		_net_send_line(net, "<PLAYER ID='%s'/>", name);
+		_net_send_line(net, "<UPDATE TYPE='player' ACTION='delete' "
+			       "ROOM='%d' TOROOM='%d'>",
+			       room_id, other_room_id);
+		name_quoted = ggz_xml_escape(name);
+		_net_send_line(net, "<PLAYER ID='%s'/>", name_quoted);
+		ggz_free(name_quoted);
 		return _net_send_line(net, "</UPDATE>");
 
 	case GGZ_PLAYER_UPDATE_ADD:
@@ -706,7 +763,9 @@ GGZReturn net_send_player_update(GGZNetIO *net,
 		p2 = *player;
 		pthread_rwlock_unlock(&player->stats_lock);
 		pthread_rwlock_unlock(&player->lock);
-		_net_send_line(net, "<UPDATE TYPE='player' ACTION='add' ROOM='%d'>", room);
+		_net_send_line(net, "<UPDATE TYPE='player' ACTION='add' "
+			       "ROOM='%d' FROMROOM='%d'>",
+			       room_id, other_room_id);
 		net_send_player(net, &p2);
 		return _net_send_line(net, "</UPDATE>");
 
@@ -719,7 +778,8 @@ GGZReturn net_send_player_update(GGZNetIO *net,
 		}
 		p2 = *player;
 		pthread_rwlock_unlock(&player->lock);
-		_net_send_line(net, "<UPDATE TYPE='player' ACTION='lag' ROOM='%d'>", room);
+		_net_send_line(net, "<UPDATE TYPE='player' ACTION='lag' "
+			       "ROOM='%d'>", room_id);
 		_net_send_player_lag(net, &p2);
 		return _net_send_line(net, "</UPDATE>");
 	case GGZ_PLAYER_UPDATE_STATS:
@@ -733,7 +793,8 @@ GGZReturn net_send_player_update(GGZNetIO *net,
 		p2 = *player;
 		pthread_rwlock_unlock(&player->stats_lock);
 		pthread_rwlock_unlock(&player->lock);
-		_net_send_line(net, "<UPDATE TYPE='player' ACTION='stats' ROOM='%d'>", room);
+		_net_send_line(net, "<UPDATE TYPE='player' ACTION='stats' "
+			       "ROOM='%d'>", room_id);
 		_net_send_player_stats(net, player);
 		return _net_send_line(net, "</UPDATE>");
 	}
@@ -744,12 +805,10 @@ GGZReturn net_send_player_update(GGZNetIO *net,
 
 
 GGZReturn net_send_table_update(GGZNetIO *net, GGZTableUpdateType opcode,
-				GGZTable *table, void* seat_data)
+				GGZTable *table, void* seat_data,
+				int room_id)
 {
 	char *action = NULL;
-	int room;
-
-	room = player_get_room(net->client->data);
 
 	switch (opcode) {
 	case GGZ_TABLE_UPDATE_DELETE:
@@ -782,7 +841,7 @@ GGZReturn net_send_table_update(GGZNetIO *net, GGZTableUpdateType opcode,
 
 	/* Always send opcode */
 	_net_send_line(net, "<UPDATE TYPE='table' ACTION='%s' ROOM='%d'>",
-		       action, room);
+		       action, room_id);
 
 	switch (opcode) {
 	case GGZ_TABLE_UPDATE_DELETE:
@@ -817,6 +876,37 @@ GGZReturn net_send_table_update(GGZNetIO *net, GGZTableUpdateType opcode,
 }
 
 
+GGZReturn _net_send_room_player_count(GGZNetIO *net,
+				      int index, int player_count)
+{
+	return _net_send_line(net, "<ROOM ID='%d' PLAYERS='%d'/>",
+			      index, player_count);
+}
+
+
+GGZReturn net_send_room_update(GGZNetIO *net, GGZRoomUpdateType opcode,
+			       int index, int player_count)
+{
+	char *action = NULL;
+
+	switch (opcode) {
+	case GGZ_ROOM_UPDATE_PLAYER_COUNT:
+		action = "players";
+		break;
+	}
+
+	_net_send_line(net, "<UPDATE TYPE='room' ACTION='%s'>", action);
+
+	switch (opcode) {
+	case GGZ_ROOM_UPDATE_PLAYER_COUNT:
+		_net_send_room_player_count(net, index, player_count);
+		break;
+	}
+
+	return _net_send_line(net, "</UPDATE>");
+}
+
+
 GGZReturn net_send_update_result(GGZNetIO *net, GGZClientReqError status)
 {
 	return _net_send_result(net, "update", status);
@@ -837,26 +927,28 @@ GGZReturn net_send_ping(GGZNetIO *net)
 
 /* Check for incoming data */
 #if 0
-static int net_data_is_pending(GGZNetIO *net)
+static bool net_data_is_pending(GGZNetIO *net)
 {
-	int pending = 0;
+	int result;
 	struct pollfd fd[1] = {{net->fd, POLLIN, 0}};
 
 	if (net && net->fd != -1) {
-	
-	dbg_msg(GGZ_DBG_CONNECTION, "Checking for net events");	
-	if ( (pending = poll(fd, 1, 0)) < 0) {
-		if (errno == EINTR)
-			/* Ignore interruptions */
-			pending = 0;
-		else 
-			err_sys_exit("poll failed in ggzcore_server_data_is_pending");
-	}
-	else if (pending)
-		dbg_msg(GGZ_DBG_CONNECTION, "Found a net event!");
+		dbg_msg(GGZ_DBG_CONNECTION, "Checking for net events");
+		result = poll(fd, 1, 0);
+		if (result < 0) {
+			if (errno == EINTR) {
+				/* Ignore interruptions */
+				return false;
+			} else {
+				err_sys_exit("poll failed in ggzcore_server_data_is_pending");
+			}
+
+		} else if (result > 0) {
+			return true;
+		}
 	}
 
-	return pending;
+	return false;
 }
 #endif
 
@@ -874,7 +966,7 @@ GGZPlayerHandlerStatus net_read_data(GGZNetIO *net)
 		return GGZ_REQ_OK;
 
 	/* Set flag in case we get called recursively */
-	net->parsing = 1;
+	net->parsing = true;
 
 	/* Get a buffer to hold the data */
 	if (!(buf = XML_GetBuffer(net->parser, BUFFSIZE)))
@@ -882,11 +974,11 @@ GGZPlayerHandlerStatus net_read_data(GGZNetIO *net)
 
 	/* Read in data from socket */
 	if ( (len = ggz_tls_read(net->fd, buf, BUFFSIZE)) < 0) {
-		
+
 		/* If it's a non-blocking socket and there isn't data,
                    we get EAGAIN.  It's safe to just return */
 		if (errno == EAGAIN) {
-			net->parsing = 0;
+			net->parsing = false;
 			return GGZ_REQ_OK;
 		}
 
@@ -895,7 +987,7 @@ GGZPlayerHandlerStatus net_read_data(GGZNetIO *net)
 	}
 
 	_net_dump_data(net, buf, len);
-	
+
 	/* If len == 0 then we've reached EOF */
 	done = (len == 0);
 
@@ -913,47 +1005,47 @@ GGZPlayerHandlerStatus net_read_data(GGZNetIO *net)
 		_net_send_result(net, "protocol", E_BAD_XML);
 		done = 1;
 	}
-	
+
 	/* If we saw any tags clear the byte counter, otherwise increment it*/
 	if (net->tag_seen) {
 		net->byte_count = 0;
-		net->tag_seen = 0;
+		net->tag_seen = false;
 	}
 	else {
 		net->byte_count += len;
 		/* If we haven't seen a tag in a while, it's an error */
 		if (net->byte_count > MAX_CHAT_LEN) {
 			dbg_msg(GGZ_DBG_XML, "Error: player overflowed XML buffer");
-			
+
 			_net_send_result(net, "protocol", E_BAD_OPTIONS);
 			done = 1;
 		}
 	}
-	   
 
 	/* Clear the flag now that we've completed this round of parsing */
-	net->parsing = 0;
+	net->parsing = false;
 
-	return (done ? GGZ_REQ_DISCONNECT: GGZ_REQ_OK);
+	return (done ? GGZ_REQ_DISCONNECT : GGZ_REQ_OK);
 }
 
 
 /********** Callbacks for XML parser **********/
-static void _net_parse_start_tag(void *data, const char *el, const char **attr)
+static void _net_parse_start_tag(void *data, const char *el,
+				 const char **attr)
 {
 	GGZNetIO *net = (GGZNetIO*)data;
 	GGZXMLElement *element;
 
 	dbg_msg(GGZ_DBG_XML, "New %s element", el);
-	
+
 	/* Create new element object */
-	element = _net_new_element((char*)el, (char**)attr);
+	element = _net_new_element(el, attr);
 
 	/* Put element on stack so we can process its children */
 	ggz_stack_push(net->stack, element);
 
 	/* Mark that we've seen a tag */
-	net->tag_seen = 1;
+	net->tag_seen = true;
 }
 
 
@@ -961,7 +1053,7 @@ static void _net_parse_end_tag(void *data, const char *el)
 {
 	GGZNetIO *net = (GGZNetIO*)data;
 	GGZXMLElement *element;
-	
+
 	/* Pop element off stack */
 	element = ggz_stack_pop(net->stack);
 
@@ -974,7 +1066,7 @@ static void _net_parse_end_tag(void *data, const char *el)
 	ggz_xmlelement_free(element);
 
 	/* Mark that we've seen a tag */
-	net->tag_seen = 1;
+	net->tag_seen = true;
 }
 
 
@@ -994,7 +1086,8 @@ static void _net_dump_data(GGZNetIO *net, char *data, int size)
 		write(net->dump_file, data, size);
 }
 
-static GGZXMLElement* _net_new_element(char *tag, char **attrs)
+static GGZXMLElement* _net_new_element(const char *tag,
+				       const char * const *attrs)
 {
 	void (*process_func)();
 
@@ -1008,6 +1101,8 @@ static GGZXMLElement* _net_new_element(char *tag, char **attrs)
 		process_func = _net_handle_name;
 	else if (strcasecmp(tag, "PASSWORD") == 0)
 		process_func = _net_handle_password;
+	else if (strcasecmp(tag, "EMAIL") == 0)
+		process_func = _net_handle_email;
 	else if (strcasecmp(tag, "UPDATE") == 0)
 		process_func = _net_handle_update;
 	else if (strcasecmp(tag, "LIST") == 0)
@@ -1016,6 +1111,8 @@ static GGZXMLElement* _net_new_element(char *tag, char **attrs)
 		process_func = _net_handle_enter;
 	else if (strcasecmp(tag, "CHAT") == 0)
 		process_func = _net_handle_chat;
+	else if (strcasecmp(tag, "INFO") == 0)
+		process_func = _net_handle_info;
 	else if (strcasecmp(tag, "JOIN") == 0)
 		process_func = _net_handle_join;
 	else if (strcasecmp(tag, "LEAVE") == 0)
@@ -1040,7 +1137,7 @@ static GGZXMLElement* _net_new_element(char *tag, char **attrs)
 		process_func = _net_handle_tls_start;
 	else
 		process_func = NULL;
-	
+
 	return ggz_xmlelement_new(tag, attrs, process_func, NULL);
 }
 
@@ -1056,7 +1153,7 @@ static void _net_handle_session(GGZNetIO *net, GGZXMLElement *element)
 /* Functions for <CHANNEL> tag */
 static void _net_handle_channel(GGZNetIO *net, GGZXMLElement *element)
 {
-	char *id;
+	const char *id;
 
 	if (!element) return;
 
@@ -1083,7 +1180,7 @@ static void _net_handle_login(GGZNetIO *net, GGZXMLElement *element)
 {
 	GGZLoginType login_type;
 	GGZAuthData *auth;	
-	char *type;
+	const char *type;
 
 	if (!element) return;
 
@@ -1125,7 +1222,7 @@ static void _net_handle_login(GGZNetIO *net, GGZXMLElement *element)
 
 	client_set_type(net->client, GGZ_CLIENT_PLAYER);
 	login_player(login_type, net->client->data,
-		     auth->name, auth->password);
+		     auth->name, auth->password, auth->email);
 
 	/* Free up any resources we allocated */
 	_net_authdata_free(auth);
@@ -1135,17 +1232,20 @@ static void _net_handle_login(GGZNetIO *net, GGZXMLElement *element)
 static GGZAuthData* _net_authdata_new(void)
 {
 	GGZAuthData *data = ggz_malloc(sizeof(GGZAuthData));
-	
+
 	data->name = NULL;
 	data->password = NULL;
-	
+	data->email = NULL;
+
 	return data;
 }
 
 
 static void _net_auth_set_name(GGZXMLElement *auth, char *name)
 {
-	GGZAuthData *data = ggz_xmlelement_get_data(auth);
+	GGZAuthData *data;
+
+	data = ggz_xmlelement_get_data(auth);
 
 	/* If data doesn't already exist, create it */
 	if (!data) {
@@ -1161,7 +1261,7 @@ static void _net_auth_set_name(GGZXMLElement *auth, char *name)
 static void _net_auth_set_password(GGZXMLElement *auth, char *password)
 {
 	GGZAuthData *data;
-	
+
 	data = ggz_xmlelement_get_data(auth);
 
 	/* If data doesn't already exist, create it */
@@ -1175,6 +1275,23 @@ static void _net_auth_set_password(GGZXMLElement *auth, char *password)
 }
 
 
+static void _net_auth_set_email(GGZXMLElement *auth, char *email)
+{
+	GGZAuthData *data;
+
+	data = ggz_xmlelement_get_data(auth);
+
+	/* If data doesn't already exist, create it */
+	if (!data) {
+		data = _net_authdata_new();
+		ggz_xmlelement_set_data(auth, data);
+	}
+
+	if (data->email) ggz_free(data->email); /* duplicate */
+	data->email = email;
+}
+
+
 static void _net_authdata_free(GGZAuthData *data)
 {
 	if (!data) return;
@@ -1191,7 +1308,7 @@ static void _net_authdata_free(GGZAuthData *data)
 static void _net_handle_name(GGZNetIO *net, GGZXMLElement *element)
 {
 	char *name;
-	char *parent_tag;
+	const char *parent_tag;
 	GGZXMLElement *parent;
 
 	if (!element) return;
@@ -1219,7 +1336,7 @@ static void _net_handle_name(GGZNetIO *net, GGZXMLElement *element)
 static void _net_handle_password(GGZNetIO *net, GGZXMLElement *element)
 {
 	char *password;
-	char *parent_tag;
+	const char *parent_tag;
 	GGZXMLElement *parent;
 
 	if (!element) return;
@@ -1243,10 +1360,38 @@ static void _net_handle_password(GGZNetIO *net, GGZXMLElement *element)
 }
 
 
+/* Functions for <EMAIL> tag */
+static void _net_handle_email(GGZNetIO *net, GGZXMLElement *element)
+{
+	char *email;
+	const char *parent_tag;
+	GGZXMLElement *parent;
+
+	if (!element) return;
+
+	/* Get parent off top of stack */
+	parent = ggz_stack_top(net->stack);
+	if (!parent) {
+		/* If there are no elements above us, it's a protocol error */
+		_net_send_result(net, "protocol", E_BAD_OPTIONS);
+		return;
+	}
+
+	parent_tag = ggz_xmlelement_get_tag(parent);
+	if (strcasecmp(parent_tag, "LOGIN") != 0) {
+		_net_send_result(net, "protocol", E_BAD_OPTIONS);
+		return;
+	}
+
+	email = ggz_strdup(ggz_xmlelement_get_text(element));
+	_net_auth_set_email(parent, email);
+}
+
+
 /* Functions for <UPDATE> tag */
 static void _net_handle_update(GGZNetIO *net, GGZXMLElement *update)
 {
-	char *type, *action, *room;
+	const char *type, *action, *room;
 
 	if (!update) return;
 	if (!check_playerconn(net, "update")) return;
@@ -1337,9 +1482,9 @@ static void _net_handle_update(GGZNetIO *net, GGZXMLElement *update)
 /* Functions for <LIST> tag */
 static void _net_handle_list(GGZNetIO *net, GGZXMLElement *element)
 {
-	char *type;
+	const char *type;
 	int verbose;
-	
+
 	if (!element) return;
 	if (!check_playerconn(net, "list")) return;
 
@@ -1382,9 +1527,9 @@ static void _net_handle_enter(GGZNetIO *net, GGZXMLElement *enter)
 /* Functions for <CHAT> tag */
 static void _net_handle_chat(GGZNetIO *net, GGZXMLElement *element)
 {
-	char *type_str, *to, *msg;
+	const char *type_str, *to, *msg;
         GGZChatType type;
-		
+
 	if (!element) return;
 	if (!check_playerconn(net, "chat")) return;
 
@@ -1403,6 +1548,19 @@ static void _net_handle_chat(GGZNetIO *net, GGZXMLElement *element)
 }
 
 
+/* Functions for <INFO> tag */
+static void _net_handle_info(GGZNetIO *net, GGZXMLElement *element)
+{
+	int seat_num;
+
+	if (!element) return;
+
+	seat_num = str_to_int(ggz_xmlelement_get_attr(element, "SEAT"), -1);
+
+	player_table_info(net->client->data, seat_num);
+}
+
+
 /* Functions for <JOIN> tag */
 static void _net_handle_join(GGZNetIO *net, GGZXMLElement *element)
 {
@@ -1442,7 +1600,7 @@ static void _net_handle_leave(GGZNetIO *net, GGZXMLElement *element)
 /* Functions for <RESEAT> tag */
 static void _net_handle_reseat(GGZNetIO *net, GGZXMLElement *element)
 {
-	char *action;
+	const char *action;
 	int seat_num;
 	GGZReseatType type;
 
@@ -1502,7 +1660,7 @@ static void _net_handle_table(GGZNetIO *net, GGZXMLElement *element)
 	GGZTable *table, *old_data;
 	GGZTableData *data;
 	GGZSeatData *seat;
-	char *desc = NULL, *parent_tag;
+	const char *desc = NULL, *parent_tag;
 	GGZList *seats = NULL;
 	GGZListEntry *entry;
 	GGZXMLElement *parent;
@@ -1538,7 +1696,6 @@ static void _net_handle_table(GGZNetIO *net, GGZXMLElement *element)
 	table->type = str_to_int(ggz_xmlelement_get_attr(element, "GAME"), -1);
 	num_seats = str_to_int(ggz_xmlelement_get_attr(element, "SEATS"), 0);
 
-#ifdef UNLIMITED_SEATS
 	if (num_seats > 0) {
 		int i;
 		/* FIXME: what if num_seats is really large? */
@@ -1552,7 +1709,6 @@ static void _net_handle_table(GGZNetIO *net, GGZXMLElement *element)
 			table->seat_names[i][0] = '\0';
 		}
 	}
-#endif
 
 	/* If room was specified, use it, otherwise use players current room */
 	table->room = player_get_room(net->client->data);
@@ -1570,13 +1726,7 @@ static void _net_handle_table(GGZNetIO *net, GGZXMLElement *element)
 		seat = ggz_list_get_data(entry);
 		seat_type = ggz_string_to_seattype(seat->type);
 
-		if (seat->index < 0 ||
-#ifdef UNLIMITED_SEATS
-		    seat->index >= table->num_seats
-#else
-		    seat->index >= MAX_TABLE_SIZE
-#endif
-		    ) {
+		if (seat->index < 0 || seat->index >= table->num_seats) {
 			/* We treat this as an error -
 			   we could just skip the seat instead */
 			if (data) _net_tabledata_free(data);
@@ -1586,9 +1736,14 @@ static void _net_handle_table(GGZNetIO *net, GGZXMLElement *element)
 
 		switch (seat_type) {
 		case GGZ_SEAT_OPEN:
-		case GGZ_SEAT_BOT:
 		case GGZ_SEAT_NONE:
 			break;
+		case GGZ_SEAT_BOT:
+			if (seat->name) {
+				strcpy(table->seat_names[seat->index],
+				       seat->name);
+			}
+			break;
 		case GGZ_SEAT_RESERVED:
 			/* We verify that this is a real,
 			   registered player name. */
@@ -1616,6 +1771,10 @@ static void _net_handle_table(GGZNetIO *net, GGZXMLElement *element)
 				strcpy(table->seat_names[seat->index],
 				       seat->name);
 			break;
+		case GGZ_SEAT_ABANDONED:
+			/* Clients can't do this. */
+			seat_type = GGZ_SEAT_NONE;
+			break;
 		}
 		table->seat_types[seat->index] = seat_type;
 	}
@@ -1643,7 +1802,7 @@ static void _net_table_add_seat(GGZXMLElement *table, GGZSeatData *seat)
 }
 
 
-static void _net_table_set_desc(GGZXMLElement *table, char *desc)
+static void _net_table_set_desc(GGZXMLElement *table, const char *desc)
 {
 	struct _GGZTableData *data = ggz_xmlelement_get_data(table);
 
@@ -1691,7 +1850,7 @@ static void _net_handle_seat(GGZNetIO *net, GGZXMLElement *element)
 {
 	struct _GGZSeatData seat;
 	GGZXMLElement *parent;
-	char *parent_tag;
+	const char *parent_tag;
 
 	/* Get parent off top of stack */
 	parent = ggz_stack_top(net->stack);
@@ -1748,7 +1907,7 @@ static void _net_seat_free(GGZSeatData *seat)
 static void _net_handle_desc(GGZNetIO *net, GGZXMLElement *element)
 {
 	char *desc = NULL;
-	char *parent_tag;
+	const char *parent_tag;
 	GGZXMLElement *parent;
 
 	/* Get parent off top of stack */
@@ -1814,19 +1973,18 @@ static int str_to_int(const char *str, int dflt)
 }
 
 
-static int check_playerconn(GGZNetIO *net, const char *type)
+static bool check_playerconn(GGZNetIO *net, const char *type)
 {
 	if (net->client->data)
-		return 1;
+		return true;
 
 	/* This should only be caused by a malicious client. */
 	dbg_msg(GGZ_DBG_CONNECTION,
 		"Client requested %s before login.", type);
 
-	/* FIXME: maybe we should just disconnect them. */
 	_net_send_result(net, type, E_NOT_LOGGED_IN);
 
-	return 0;
+	return false;
 }
 
 
@@ -1852,8 +2010,12 @@ static GGZReturn _net_send_table_spectator(GGZNetIO *net, GGZTable *table,
 
 static GGZReturn _net_send_table_desc(GGZNetIO *net, GGZTable *table)
 {
+	char *description_quoted;
+
 	_net_send_line(net, "<TABLE ID='%d'>", table->index);
-	_net_send_line(net, "<DESC>%s</DESC>", table->desc);
+	description_quoted = ggz_xml_escape(table->desc);
+	_net_send_line(net, "<DESC>%s</DESC>", description_quoted);
+	ggz_free(description_quoted);
 	return _net_send_line(net, "</TABLE>");
 }
 
@@ -1873,6 +2035,7 @@ static GGZReturn _net_send_seat(GGZNetIO *net, GGZTableSeat *seat)
 	switch (seat->type) {
 	case GGZ_SEAT_RESERVED:
 	case GGZ_SEAT_PLAYER:
+	case GGZ_SEAT_ABANDONED:
 		name = seat->name;
 		break;
 	case GGZ_SEAT_OPEN:
@@ -1881,6 +2044,9 @@ static GGZReturn _net_send_seat(GGZNetIO *net, GGZTableSeat *seat)
 		name = NULL;
 		break;
 	}
+
+	if (seat->type == GGZ_SEAT_BOT && seat->name[0] != '\0')
+		name = seat->name;
 	
 	if (name)
 		return _net_send_line(net,
@@ -1923,7 +2089,7 @@ static GGZReturn _net_send_login_anon_status(GGZNetIO *net,
 
 static GGZReturn _net_send_login_new_status(GGZNetIO *net,
 					    GGZClientReqError status,
-					    char *password)
+					    const char *password)
 {
 	/* Try to send login status */
 	_net_send_line(net, "<RESULT ACTION='login' CODE='%s'>",
@@ -1947,6 +2113,25 @@ static GGZReturn _net_send_pong(GGZNetIO *net, const char *id)
 }
 
 
+/* Send <INFO> tag. */
+GGZReturn net_send_info_list_begin(GGZNetIO *net)
+{
+	return _net_send_line(net, "<INFO>");
+}
+
+GGZReturn net_send_info(GGZNetIO *net, int num, const char *realname,
+			const char *photo, const char *host)
+{
+	return _net_send_line(net, "<PLAYERINFO SEAT='%d' REALNAME='%s' PHOTO='%s' HOST='%s'/>",
+		num, (realname ? realname : ""), (photo ? photo : ""), (host ? host : ""));
+}
+
+GGZReturn net_send_info_list_end(GGZNetIO *net)
+{
+	return _net_send_line(net, "</INFO>");
+}
+
+
 static GGZReturn _net_send_line(GGZNetIO *net, char *line, ...)
 {
 	char buf[4096];
