@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 10/15/99
  * Desc: Parse command-line arguments and conf file
- * $Id: parse_opt.c,v 1.104 2003/01/22 13:50:38 dr_maux Exp $
+ * $Id: parse_opt.c 7861 2006-02-13 07:02:50Z josef $
  *
  * Copyright (C) 1999-2002 Brent Hendricks.
  *
@@ -29,7 +29,7 @@
 
 #include <ctype.h>
 #include <dirent.h>
-#include <popt.h>
+#include <getopt.h>
 #include <pthread.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -49,12 +49,6 @@
 #include "util.h"
 
 
-/* Stuff from control.c we need access to */
-extern Options opt;
-extern struct GGZState state;
-extern struct GameInfo game_types[MAX_GAME_TYPES];
-
-
 /* Private file parsing functions */
 static void get_config_options(int);
 static void parse_game(char *, char *);
@@ -72,12 +66,12 @@ struct LogTypes {
 	unsigned type;
 };
 static const struct LogTypes log_types[] = {
-	{ "all",		GGZ_LOG_ALL },
+	{ "all",			GGZ_LOG_ALL },
 	{ "notices",		GGZ_LOG_NOTICE },
 	{ "connections",	GGZ_LOG_CONNECTION_INFO },
 	{ "security",		GGZ_LOG_SECURITY },
-	{ "update",		GGZ_LOG_UPDATE },
-	{ "tables",		GGZ_LOG_TABLES }
+	{ "update",			GGZ_LOG_UPDATE },
+	{ "tables",			GGZ_LOG_TABLES }
 };
 static int num_log_types = sizeof(log_types) / sizeof(log_types[0]);
 
@@ -112,53 +106,131 @@ static char **r_list = NULL;
 /* Defaults for Admin items */
 #define ADMIN_ERR     "<You must set this parameter in ggzd.conf>"
 
-static const struct poptOption args[] = {
-	
-	{"foreground", 'F', POPT_ARG_NONE, &opt.foreground, 0,
-	 "Tells ggzd to run in the foreground"},
-	{"file", 'f', POPT_ARG_STRING, &opt.local_conf, 0, 
-	 "Configuration file", "FILE"},
-	{"log", 'l', POPT_ARG_INT, &log_info.log_types, 0,
-	 "Types of logging to perform", "LEVEL"},
-	{"port", 'p', POPT_ARG_INT, &opt.main_port, 0,
-	 "GGZ port number", "PORT"},
-	{"version", 'V', POPT_ARG_NONE, NULL, 1},
-	POPT_AUTOHELP {NULL, '\0', 0, NULL, 0}	/* end the list */
+static struct option options[] = {
+	{"foreground", no_argument, 0, 'F'},
+	{"file", required_argument, 0, 'f'},
+	{"log", required_argument, 0, 'l'},
+	{"port", required_argument, 0, 'p'},
+	{"help", no_argument, 0, 'h'},
+	{"version", no_argument, 0, 'v'},
+	{"specs", no_argument, 0, 's'},
+	{0, 0, 0, 0}
 };
 
-/* Parse command-line options */
-void parse_args(int argc, const char *argv[])
+/* Convert a string into a list */
+/* FIXME: should be made available from ggz_read_list() */
+static char **_ggz_string_to_list(const char *s, const char *sep)
 {
+	char *sc = ggz_strdup(s);
+	char **list = NULL;
+	int count = 0;
+	char *tok;
 
-	poptContext context = poptGetContext(NULL, argc, argv, args, 0);
-	int rc;
+	tok = strtok(sc, sep);
+	while(tok)
+	{
+		count++;
+		list = (char**)ggz_realloc(list, (count + 1) * sizeof(char**));
+		list[count - 1] = ggz_strdup(tok);
+		list[count] = NULL;
+		tok = strtok(NULL, sep);
+	}
+	ggz_free(sc);
+
+	return list;
+}
 
-	while ((rc = poptGetNextOpt(context)) != -1) {
-		switch (rc) {
-		case 1:	/* Version command */
-			printf("GGZ Gaming Zone server: version %s\n", 
-			       VERSION);
-			poptFreeContext(context);
-			exit(0);
 #ifdef DEBUG
-		case 2: /* They specified --debug */
-			log_info.popt_dbg++;
-			break;
+#define SPEC_DEBUG 1
+#else
+#define SPEC_DEBUG 0
 #endif
-		case POPT_ERROR_NOARG:
-		case POPT_ERROR_BADOPT:
-		case POPT_ERROR_BADNUMBER:
-		case POPT_ERROR_OVERFLOW:
-			err_msg("%s: %s", poptBadOption(context, 0),
-				poptStrerror(rc));
-			poptFreeContext(context);
-			exit(-1);
-			break;
+#define SPEC_DB DATABASE_TYPE
+#ifdef WITH_HOWL
+#define SPEC_HOWL 1
+#else
+#define SPEC_HOWL 0
+#endif
+
+static void dump_specs(void)
+{
+	int tls = ggz_tls_support_query();
+
+	parse_conf_file();
+
+	printf("GGZ Gaming Zone server (ggzd) specifications\n");
+	printf("Version: %s\n", VERSION);
+	printf("Protocol version: %i\n", GGZ_CS_PROTO_VERSION);
+	printf("Configuration: %s/ggzd.conf [%s]\n",
+		GGZDCONFDIR,
+		(opt.conf_valid ? "ok" : "not ok"));
+	printf("Debugging: %s\n", (SPEC_DEBUG ? "yes" : "no"));
+	printf("Database backend: %s\n", SPEC_DB);
+	printf("Zeroconf support: %s [%s]\n",
+		(SPEC_HOWL ? "yes" : "no"),
+		(opt.conf_valid ? (opt.announce_lan ? "used" : "not used") : "unknown"));
+	printf("TLS support: %s [%s]\n",
+		(tls ? "yes" : "no"),
+		(opt.conf_valid ? (opt.tls_use ? "used" : "not used") : "unknown"));
+}
+
+/* Parse command-line options */
+void parse_args(int argc, char *argv[])
+{
+	int sopt, optindex;
+	int t_count;
+	char **t_list;
+
+	while(1)
+	{
+		sopt = getopt_long(argc, argv, "vhsFf:l:p:", options, &optindex);
+		if(sopt == -1) break;
+		switch(sopt)
+		{
+			case 'F':
+				opt.foreground = 1;
+				break;
+			case 'f':
+				opt.local_conf = optarg;
+				break;
+			case 'l':
+				/*log_info.log_types = atoi(optarg);*/
+				t_list = _ggz_string_to_list(optarg, ",");
+				t_count = 0;
+				while((t_list) && (t_list[t_count]))
+					t_count++;
+				log_info.log_types_console = parse_log_types(t_count, t_list);
+				break;
+			case 'p':
+				opt.main_port = atoi(optarg);
+				break;
+			case 'h':
+	 			printf("GGZD - The main server of the GGZ Gaming Zone\n"),
+	 			printf("Copyright (C) 1999 - 2002 Brent Hendricks\n"),
+	 			printf("Copyright (C) 2003 - 2006 The GGZ Gaming Zone developers\n"),
+	 			printf("\n"),
+	 			printf("[-F | --foreground ] Tells ggzd to run in the foreground\n"),
+				printf("[-f | --file <file>] Configuration file\n"),
+				printf("[-l | --log <types>] Types of logging to the console to perform\n"),
+				printf("[-p | --port <port>] GGZ port number\n"),
+				printf("[-s | --specs      ] Display compilation specifications\n"),
+				printf("[-v | --version    ] Display version number\n"),
+				printf("[-h | --help       ] Display this help screen\n"),
+				exit(0);
+				break;
+			case 'v':
+				printf("%s\n", VERSION);
+				exit(0);
+				break;
+			case 's':
+				dump_specs();
+				exit(0);
+				break;
+			default:
+				exit(0);
+				break;
 		}
 	}
-
-	poptFreeContext(context);
-
 }
 
 
@@ -190,8 +262,11 @@ void parse_conf_file(void)
 	}
 
 	/* Get options from the file */
-	if(c_handle >= 0)
+	opt.conf_valid = false;
+	if(c_handle >= 0) {
 		get_config_options(c_handle);
+		opt.conf_valid = true;
+	}
 
 	/* Add any defaults which were not config'ed */
 
@@ -239,19 +314,14 @@ static void get_config_options(int ch)
 	if(opt.main_port == 0)
 		opt.main_port = ggz_conf_read_int(ch, "General", "Port",
 						  DEFAULT_GGZD_PORT);
+	opt.interface = ggz_conf_read_string(ch, "General", "Interface",
+					  NULL);
 	opt.admin_name = ggz_conf_read_string(ch, "General", "AdminName",
 					  "<unconfigured>");
 	opt.admin_email = ggz_conf_read_string(ch, "General", "AdminEmail",
 					  "<unconfigured>");
 	opt.server_name = ggz_conf_read_string(ch, "General", "ServerName",
 					   "An Unconfigured GGZ Server");
-	ggz_conf_read_list(ch, "General", "IPBanList", &t_count, &t_list);
-	client_set_ip_ban_list(t_count, t_list);
-	for (i = 0; i < t_count; i++)
-	  ggz_free(t_list[i]);
-	if (t_list)
-		ggz_free(t_list);
-	t_count = 0;
 
 	/* Encryption in [General] */
 	opt.tls_use = ggz_conf_read_int(ch, "General", "EncryptionUse", 0);
@@ -271,6 +341,10 @@ static void get_config_options(int ch)
 	opt.dbpassword = ggz_conf_read_string(ch, "General", "DatabasePassword", NULL);
 	opt.dbhashing = ggz_conf_read_string(ch, "General", "DatabaseHashing", NULL);
 
+	/* Announcements in [General] */
+	opt.announce_lan = ggz_conf_read_int(ch, "General", "AnnounceLAN", 0);
+	opt.announce_metaserver = ggz_conf_read_int(ch, "General", "AnnounceMetaserver", 0);
+
 	/* [Games] */
 	ggz_conf_read_list(ch, "Games", "GameList", &g_count, &g_list);
 	if(g_count == 0) {
@@ -285,13 +359,14 @@ static void get_config_options(int ch)
 
 	/* [Files] */
 	opt.motd_file = ggz_conf_read_string(ch, "Files", "MOTD", NULL);
+	opt.motd_web = ggz_conf_read_string(ch, "Files", "WebMOTD", NULL);
 
 	/* [Logs] */
 	opt.dump_file = ggz_conf_read_string(ch, "Logs", "DumpFile", NULL);
 	log_info.log_fname = ggz_conf_read_string(ch, "Logs", "LogFile", NULL);
 	ggz_conf_read_list(ch, "Logs", "LogTypes", &t_count, &t_list);
 	if(t_count > 0)
-		log_info.log_types |= parse_log_types(t_count, t_list);
+		log_info.log_types = parse_log_types(t_count, t_list);
 #ifdef DEBUG
 	log_info.dbg_fname = ggz_conf_read_string(ch, "Logs", "DebugFile", NULL);
 	t_count = 0;
@@ -333,11 +408,19 @@ static void get_config_options(int ch)
 
 	/* Miscellaneous */
 	opt.perform_lookups = ggz_conf_read_int(ch, "Miscellaneous", "HostnameLookup", 0);
-	opt.ping_freq = ggz_conf_read_int(ch, "Miscellaneous", "LagFrequency", 10);
-	opt.lag_class[0] = ggz_conf_read_int(ch, "Miscellaneous","LagClass1", 500);
-	opt.lag_class[1] = ggz_conf_read_int(ch, "Miscellaneous","LagClass2", 1000);
-	opt.lag_class[2] = ggz_conf_read_int(ch, "Miscellaneous","LagClass3", 2000);
-	opt.lag_class[3] = ggz_conf_read_int(ch, "Miscellaneous","LagClass4", 5000);
+	opt.ping_freq = ggz_conf_read_int(ch, "Miscellaneous", "PingFrequency", 10);
+	for (i = 0; i < 4; i++) {
+		char buf[1024];
+		int defaults[4] = {500, 1000, 2000, 5000};
+
+		/* The time in the config file is in milliseconds. */
+		snprintf(buf, sizeof(buf), "LagClass%d", i + 1);
+		opt.lag_class_time[i] = ggz_conf_read_int(ch, "Miscellaneous",
+							  buf, defaults[i]);
+		opt.lag_class_time[i] /= 1000.0;
+	}
+	opt.room_update_freq = ggz_conf_read_int(ch, "Miscellaneous",
+						 "RoomUpdateFrequency", 60);
 
 	ggz_conf_cleanup();
 }
@@ -411,6 +494,11 @@ static void parse_game(char *name, char *dir)
 	GameInfo game_info;
 	int len, num_args;
 	char *tmp;
+	int argcp;
+	char **argvp;
+	int ret, i;
+	ggzdbPlayerEntry db_pe;
+	GGZDBResult db_status;
 
 	/* Check to see if we are allocating too many games */
 	if(state.types == MAX_GAME_TYPES) {
@@ -481,7 +569,7 @@ static void parse_game(char *name, char *dir)
 
 	/* [TableOptions] */
 	game_info.allow_leave = ggz_conf_read_int(ch, "TableOptions", 
-						  "AllowLeave",0);
+						  "AllowLeave", 0);
 	game_info.kill_when_empty = ggz_conf_read_int(ch, "TableOptions",
 						      "KillWhenEmpty", 1);
 
@@ -489,6 +577,8 @@ static void parse_game(char *name, char *dir)
 						    "Records", 0);
 	game_info.stats_ratings = ggz_conf_read_int(ch, "Statistics",
 						    "Ratings", 0);
+	game_info.stats_highscores = ggz_conf_read_int(ch, "Statistics",
+						       "Highscores", 0);
 
 	tmp = ggz_conf_read_string(ch, "TableOptions", "BotsAllowed", "");
 	game_info.bot_allow_list = ggz_numberlist_read(tmp);
@@ -501,6 +591,62 @@ static void parse_game(char *name, char *dir)
 	game_info.allow_spectators = ggz_conf_read_int(ch, "TableOptions",
 						       "AllowSpectators", 0);
 
+	game_info.allow_peers = ggz_conf_read_int(ch, "TableOptions", 
+						  "AllowPeers", 0);
+
+	/* Named bots */
+	game_info.named_bots = NULL;
+
+	ret = ggz_conf_get_keys(ch, "NamedBots", &argcp, &argvp);
+	if(ret == 0) {
+		game_info.named_bots = (char***)ggz_malloc((argcp + 1) * sizeof(char**));
+		for (i = 0; i < argcp; i++) {
+			dbg_msg(GGZ_DBG_CONFIGURATION,
+				"Found amed bot <%s>.",
+				argvp[i]);
+			tmp = ggz_conf_read_string(ch, "NamedBots", argvp[i], NULL);
+			game_info.named_bots[i] = (char**)ggz_malloc(2 * sizeof(char*));
+			game_info.named_bots[i][0] = ggz_strdup(argvp[i]);
+			game_info.named_bots[i][1] = ggz_strdup(tmp);
+			ggz_free(tmp);
+			ggz_free(argvp[i]);
+		}
+		game_info.named_bots[argcp] = NULL;
+		ggz_free(argvp);
+	}
+
+	if (game_info.named_bots) {
+		for (i = 0; game_info.named_bots[i]; i++) {
+			snprintf(db_pe.handle, sizeof(db_pe.handle), "%s",
+				game_info.named_bots[i][0]);
+			db_status = ggzdb_player_get(&db_pe);
+			if ((db_pe.perms & PERMS_CHAT_BOT)
+			&& (db_status != GGZDB_ERR_NOTFOUND)) {
+				dbg_msg(GGZ_DBG_CONFIGURATION,
+					"named_bot <%s> registered already.",
+					game_info.named_bots[i][0]);
+				continue;
+			}
+
+			db_pe.user_id = ggzdb_player_next_uid();
+			db_pe.perms = PERMS_DEFAULT_SETTINGS | PERMS_CHAT_BOT;
+			db_pe.last_login = time(NULL);
+			snprintf(db_pe.handle, sizeof(db_pe.handle), "%s",
+				game_info.named_bots[i][0]);
+			strncpy(db_pe.password, "", sizeof(db_pe.password));
+			strncpy(db_pe.email, "", sizeof(db_pe.email));
+			strncpy(db_pe.name, "", sizeof(db_pe.name));
+
+			if (ggzdb_player_add(&db_pe) == GGZDB_ERR_DUPKEY) {
+				err_msg("Could not register named bot <%s>.",
+					game_info.named_bots[i][0]);
+			} else {
+				log_msg(GGZ_LOG_SECURITY, "NEWACCT (%u) (named bot) for %s",
+					db_pe.user_id, game_info.named_bots[i][0]);
+			}
+		}
+	}
+
 	/* Set up data_dir. */
 	len = strlen(opt.data_dir) + strlen("/gamedata/")
               + strlen(game_info.name) + 1;
@@ -705,9 +851,13 @@ static unsigned parse_log_types(int num, char **entry)
 		for(j=0; j<num_log_types; j++)
 			if(!strcasecmp(entry[i], log_types[j].name))
 				break;
-		if(j == num_log_types)
+		if(j == num_log_types) {
 			err_msg("Config: Invalid log type '%s' specified",
 				entry[i]);
+			err_msg("Possible types are:");
+			for(j = 0; j < num_log_types; j++)
+				err_msg("- %s", log_types[j].name);
+		}
 		else {
 			dbg_msg(GGZ_DBG_CONFIGURATION,
 				"%s added to log types", log_types[j].name);
@@ -739,9 +889,13 @@ static void parse_dbg_types(int num, char **entry)
 				/* Enable all debugging types */
 				for (j = 0; dbg_types[j]; j++)
 					ggz_debug_enable(dbg_types[j]);
-			} else
+			} else {
 				err_msg("Config: Invalid debug type '%s' specified",
 					entry[i]);
+				err_msg("Possible types are:");
+				for(j = 0; dbg_types[j]; j++)
+					err_msg("- %s", dbg_types[j]);
+			}
 		} else {
 			/* FIXME: how does this work if debugging isn't set up
 			   yet??? */
