@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 10/18/99
  * Desc: Functions for handling players
- * $Id: players.c,v 1.191 2002/12/07 06:13:17 jdorje Exp $
+ * $Id: players.c 7879 2006-02-27 12:16:44Z josef $
  *
  * Desc: Functions for handling players.  These functions are all
  * called by the player handler thread.  Since this thread is the only
@@ -65,11 +65,6 @@
 #include "transit.h"
 
 
-/* Server wide data structures*/
-extern struct GameInfo game_types[MAX_GAME_TYPES];
-extern struct GGZState state;
-extern Options opt;
-
 pthread_key_t player_key;
 
 
@@ -80,8 +75,9 @@ static GGZClientReqError player_transit(GGZPlayer* player,
 					const int table_index,
 					const int seat_index,
 					int reason);
+static GGZPlayerHandlerStatus player_send_room_update(GGZPlayer *player);
 static GGZPlayerHandlerStatus player_send_ping(GGZPlayer *player);
-static int player_get_time_since_ping(GGZPlayer *player);
+static ggztime_t player_get_time_since_ping(GGZPlayer *player);
 
 
 /**
@@ -103,8 +99,8 @@ GGZPlayer* player_new(GGZClient *client)
 	player->client = client;
 	player->table = -1;
 	player->game_fd = -1;
-	player->launching = 0;
-	player->transit = 0;
+	player->launching = false;
+	player->transit = false;
 	player->room = -1;
 	player->login_status = GGZ_LOGIN_NONE;
 	player->perms = PERMS_DEFAULT_ANON;
@@ -113,11 +109,22 @@ GGZPlayer* player_new(GGZClient *client)
 	player->my_events_head = NULL;
 	player->my_events_tail = NULL;
 	player->lag_class = 1;			/* Assume they are low lag */
-	player->next_ping = 0;			/* Don't ping until login */
+
+	/* Don't ping until login */
+	player->is_ping_sent = false;
+	player->next_ping_time = 0.0;
+	player->sent_ping_time = 0.0;
+
+	/* Don't update until login. */
+	player->is_room_update_needed = false;
+	player->next_room_update_time = 0.0;
+
 	pthread_rwlock_init(&player->stats_lock, NULL);
 	player->wins = -1;
 	player->losses = -1;
 	player->ties = -1;
+	player->rating = -1;
+	player->highscore = -1;
 
 	return player;
 }
@@ -143,14 +150,14 @@ void player_logout(GGZPlayer* player)
 {
 	long connect_time;
 	int hours, mins, secs;
-	int anon = 0;
+	bool anon = false;
 
 	pthread_rwlock_rdlock(&player->lock);
 	/* There's no need to remove them if they aren't "here" */
-	if(strcmp(player->name, "<none>")) {
+	if (strcasecmp(player->name, "<none>") != 0) {
 		connect_time = (long)time(NULL) - player->login_time;
 		if(player->login_status == GGZ_LOGIN_ANON)
-			anon = 1;
+			anon = true;
 		pthread_rwlock_unlock(&player->lock);
 
 		/* Take their name off the hash list */
@@ -206,7 +213,7 @@ void player_logout(GGZPlayer* player)
  */
 GGZReturn player_updates(GGZPlayer* player)
 {
-	int changed = 0;
+	bool changed = false;
 
 	/* There's a bit of a problem here: because the room events will all
 	   be in order, and the player events will all be in order, but
@@ -225,23 +232,31 @@ GGZReturn player_updates(GGZPlayer* player)
 			return GGZ_ERROR;
 
 	/* Lowest priority update - send a PING / check lag */
-	if(player->next_ping && time(NULL) >= player->next_ping) {
+	if (!player->is_ping_sent
+	    && get_current_time() >= player->next_ping_time) {
 		if(player_send_ping(player) < 0)
 			return GGZ_ERROR;
-	} else if (player->next_ping == 0) {
+	} else if (player->is_ping_sent) {
 		/* We're waiting for a PONG, but if we wait too long
 		   we should increase the lag class without waiting */
-		
 		while (player->lag_class < 5
-		       && player_get_time_since_ping(player) >=
-		             opt.lag_class[player->lag_class - 1]) {
+		       && (player_get_time_since_ping(player)
+			   >= opt.lag_class_time[player->lag_class - 1])) {
 			player->lag_class++;
-			changed = 1;
+			changed = true;
 		}
 		
 		if (changed)
 			room_update_event(player->name, GGZ_PLAYER_UPDATE_LAG,
-					  player->room);
+					  player->room, -1);
+	}
+
+	if (player->is_room_update_needed
+	    && get_current_time() >= player->next_room_update_time) {
+		if (player_send_room_update(player) < 0) {
+			return GGZ_ERROR;
+		}
+		player->next_room_update_time += opt.room_update_freq;
 	}
 
 	return GGZ_OK;
@@ -274,6 +289,8 @@ GGZPlayerType player_get_type(GGZPlayer *player)
 		type = GGZ_PLAYER_GUEST;
 	else if(perms_is_admin(player))
 		type = GGZ_PLAYER_ADMIN;
+	else if(perms_is_bot(player))
+		type = GGZ_PLAYER_BOT;
 	else
 		type = GGZ_PLAYER_NORMAL;
 
@@ -323,21 +340,11 @@ GGZPlayerHandlerStatus player_table_launch(GGZPlayer* player, GGZTable *table)
 			return GGZ_REQ_DISCONNECT;
 		return GGZ_REQ_FAIL;
 	}
-	
-#ifndef UNLIMITED_SEATS
-	/* Silly client. Tables are only so big*/
-	if (seats_num(table) > MAX_TABLE_SIZE) {
-		dbg_msg(GGZ_DBG_TABLE, 
-			"%s tried to launch a table with > %d seats",
-			player->name, MAX_TABLE_SIZE);
-		if (net_send_table_launch(player->client->net, E_BAD_OPTIONS) < 0)
-			return GGZ_REQ_DISCONNECT;
-		return GGZ_REQ_FAIL;
-	}
-#else
+
+	/* We used to check for _max_ table size here. Now there is no
+	 * max table size. */
 	/* Shouldn't some check be done?  Note that the number of players is
 	   checked against the player_allow_list later.  --JDS */
-#endif
 	
 	/* Don't allow multiple table launches */
 	if (player->table != -1 || player->launching) {
@@ -378,7 +385,7 @@ GGZPlayerHandlerStatus player_table_launch(GGZPlayer* player, GGZTable *table)
 	if (status == E_OK) {
 		/* Mark player as launching table so we can't launch more */
 		pthread_rwlock_wrlock(&player->lock);
-		player->launching = 1;
+		player->launching = true;
 		pthread_rwlock_unlock(&player->lock);
 	} else {
 		if (net_send_table_launch(player->client->net, status) < 0)
@@ -396,9 +403,9 @@ GGZEventFuncReturn player_launch_callback(void* target, size_t size,
 	GGZLaunchEventData *event = data;
 	GGZPlayer* player = (GGZPlayer*)target;
 
-	/* Launch compleyed */
+	/* Launch completed */
 	pthread_rwlock_wrlock(&player->lock);
-	player->launching = 0;
+	player->launching = false;
 	pthread_rwlock_unlock(&player->lock);
 
 	dbg_msg(GGZ_DBG_TABLE, "%s launch result: %d",
@@ -417,8 +424,8 @@ GGZEventFuncReturn player_launch_callback(void* target, size_t size,
 }
 
 /* Check if the player has permissions to modify the table.  This will need
-   to be more extensive later.  Return the table if they have
-   permissions, otherwise NULL. */
+   to be more extensive later.  Return the table (with write-lock) if they
+   have permissions, otherwise NULL. */
 static GGZTable *check_table_perms(GGZPlayer *player,
 				   int room_id, int table_id)
 {
@@ -429,10 +436,13 @@ static GGZTable *check_table_perms(GGZPlayer *player,
 	if (!table)
 		return NULL;
 
-	if (perms_is_admin(player))
+	/* perms_check does a read-lock on the player.  This will succeed
+	 * since we are the player thread. */
+	if (perms_check(player, PERMS_EDIT_TABLES)) {
 		return table;
+	}
 
-	if (strcmp(table->owner, player->name) != 0) {
+	if (strcasecmp(table->owner, player->name) != 0) {
 		pthread_rwlock_unlock(&table->lock);
 		dbg_msg(GGZ_DBG_TABLE, "%s tried to modify bad table",
 			player->name);
@@ -475,7 +485,7 @@ GGZPlayerHandlerStatus player_table_seat_update(GGZPlayer *player,
 						int room_id, int table_id,
 						GGZTableSeat *seat)
 {
-	int allow;
+	bool allow;
 
 	GGZTable *table = check_table_perms(player, room_id, table_id);
 	if (!table) {
@@ -524,7 +534,7 @@ GGZPlayerHandlerStatus player_table_boot_update(GGZPlayer *player,
 {
 	GGZPlayer *them;
 	GGZClientReqError status;
-	int allow;
+	bool allow;
 
 	GGZTable *table = check_table_perms(player, room_id, table_id);
 	if (!table) {
@@ -682,7 +692,7 @@ GGZPlayerHandlerStatus player_table_leave(GGZPlayer* player,
 {
 	int gametype;
 	GGZClientReqError status;
-	char allow;
+	bool allow;
 	GGZTableState state;
 	GGZTable *table;
 	GGZTransitType transit;
@@ -716,7 +726,7 @@ GGZPlayerHandlerStatus player_table_leave(GGZPlayer* player,
 
 		/* FIXME: make sure the player really is a spectator!
 		   Otherwise the table could break! */
-		allow = 1;
+		allow = true;
 		state = GGZ_TABLE_WAITING;
 
 		transit = GGZ_TRANSIT_LEAVE_SPECTATOR;
@@ -770,7 +780,8 @@ GGZPlayerHandlerStatus player_table_reseat(GGZPlayer *player,
 					   int seat_num)
 {
 	GGZTable *table;
-	int gametype, allow;
+	int gametype;
+	bool allow;
 	GGZClientReqError status;
 	GGZTransitType action;
 
@@ -903,6 +914,89 @@ GGZPlayerHandlerStatus player_table_leave_spectator(GGZPlayer* player)
 #endif
 
 
+GGZPlayerHandlerStatus player_table_info(GGZPlayer *player, int seat_num)
+{
+	GGZTable *table;
+	GameInfo gametype;
+	int i, do_send;
+	int first, last;
+	const char *realname, *photo, *host;
+	int allow_hostname;
+	ggzdbPlayerEntry entry;
+	ggzdbPlayerExtendedEntry extentry;
+	GGZDBResult status;
+
+	table = table_lookup(player->room, player->table);
+	if (!table) {
+		dbg_msg(GGZ_DBG_TABLE,
+			"%s tried to get player info about non-existing table %d:%d.",
+			player->name, player->room, player->table);
+		return GGZ_REQ_FAIL;
+	}
+
+	first = 0;
+	last = table->num_seats;
+	if (seat_num > -1) {
+		first = seat_num;
+		last = seat_num + 1;
+	}
+
+	net_send_info_list_begin(player->client->net);
+
+	for (i = first; i < last; i++) {
+		if (table->seat_types[i] == GGZ_SEAT_PLAYER) {
+			snprintf(entry.handle, sizeof(entry.handle), player->name);
+			status = ggzdb_player_get(&entry);
+
+			do_send = 1;
+			realname = NULL;
+			photo = NULL;
+			host = NULL;
+
+			if (status == GGZDB_NO_ERROR) {
+				realname = ggz_strdup(entry.name);
+
+				snprintf(extentry.handle, sizeof(extentry.handle), player->name);
+				status = ggzdb_player_get_extended(&extentry);
+				if (status == GGZDB_NO_ERROR) {
+					photo = ggz_strdup(extentry.photo);
+				}
+			} else {
+				do_send = 0;
+			}
+
+			allow_hostname = 0;
+
+			pthread_rwlock_rdlock(&game_types[i].lock);
+			gametype = game_types[table->type];
+			allow_hostname = gametype.allow_peers;
+			pthread_rwlock_unlock(&game_types[i].lock);
+
+			if (allow_hostname) {
+				host = ggz_getpeername(net_get_fd(player->client->net), 0);
+				do_send = 1;
+			}
+
+			if (do_send) {
+				net_send_info(player->client->net, i,
+					realname, photo, host);
+
+			}
+
+			ggz_free(realname);
+			ggz_free(photo);
+			ggz_free(host);
+		}
+	}
+	
+	net_send_info_list_end(player->client->net);
+
+	pthread_rwlock_unlock(&table->lock);
+
+	return GGZ_REQ_OK;
+}
+
+
 static void find_player_at_table(GGZPlayer *player,
 				 int table_index,
 				 int *spectating,
@@ -1077,7 +1171,7 @@ static GGZClientReqError player_transit(GGZPlayer* player,
 
 	/* Mark player as "in transit" */
 	pthread_rwlock_wrlock(&player->lock);
-	player->transit = 1;
+	player->transit = true;
 	pthread_rwlock_unlock(&player->lock);
 
 	return E_OK;
@@ -1124,7 +1218,8 @@ GGZPlayerHandlerStatus player_list_players(GGZPlayer* player)
 	}
 	pthread_rwlock_unlock(&rooms[room].lock);
 
-	if (net_send_player_list_count(player->client->net, count) < 0) {
+	if (net_send_player_list_count(player->client->net,
+				       player_get_room(player), count) < 0) {
 		ggz_free(data);
 		return GGZ_REQ_DISCONNECT;
 	}
@@ -1218,7 +1313,8 @@ GGZPlayerHandlerStatus player_list_tables(GGZPlayer* player, int type,
 	count = table_search(player->name, player->room, type, global, 
 			     &my_tables);
 	
-	if (net_send_table_list_count(player->client->net, count) < 0)
+	if (net_send_table_list_count(player->client->net,
+				      player_get_room(player), count) < 0)
 		return GGZ_REQ_DISCONNECT;
 	
 	/* Don`t proceed if there was an error, or no tables found*/
@@ -1244,7 +1340,7 @@ GGZPlayerHandlerStatus player_list_tables(GGZPlayer* player, int type,
 
 
 GGZPlayerHandlerStatus player_chat(GGZPlayer* player, GGZChatType type,
-				   char *target, char *msg)
+				   const char *target, const char *msg)
 {
 	int target_room=-1;	/* FIXME - this should come from net.c if we */
 				/* are going to support per-room announce... */
@@ -1322,41 +1418,68 @@ GGZPlayerHandlerStatus player_motd(GGZPlayer* player)
 }
 
 
-GGZPlayerHandlerStatus player_send_ping(GGZPlayer *player)
+GGZPlayerHandlerStatus player_send_room_update(GGZPlayer *player)
 {
-	/* Send a ping and mark send time */
-	if(net_send_ping(player->client->net) < 0)
-		return GGZ_REQ_DISCONNECT;
-	gettimeofday(&player->sent_ping, NULL);
+	int i;
+	time_t curr_time = time(NULL);
 
-	/* Set next_ping to zero so we won't ping until we've got a reply */
-	player->next_ping = 0;
+	if (opt.room_update_freq == 0) {
+		/*err_msg("Doing a room update even though they're disabled.");*/
+		return GGZ_REQ_OK;
+	}
+
+	for (i = 0; i < room_info.num_rooms; i++) {
+		int player_count;
+		time_t last_player_change;
+
+		pthread_rwlock_rdlock(&rooms[i].lock);
+		player_count = rooms[i].player_count;
+		last_player_change = rooms[i].last_player_change;
+		pthread_rwlock_unlock(&rooms[i].lock);
+
+		if (curr_time - last_player_change <= opt.room_update_freq) {
+			net_send_room_update(player->client->net,
+					     GGZ_ROOM_UPDATE_PLAYER_COUNT, i,
+					     player_count);
+		}
+	}
 
 	return GGZ_REQ_OK;
 }
 
 
-/* Determine time from sent_ping to now, in msec */
-static int player_get_time_since_ping(GGZPlayer *player)
+GGZPlayerHandlerStatus player_send_ping(GGZPlayer *player)
 {
-	struct timeval tv;
+	/* Do not send ping if frequency is set to 0 */
+	if (opt.ping_freq <= 0)
+		return GGZ_REQ_OK;
 
-	gettimeofday(&tv, NULL);
-	return ((tv.tv_sec - player->sent_ping.tv_sec) * 1000)
-	       + ((tv.tv_usec - player->sent_ping.tv_usec) / 1000);
+	/* Send a ping and mark send time */
+	if (net_send_ping(player->client->net) < 0)
+		return GGZ_REQ_DISCONNECT;
+
+	player->sent_ping_time = get_current_time();
+	player->is_ping_sent = true;
+	player->next_ping_time = 0.0;
+
+	return GGZ_REQ_OK;
+}
+
+
+static ggztime_t player_get_time_since_ping(GGZPlayer *player)
+{
+	return get_current_time() - player->sent_ping_time;
 }
 
 
 void player_handle_pong(GGZPlayer *player)
 {
-	int msec, lag_class;
-	int i;
-	
-	msec = player_get_time_since_ping(player);
+	ggztime_t waited = player_get_time_since_ping(player);
+	int lag_class, i;
 
 	lag_class = 1;
 	for(i=0; i<4; i++) {
-		if(msec >= opt.lag_class[i])
+		if (waited >= opt.lag_class_time[i])
 			lag_class++;
 		else
 			break;
@@ -1366,11 +1489,13 @@ void player_handle_pong(GGZPlayer *player)
 		player->lag_class = lag_class;
 		if(player->room >= 0)
 			room_update_event(player->name, GGZ_PLAYER_UPDATE_LAG,
-					  player->room);
+					  player->room, -1);
 	}
 
 	/* Queue our next ping */
-	player->next_ping = time(NULL) + opt.ping_freq;
+	player->is_ping_sent = false;
+	player->sent_ping_time = 0.0;
+	player->next_ping_time = get_current_time() + opt.ping_freq;
 }
 
 
