@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 10/27/2002
  * Desc: Functions for calculating statistics
- * $Id: stats.h,v 1.2 2002/10/28 00:09:53 jdorje Exp $
+ * $Id: stats.h 7067 2005-03-28 19:30:35Z josef $
  *
  * Copyright (C) 2002 GGZ Development Team.
  *
@@ -33,6 +33,8 @@
 GGZReturn stats_lookup(ggzdbPlayerGameStats *stats);
 
 void report_statistics(int room, int gametype,
-		       GGZdModGameReportData *report);
+		       const GGZdModGameReportData *report, const char *savegame);
+
+void report_savegame(int gametype, const char *owner, const char *savegame);
 
 #endif /* _GGZ_STATS_H */
