@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 1/9/00
  * Desc: Functions for handling tables
- * $Id: table.c,v 1.151 2003/02/15 01:41:20 jdorje Exp $
+ * $Id: table.c 7612 2005-11-07 10:39:49Z josef $
  *
  * Copyright (C) 1999-2002 Brent Hendricks.
  *
@@ -69,11 +69,6 @@ typedef struct {
 	int seat;
 } GGZTableEventData;
 
-/* Server wide data structures*/
-extern struct GameInfo game_types[MAX_GAME_TYPES];
-extern struct GGZState state;
-extern Options opt;
-
 /* Local functions for handling tables */
 static GGZTable *table_copy(GGZTable *table);
 static GGZReturn table_check(GGZTable* table);
@@ -84,19 +79,23 @@ static void  table_loop(GGZTable* table);
 static void  table_remove(GGZTable* table);
 
 /* Handlers for ggzdmod events */
-static void table_handle_state(GGZdMod *mod, GGZdModEvent event, void *data);
-static void table_log(GGZdMod *ggzdmod, GGZdModEvent event, void *data);
+static void table_handle_state(GGZdMod *mod, GGZdModEvent event,
+			       const void *data);
+static void table_log(GGZdMod *ggzdmod, GGZdModEvent event, const void *data);
 static void table_game_report(GGZdMod *ggzdmod, GGZdModEvent event,
-			      void *data);
+			      const void *data);
+static void table_game_savegame(GGZdMod *ggzdmod, GGZdModEvent event,
+			      const void *data);
 static void table_game_req_num_seats(GGZdMod *ggzdmod, GGZdModEvent event,
-				     void *data);
+				     const void *data);
 static void table_game_req_boot(GGZdMod *ggzdmod, GGZdModEvent event,
-				void *data);
+				const void *data);
 static void table_game_req_bot(GGZdMod *ggzdmod, GGZdModEvent event,
-			       void *data);
+			       const void *data);
 static void table_game_req_open(GGZdMod *ggzdmod, GGZdModEvent event,
-				void *data);
-static void table_error(GGZdMod *ggzdmod, GGZdModEvent event, void *data);
+				const void *data);
+static void table_error(GGZdMod *ggzdmod, GGZdModEvent event,
+			const void *data);
 
 static GGZReturn table_event_enqueue(GGZTable* table,
 				     GGZTableUpdateType opcode);
@@ -120,9 +119,6 @@ static int   type_match_table(int type, GGZTable* table);
 GGZTable* table_new(void)
 {
 	GGZTable *table;
-#if !defined UNLIMITED_SEATS || !defined UNLIMITED_SPECTATORS
-	int i;
-#endif
 	
 	/* Allocate a new table structure */
 	table = ggz_malloc(sizeof(*table));
@@ -134,22 +130,14 @@ GGZTable* table_new(void)
 	table->state = GGZ_TABLE_CREATED;
 	table->ggzdmod = NULL;
 
-#ifdef UNLIMITED_SEATS
 	table->num_seats = 0;
 	table->seat_types = NULL;
 	table->seat_names = NULL;
-#else
-	for (i = 0; i < MAX_TABLE_SIZE; i++)
-		table->seat_types[i] = GGZ_SEAT_NONE;
-#endif
 
-#ifdef UNLIMITED_SPECTATORS
 	table->max_num_spectators = 0;
 	table->spectators = NULL;
-#else
-	for (i = 0; i < MAX_TABLE_SPECTATORS; i++)
-		table->spectators[i][0] = '\0';
-#endif
+
+	table->savegame = NULL;
 
 	return table;
 }
@@ -172,7 +160,6 @@ static GGZTable *table_copy(GGZTable *table)
 	memcpy(new_table, table, size);
 
 	/* Some elements need to be copied explicitly. */
-#ifdef UNLIMITED_SEATS
 	if (new_table->num_seats > 0) {
 		size = new_table->num_seats * sizeof(*new_table->seat_names);
 		new_table->seat_names = ggz_malloc(size);
@@ -182,8 +169,7 @@ static GGZTable *table_copy(GGZTable *table)
 		new_table->seat_types = ggz_malloc(size);
 		memcpy(new_table->seat_types, table->seat_types, size);
 	}
-#endif
-#ifdef UNLIMITED_SPECTATORS
+
 	if (new_table->max_num_spectators > 0) {
 		size = new_table->max_num_spectators
 			* sizeof(*new_table->spectators);
@@ -191,12 +177,12 @@ static GGZTable *table_copy(GGZTable *table)
 		memcpy(new_table->spectators, table->spectators, size);
 	} else
 		new_table->spectators = NULL;
-#endif
 
 	/* Some elements just shouldn't be copied. */
 	new_table->ggzdmod = NULL;
 	new_table->events_head = NULL;
 	new_table->events_tail = NULL;
+	new_table->savegame = NULL;
 	/* What about table->lock? */
 
 	return new_table;
@@ -408,7 +394,8 @@ static void* table_new_thread(void *index_ptr)
 	else {
 		dbg_msg(GGZ_DBG_TABLE, "Table %d failed to start game module", 
 			table->index);
-		table_launch_event(table->owner, E_LAUNCH_FAIL, 0);
+		/* ggzdmod already does that */
+		/*table_launch_event(table->owner, E_LAUNCH_FAIL, 0);*/
 	}
 	
 	table_remove(table);
@@ -455,6 +442,8 @@ static GGZReturn table_start_game(GGZTable *table)
         ggzdmod_set_handler(table->ggzdmod, GGZDMOD_EVENT_LOG, &table_log);
 	ggzdmod_set_handler(table->ggzdmod, GGZDMOD_EVENT_GAMEREPORT,
 			    &table_game_report);
+	ggzdmod_set_handler(table->ggzdmod, GGZDMOD_EVENT_SAVEGAMEREPORT,
+			    &table_game_savegame);
 	ggzdmod_set_handler(table->ggzdmod, GGZDMOD_EVENT_REQ_NUM_SEATS,
 			    &table_game_req_num_seats);
 	ggzdmod_set_handler(table->ggzdmod, GGZDMOD_EVENT_REQ_BOOT,
@@ -471,7 +460,7 @@ static GGZReturn table_start_game(GGZTable *table)
         for (i = 0; i < num_seats; i++) {
 		seat.num = i;
 		seat.type = seats_type(table, i);
-		if (seat.type == GGZ_SEAT_RESERVED)
+		if (seat.type == GGZ_SEAT_RESERVED || seat.type == GGZ_SEAT_BOT)
 			seat.name = table->seat_names[i];
 		else
 			seat.name = NULL;
@@ -532,7 +521,7 @@ static void table_loop(GGZTable* table)
  * table_game_join handles the RSP_GAME_JOIN from the table
  * Note: table->transit_name contains allocated mem on entry
  */
-void table_game_join(GGZTable *table, char *name,
+void table_game_join(GGZTable *table, const char *name,
 		     GGZJoinType reason, int num)
 {
 #if 0
@@ -555,10 +544,10 @@ void table_game_join(GGZTable *table, char *name,
 		/* If player notification failed, they must've logged out.
 		   We have to get them out of the table, but since they've
 		   already joined we have to do this the long way. */
-		struct GGZTableSeat seat = {index: num,
-					    type: GGZ_SEAT_OPEN,
-					    name: "",
-					    fd: -1};
+		struct GGZTableSeat seat = {.index = num,
+					    .type = GGZ_SEAT_OPEN,
+					    .name = "",
+					    .fd = -1};
 
 		dbg_msg(GGZ_DBG_TABLE, "%s logged out during join", name);
 		transit_seat_event(table->room, table->index,
@@ -579,7 +568,8 @@ void table_game_join(GGZTable *table, char *name,
 	   may still be necessary to always double-check. */
 	/* Remove any seat reservations this player may have. */
 	for (i = 0; i < table->num_seats; i++) {
-		if (table->seat_types[i] == GGZ_SEAT_RESERVED
+		if ((table->seat_types[i] == GGZ_SEAT_RESERVED
+		     || table->seat_types[i] == GGZ_SEAT_ABANDONED)
 		    && !strcasecmp(table->seat_names[i], name)) {
 			pthread_rwlock_wrlock(&table->lock);
 			table->seat_types[i] = GGZ_SEAT_OPEN;
@@ -590,21 +580,20 @@ void table_game_join(GGZTable *table, char *name,
 		}
 	}
 #endif
-
-	dbg_msg(GGZ_DBG_TABLE, 
-		"%s in seat %d at table %d of room %d",
-		name, num,
-		table->index, table->room);
+	log_msg(GGZ_LOG_TABLES,
+		"TABLE_JOIN - %s joined seat %d at table %d of room %s in game of %s",
+		name, num, table->index, rooms[table->room].name,
+		game_types[rooms[table->room].game_type].name);
 }
 
 
 /*
  * table_game_leave handles the RSP_GAME_LEAVE from the table
  */
-void table_game_leave(GGZTable *table, char *caller,
+void table_game_leave(GGZTable *table, const char *caller,
 		      GGZLeaveType reason, int num)
 {
-	char empty;
+	bool empty;
 	char player[MAX_USER_NAME_LEN + 1];
 	strcpy(player, table->seat_names[num]);
 
@@ -612,10 +601,11 @@ void table_game_leave(GGZTable *table, char *caller,
 		table->index, table->room);
 
 	/* Vacate seat */
-	dbg_msg(GGZ_DBG_TABLE, 
-		"%s left seat %d at table %d of room %d",
-		player, num, table->index, table->room);
-
+	log_msg(GGZ_LOG_TABLES,
+		"TABLE_LEAVE - %s left seat %d at table %d of room %s in game of %s",
+		player, num, table->index, rooms[table->room].name,
+		game_types[rooms[table->room].game_type].name);
+	
 	pthread_rwlock_wrlock(&table->lock);
 	table->seat_types[num] = GGZ_SEAT_OPEN;
 	pthread_rwlock_unlock(&table->lock);
@@ -644,7 +634,7 @@ void table_game_leave(GGZTable *table, char *caller,
 
 void table_game_reseat(GGZTable *table,
 		       GGZReseatType op,
-		       char *name,
+		       const char *name,
 		       int old_seat, int new_seat)
 {
 	int empty = 0;
@@ -689,10 +679,10 @@ void table_game_reseat(GGZTable *table,
 		/* If player notification failed, they must've logged out.
 		   We have to get them out of the table, but since they've
 		   already joined we have to do this the long way. */
-		struct GGZTableSeat seat = {index: new_seat,
-					    type: GGZ_SEAT_OPEN,
-					    name: "",
-					    fd: -1};
+		struct GGZTableSeat seat = {.index = new_seat,
+					    .type = GGZ_SEAT_OPEN,
+					    .name = "",
+					    .fd = -1};
 
 		if (op == GGZ_RESEAT_STAND)
 			transit = GGZ_TRANSIT_LEAVE_SPECTATOR;
@@ -748,7 +738,7 @@ void table_game_seatchange(GGZTable *table, GGZSeatType type, int num)
  * table_game_spectator_join handles the RSP_GAME_JOIN_SPECTATOR from the table
  * Note: table->transit_name contains allocated mem on entry
  */
-void table_game_spectator_join(GGZTable *table, char *name,
+void table_game_spectator_join(GGZTable *table, const char *name,
 			       GGZJoinType reason, int num)
 {
 	dbg_msg(GGZ_DBG_TABLE,
@@ -766,9 +756,9 @@ void table_game_spectator_join(GGZTable *table, char *name,
 		/* If player notification failed, they must've logged out.
 		   We have to get them out of the table, but since they've
 		   already joined we have to do this the long way. */
-		struct GGZTableSeat seat = {index: num,
-					    name: "",
-					    fd: -1};
+		struct GGZTableSeat seat = {.index = num,
+					    .name = "",
+					    .fd = -1};
 
 		dbg_msg(GGZ_DBG_TABLE, "%s logged out during spectator join",
 			name);
@@ -791,7 +781,7 @@ void table_game_spectator_join(GGZTable *table, char *name,
 /*
  * table_game_leave handles the RSP_GAME_LEAVE_SPECTATOR from the table
  */
-void table_game_spectator_leave(GGZTable *table, char *caller,
+void table_game_spectator_leave(GGZTable *table, const char *caller,
 				GGZLeaveType reason, int num)
 {
 	char player[MAX_USER_NAME_LEN + 1];
@@ -822,11 +812,15 @@ void table_game_spectator_leave(GGZTable *table, char *caller,
 }
 
 
-static void table_handle_state(GGZdMod *mod, GGZdModEvent event, void *data)
+static void table_handle_state(GGZdMod *mod, GGZdModEvent event,
+			       const void *data)
 {
 	GGZTable* table = ggzdmod_get_gamedata(mod);
-	/* GGZdModState prev = *(GGZdModState*)data; */
+#if 0
+	const GGZdModState *prev = data;
+#endif
 	GGZdModState cur = ggzdmod_get_state(mod);
+	GGZTableState origstate;
 
 	switch (cur) {
 	case GGZDMOD_STATE_WAITING:
@@ -852,7 +846,7 @@ static void table_handle_state(GGZdMod *mod, GGZdModEvent event, void *data)
 		pthread_rwlock_wrlock(&table->lock);
 		table->state = GGZ_TABLE_PLAYING;
 		pthread_rwlock_unlock(&table->lock);
-		
+
 		table_event_enqueue(table, GGZ_TABLE_UPDATE_STATE);
 		dbg_msg(GGZ_DBG_TABLE, "Table %d in room %d now full/playing",
 			table->index, table->room);
@@ -865,9 +859,12 @@ static void table_handle_state(GGZdMod *mod, GGZdModEvent event, void *data)
 
 		/* Mark table as done, so people don't attempt transits */
 		pthread_rwlock_wrlock(&table->lock);
-		table->state = GGZ_TABLE_DONE;
+		origstate = table->state;
+		if (origstate != GGZ_TABLE_CREATED)
+			table->state = GGZ_TABLE_DONE;
 		pthread_rwlock_unlock(&table->lock);
-		table_event_enqueue(table, GGZ_TABLE_UPDATE_STATE);
+		if (origstate != GGZ_TABLE_CREATED)
+			table_event_enqueue(table, GGZ_TABLE_UPDATE_STATE);
 		return;
 
 	case GGZDMOD_STATE_CREATED:
@@ -880,11 +877,11 @@ static void table_handle_state(GGZdMod *mod, GGZdModEvent event, void *data)
 }
 
 
-static void table_log(GGZdMod *ggzdmod, GGZdModEvent event, void *data)
+static void table_log(GGZdMod *ggzdmod, GGZdModEvent event, const void *data)
 {
 	GGZTable* table = ggzdmod_get_gamedata(ggzdmod);
 	int type;
-	char *game_name, *msg = data;
+	const char *game_name, *msg = data;
 
 	if (log_info.options & GGZ_LOGOPT_INC_GAMETYPE) {
 		pthread_rwlock_rdlock(&table->lock);
@@ -895,25 +892,38 @@ static void table_log(GGZdMod *ggzdmod, GGZdModEvent event, void *data)
 		game_name = ggz_strdup(game_types[type].name);
 		pthread_rwlock_unlock(&game_types[type].lock);
 		
-		dbg_msg(GGZ_DBG_TABLE, "(%s) %s", game_name, msg);
+		dbg_msg(GGZ_DBG_GAME_MSG, "(%s) %s", game_name, msg);
 		
 		ggz_free(game_name);
 	} else {
 		/* It's important that we use the %s so that we can
 		   safely log messages containing % and other printf
 		   meta-characters. */
-		dbg_msg(GGZ_DBG_TABLE, "%s", msg);
+		dbg_msg(GGZ_DBG_GAME_MSG, "%s", msg);
 	}
 }
 
 
 static void table_game_report(GGZdMod *ggzdmod,
-			      GGZdModEvent event, void *data)
+			      GGZdModEvent event, const void *data)
 {
 	GGZTable *table = ggzdmod_get_gamedata(ggzdmod);
-	GGZdModGameReportData *report = data;
+	const GGZdModGameReportData *report = data;
 
-	report_statistics(table->room, table->type, report);
+	report_statistics(table->room, table->type, report, table->savegame);
+}
+
+
+static void table_game_savegame(GGZdMod *ggzdmod,
+			      GGZdModEvent event, const void *data)
+{
+	GGZTable *table = ggzdmod_get_gamedata(ggzdmod);
+
+	if (table->savegame)
+		ggz_free(table->savegame);
+	table->savegame = ggz_strdup((char*)data);
+
+	report_savegame(table->type, table->owner, table->savegame);
 }
 
 
@@ -929,10 +939,11 @@ static void table_change_num_seats(GGZTable *table, int num_seats)
 
 
 static void table_game_req_num_seats(GGZdMod *ggzdmod, GGZdModEvent event,
-				     void *data)
+				     const void *data)
 {
 	GGZTable *table = ggzdmod_get_gamedata(ggzdmod);	
-	int num_seats = *(int*)data;
+	const int *num_seats_ptr = data;
+	const int num_seats = *num_seats_ptr;
 	int old_seats = table->num_seats;
 	int seat;
 
@@ -988,11 +999,12 @@ static void table_game_req_num_seats(GGZdMod *ggzdmod, GGZdModEvent event,
 
 
 static void table_game_req_boot(GGZdMod *ggzdmod,
-				GGZdModEvent event, void *data)
+				GGZdModEvent event, const void *data)
 {
 	GGZTable *table = ggzdmod_get_gamedata(ggzdmod);
-	char *name = data;
-	int seat_num, is_spectator = 0, found = 1;
+	const char *name = data;
+	int seat_num;
+	bool is_spectator = false, found = true;
 	int len = strlen(name);
 	GGZTransitType transit;
 	GGZTableUpdateType update;
@@ -1012,14 +1024,14 @@ static void table_game_req_boot(GGZdMod *ggzdmod,
 			break;
 	}
 	if (seat_num == table->num_seats) {
-		is_spectator = 1;
+		is_spectator = true;
 		for (seat_num = 0; seat_num < table->max_num_spectators;
 		     seat_num++) {
-			if (strcmp(table->spectators[seat_num], name) == 0)
+			if (strcasecmp(table->spectators[seat_num], name) == 0)
 				break;
 		}
 		if (seat_num == table->max_num_spectators)
-			found = 0;
+			found = false;
 	}
 
 	/* Update the seat. */
@@ -1027,9 +1039,9 @@ static void table_game_req_boot(GGZdMod *ggzdmod,
 		/* FIXME: send error to game... */
 		return;
 	} else if (is_spectator) {
-		GGZSpectator seat = {num: seat_num,
-				     name: NULL,
-				     fd: -1};
+		GGZSpectator seat = {.num = seat_num,
+				     .name = NULL,
+				     .fd = -1};
 		dbg_msg(GGZ_DBG_TABLE,
 			"Table %d/%d: server is booting spectator %s.",
 			table->index, table->room, name);
@@ -1037,10 +1049,10 @@ static void table_game_req_boot(GGZdMod *ggzdmod,
 		transit = GGZ_TRANSIT_LEAVE_SPECTATOR;
 		update = GGZ_TABLE_UPDATE_SPECTATOR_LEAVE;
 	} else {
-		GGZSeat seat = {num: seat_num,
-				name: NULL,
-				type: GGZ_SEAT_OPEN,
-				fd: -1};
+		GGZSeat seat = {.num = seat_num,
+				.name = NULL,
+				.type = GGZ_SEAT_OPEN,
+				.fd = -1};
 		dbg_msg(GGZ_DBG_TABLE,
 			"Table %d/%d: server is booting seat %s.",
 			table->index, table->room, name);
@@ -1071,10 +1083,11 @@ static void table_game_req_boot(GGZdMod *ggzdmod,
 
 
 static void table_game_req_bot(GGZdMod *ggzdmod,
-			       GGZdModEvent event, void *data)
+			       GGZdModEvent event, const void *data)
 {
 	GGZTable *table = ggzdmod_get_gamedata(ggzdmod);
-	int seat_num = *(int*)data;
+	const int *seat_num_ptr = data;
+	const int seat_num = *seat_num_ptr;
 	GGZSeat seat;
 
 	/* FIXME: this code overlaps with the leaving code in transit.c */
@@ -1108,10 +1121,11 @@ static void table_game_req_bot(GGZdMod *ggzdmod,
 
 
 static void table_game_req_open(GGZdMod *ggzdmod,
-				GGZdModEvent event, void *data)
+				GGZdModEvent event, const void *data)
 {
 	GGZTable *table = ggzdmod_get_gamedata(ggzdmod);
-	int seat_num = *(int*)data;
+	const int *seat_num_ptr = data;
+	const int seat_num = *seat_num_ptr;
 	GGZSeat seat;
 
 	/* FIXME: this code overlaps with the leaving code in transit.c */
@@ -1145,7 +1159,8 @@ static void table_game_req_open(GGZdMod *ggzdmod,
 }
 
 
-static void table_error(GGZdMod *ggzdmod, GGZdModEvent event, void *data)
+static void table_error(GGZdMod *ggzdmod,
+			GGZdModEvent event, const void *data)
 {
 	GGZTable* table = ggzdmod_get_gamedata(ggzdmod);
 	
@@ -1167,6 +1182,7 @@ static void table_remove(GGZTable* table)
 {
 	int room, count, index, i;
 	GGZLeaveType reason;
+	GGZTableState origstate;
 
 	/* Disconnect from the game server */
 	(void)ggzdmod_disconnect(table->ggzdmod);
@@ -1224,11 +1240,12 @@ static void table_remove(GGZTable* table)
 
 	pthread_rwlock_wrlock(&table->lock);
 	table->type = -1;
+	origstate = table->state;
 	table->state = GGZ_TABLE_ERROR;
 	pthread_rwlock_unlock(&table->lock);
 
-	table_event_enqueue(table, GGZ_TABLE_UPDATE_DELETE);
-
+	if (origstate != GGZ_TABLE_CREATED)
+		table_event_enqueue(table, GGZ_TABLE_UPDATE_DELETE);
 
 	/* Process any remaining events */
 	event_table_handle(table);
@@ -1236,7 +1253,7 @@ static void table_remove(GGZTable* table)
 
 
 /* FIXME: redo using events to notify player */
-GGZClientReqError table_launch(GGZTable *table, char* name)
+GGZClientReqError table_launch(GGZTable *table, const char *name)
 {
 	/* Fill in the table owner */
 	strcpy(table->owner, name);	
@@ -1271,7 +1288,7 @@ void table_set_desc(GGZTable *table, const char *desc)
 
 
 /* Kill the table */
-GGZClientReqError table_kill(int room, int index, char *name)
+GGZClientReqError table_kill(int room, int index, const char *name)
 {
 	char *data;
 
@@ -1306,7 +1323,7 @@ static GGZEventFuncReturn table_kill_callback(void* target, size_t size,
 					      
 
 /* Search for tables */
-int table_search(char* name, int room, int type, char global, 
+int table_search(const char *name, int room, int type, bool global, 
 		 GGZTable*** tables)
 {
 	GGZTable* t;
@@ -1380,7 +1397,7 @@ int table_search(char* name, int room, int type, char global,
 
 
 /* Find a player at a table */
-int table_find_player(int room, int index, char *name)
+int table_find_player(int room, int index, const char *name)
 {
 	int i, seats, seat = -1;
 	GGZTable *table;
@@ -1402,7 +1419,7 @@ int table_find_player(int room, int index, char *name)
 }
 
 /* Find a player at a table */
-int table_find_spectator(int room, int index, char *name)
+int table_find_spectator(int room, int index, const char *name)
 {
 	int i, spectator = -1;
 	GGZTable *table;
@@ -1552,7 +1569,8 @@ static GGZEventFuncReturn table_event_callback(void* target, size_t size,
 	}
 
 	if (net_send_table_update(player->client->net, event->opcode,
-				  event->table, update_data) < 0)
+				  event->table, update_data,
+				  player_get_room(player)) < 0)
 		return GGZ_EVENT_ERROR;
 	
 	return GGZ_EVENT_OK;
@@ -1575,20 +1593,20 @@ static GGZReturn table_launch_event(char* name,
 /* Free dynamically allocated memory associated with a table*/
 void table_free(GGZTable* table)
 {
-#ifdef UNLIMITED_SEATS
 	if (table->num_seats > 0) {
 		ggz_free(table->seat_types);
 		ggz_free(table->seat_names);
 	}
-#endif
-#ifdef UNLIMITED_SPECTATORS
-	if (table->max_num_spectators > 0)
+	if (table->max_num_spectators > 0) {
 		ggz_free(table->spectators);
-#endif
+	}
 
 	if (table->ggzdmod)
 		ggzdmod_free(table->ggzdmod);
 
+	if (table->savegame)
+		ggz_free(table->savegame);
+
 	ggz_free(table);
 }
 
