@@ -4,7 +4,7 @@
  * Project: GGZ Server
  * Date: 05/04/2002 (code moved from control.c)
  * Desc: General utility functions for ggzd
- * $Id: util.h,v 1.1 2002/05/04 23:46:47 jdorje Exp $
+ * $Id: util.h 6416 2004-11-17 22:02:24Z jdorje $
  *
  * Copyright (C) 1999-2002 Brent Hendricks.
  *
@@ -23,5 +23,20 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  */
 
+#ifndef __UTIL_H__
+#define __UTIL_H__
+
+#define MAX(a, b) ((a) > (b) ? (a) : (b))
+#define MIN(a, b) ((a) < (b) ? (a) : (b))
+
 /* Functionally similar to "mkdir -p". */
 void check_path(const char* full_path);
+
+/* Functions for handling time in an "easy" way. */
+typedef double ggztime_t;
+
+ggztime_t timeval_to_ggztime(struct timeval tv);
+struct timeval ggztime_to_timeval(ggztime_t ggztime);
+ggztime_t get_current_time(void);
+
+#endif /* __UTIL_H__ */
