@@ -4,7 +4,7 @@
  * Project: ggzdmod
  * Date: 10/14/01
  * Desc: GGZ game module functions
- * $Id: ggzdmod.c,v 1.113 2003/02/12 04:37:55 jdorje Exp $
+ * $Id: ggzdmod.c 7887 2006-03-07 09:56:51Z josef $
  *
  * This file contains the backend for the ggzdmod library.  This
  * library facilitates the communication between the GGZ server (ggzd)
@@ -13,19 +13,19 @@
  *
  * Copyright (C) 2001-2002 GGZ Development Team.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -97,7 +97,7 @@ static void call_handler(GGZdMod * ggzdmod, GGZdModEvent event, void *data)
 		fprintf(stderr,
 			"GGZDMOD: call_handler called for unknown event %d.\n"
 			"This is a bug in libggzdmod.  Please e-mail the GGZ\n"
-			"development team at ggz-dev@lists.sourceforge.net\n"
+			"development team at ggz-dev@mail.ggzgamingzone.org\n"
 			"to report it.\n", event);
 		return;
 	}
@@ -335,10 +335,10 @@ int ggzdmod_get_max_num_spectators(GGZdMod * ggzdmod)
 
 GGZSeat ggzdmod_get_seat(GGZdMod * ggzdmod, int num)
 {
-	GGZSeat seat = {num: num,
-			type: GGZ_SEAT_NONE,
-			name: NULL,
-			fd: -1};
+	GGZSeat seat = {.num = num,
+			.type = GGZ_SEAT_NONE,
+			.name = NULL,
+			.fd = -1};
 	GGZListEntry *entry;
 
 	if (CHECK_GGZDMOD(ggzdmod) && num >= 0 && num < ggzdmod->num_seats) {
@@ -352,7 +352,7 @@ GGZSeat ggzdmod_get_seat(GGZdMod * ggzdmod, int num)
 
 GGZSpectator ggzdmod_get_spectator(GGZdMod * ggzdmod, int num)
 {
-	GGZSpectator spectator = {num: num, name: NULL, fd: -1};
+	GGZSpectator spectator = {.num = num, .name = NULL, .fd = -1};
 	GGZListEntry *entry;
 
 	if (CHECK_GGZDMOD(ggzdmod)
@@ -383,15 +383,15 @@ static void _ggzdmod_set_num_seats(GGZdMod *ggzdmod, int num_seats)
 	if (num_seats > old_num) {
 		/* Initialize new seats */
 		for (i = old_num; i < num_seats; i++) {
-			GGZSeat seat = {num: i,
-					type: GGZ_SEAT_OPEN,
-					name: NULL,
-					fd: -1};
+			GGZSeat seat = {.num = i,
+					.type = GGZ_SEAT_OPEN,
+					.name = NULL,
+					.fd = -1};
 			ggz_list_insert(ggzdmod->seats, &seat);
 		}
 	} else if (num_seats < old_num) {
 		GGZListEntry *entry;
-		GGZSeat seat = {num: num_seats};
+		GGZSeat seat = {.num = num_seats};
 
 		/* Delete extra seats (FIXME: UNTESTED) */
 		entry = ggz_list_search(ggzdmod->seats, &seat);
@@ -406,6 +406,26 @@ static void _ggzdmod_set_num_seats(GGZdMod *ggzdmod, int num_seats)
 }
 
 
+char* ggzdmod_get_bot_class(GGZdMod *ggzdmod, const char *name)
+{
+	int ch;
+	char *conffile;
+	const char *game;
+	char *botclass;
+
+	game = "tictactoe";
+	conffile = (char*)ggz_malloc(strlen(GGZDCONFDIR) + strlen("games") + strlen(game) + 7);
+	sprintf(conffile, "%s/games/%s.dsc", GGZDCONFDIR, game);
+	ch = ggz_conf_parse(conffile, GGZ_CONF_RDONLY);
+	if(ch < 0) return NULL;
+	ggz_free(conffile);
+
+	botclass = ggz_conf_read_string(ch, "NamedBots", name, NULL);
+	ggz_conf_close(ch);
+
+	return botclass;
+}
+
 static void _ggzdmod_set_max_num_spectators(GGZdMod *ggzdmod,
 					    int num_spectators)
 {
@@ -556,12 +576,19 @@ int ggzdmod_set_seat(GGZdMod * ggzdmod, GGZSeat *seat)
 	/* If we're connected to the game, send a message */
 	if (ggzdmod->type == GGZDMOD_GGZ
 	    && ggzdmod->state != GGZDMOD_STATE_CREATED) {
-		if (_io_send_seat_change(ggzdmod->fd, seat) < 0)
+		if (_io_send_seat_change(ggzdmod->fd, seat) < 0) {
+			/* ggzdmod_error will most likely call
+			   ggzdmod_disconnect - but this doesn't free the
+			   seat data.  Be careful with this! */
 			_ggzdmod_error(ggzdmod,
 				       "Error writing to game");
+		}
 
 		/* We (GGZ) don't need the fd now */
-		seat->fd = -1;
+		if (seat->type == GGZ_SEAT_PLAYER) {
+			close(seat->fd);
+			seat->fd = -1;
+		}
 	}
 
 	return _ggzdmod_set_seat(ggzdmod, seat);
@@ -636,7 +663,7 @@ int ggzdmod_reseat(GGZdMod * ggzdmod,
 		   int old_seat, int was_spectator,
 		   int new_seat, int is_spectator)
 {
-	char *name;
+	const char *name;
 
 	if (!CHECK_GGZDMOD(ggzdmod) || ggzdmod->type != GGZDMOD_GGZ)
 		return -1;
@@ -682,31 +709,46 @@ int ggzdmod_reseat(GGZdMod * ggzdmod,
 	name = ggz_strdup(name);
 
 	if (was_spectator) {
-		GGZSpectator s = {num: old_seat,
-				  name: NULL,
-				  fd: -1};
+		GGZSpectator s = {.num = old_seat,
+				  .name = NULL,
+				  .fd = -1};
 		if (_ggzdmod_set_spectator(ggzdmod, &s) < 0)
 			_ggzdmod_error(ggzdmod, "ggzdmod_reseat failed");
 	} else {
-		GGZSeat s = {num: old_seat,
-			     name: NULL,
-			     type: GGZ_SEAT_OPEN,
-			     fd: -1};
+		GGZSeat s = {.num = old_seat,
+			     .name = NULL,
+			     .type = GGZ_SEAT_OPEN,
+			     .fd = -1};
+		const char *name;
+
+		name = ggz_strdup(ggzdmod_get_seat(ggzdmod, old_seat).name);
+		if (ggzdmod->state == GGZDMOD_STATE_PLAYING) {
+			/* Mark the seat as abandoned rather than open.
+			   FIXME - this should probably be done inside ggzd
+			   rather than here, which would allow it to be
+			   controlled by game options. However this will
+			   take some interface changes.  See also the comment
+			   in handle_reseat. */
+			s.name = name;
+			s.type = GGZ_SEAT_ABANDONED;
+		}
+
 		if (_ggzdmod_set_seat(ggzdmod, &s) < 0)
 			_ggzdmod_error(ggzdmod, "ggzdmod_reseat failed");
+		ggz_free(name); /* It's been copied in _ggzdmod_set_seat. */
 	}
 
 	if (is_spectator) {
-		GGZSpectator s = {num: new_seat,
-				  name: name,
-				  fd: -1};
+		GGZSpectator s = {.num = new_seat,
+				  .name = name,
+				  .fd = -1};
 		if (_ggzdmod_set_spectator(ggzdmod, &s) < 0)
 			_ggzdmod_error(ggzdmod, "ggzdmod_reseat failed");
 	} else {
-		GGZSeat s = {num: new_seat,
-			     name: name,
-			     type: GGZ_SEAT_PLAYER,
-			     fd: -1};
+		GGZSeat s = {.num = new_seat,
+			     .name = name,
+			     .type = GGZ_SEAT_PLAYER,
+			     .fd = -1};
 		if (_ggzdmod_set_seat(ggzdmod, &s) < 0)
 			_ggzdmod_error(ggzdmod, "ggzdmod_reseat failed");
 	}
@@ -964,6 +1006,27 @@ int ggzdmod_loop(GGZdMod * ggzdmod)
 	return 0;		/* should handle errors */
 }
 
+/* FIXME
+
+For seats, any change must be done in ggzdmod-ggz.  This means a single point
+ where a change can be made.  If the game server wants to change seats it
+sends a request to ggzd which is then approved (or not) and the new seat 
+data sent back.
+
+For game state, however, it is the opposite.  The change must be done in 
+ggzdmod-game.  There is currently no way for the ggzd to change the state 
+(maybe there never will be).
+
+However this conflicts.  Some seat changes should depend on the game 
+state, and the game state as known by ggzdmod-ggz (which determines seat 
+changes) may be out of date.  The solution is that the same mechanism 
+should be used for state changes as is used for seat changes.  This may 
+affect games however because it means the state change is asynchronous and 
+not immediate (there is a state-change callback in ggzdmod, that currently 
+gets invoked from inside ggzdmod_set_state; if this were changed then it 
+wouldn't be invoked until the message was sent to ggzd and the response 
+was received).
+*/
 int ggzdmod_set_state(GGZdMod * ggzdmod, GGZdModState state)
 {
 	if (!CHECK_GGZDMOD(ggzdmod))
@@ -1098,6 +1161,10 @@ static int game_fork(GGZdMod * ggzdmod)
 			/* FIXME: what to do? */
 		}
 
+		/* Setup environment variables */
+		setenv("GGZMODE", "true", 1);
+		setenv("GGZSOCKET", "3", 1);
+
 		/* FIXME: can we call ggzdmod_log() from here? */
 		execv(ggzdmod->argv[0], ggzdmod->argv);	/* run game */
 
@@ -1121,6 +1188,10 @@ static int game_fork(GGZdMod * ggzdmod)
 
 int ggzdmod_connect(GGZdMod * ggzdmod)
 {
+	char *ggzsocketstr;
+	int ggzsocket;
+	int items;
+
 	if (!CHECK_GGZDMOD(ggzdmod)) {
 		return -1;
 	}
@@ -1142,7 +1213,15 @@ int ggzdmod_connect(GGZdMod * ggzdmod)
 		}
 	} else {
 		/* For the game side we setup the fd */
-		ggzdmod->fd = 3;
+		ggzsocket = 3;
+		ggzsocketstr = getenv("GGZSOCKET");
+		if(ggzsocketstr) {
+			items = sscanf(ggzsocketstr, "%d", &ggzsocket);
+			if (items == 0) {
+				ggzsocket = 3;
+			}
+		}
+		ggzdmod->fd = ggzsocket;
 		
 		if (ggzdmod_log(ggzdmod, "GGZDMOD: Connecting to GGZ server.") < 0) {
 			ggzdmod->fd = -1;
@@ -1189,7 +1268,9 @@ int ggzdmod_disconnect(GGZdMod * ggzdmod)
 	}
 	
 	/* We no longer free the seat data here.  It will stick around until
-	   ggzdmod_free is called or it is used again. */
+	   ggzdmod_free is called or it is used again.  This is important
+	   because we don't know who the callers are or if they expect to
+	   have valid seat data to look at (see ggzdmod_set_seat). */
 
 	/* Clean up the ggzdmod object.  In theory it could now reconnect for
 	   a new game. */
@@ -1204,7 +1285,7 @@ int ggzdmod_disconnect(GGZdMod * ggzdmod)
  * module specific actions
  */
 
-int ggzdmod_log(GGZdMod * ggzdmod, char *fmt, ...)
+int ggzdmod_log(GGZdMod * ggzdmod, const char *fmt, ...)
 {
 	char buf[4096];
 	va_list ap;
@@ -1267,13 +1348,13 @@ void _ggzdmod_handle_state(GGZdMod * ggzdmod, GGZdModState state)
 		   separately change states when the launch packet
 		   is sent. */
 		set_state(ggzdmod, state);
+		/* Is this right? has the gameover happened yet? */   
+		return;
+	case GGZDMOD_STATE_CREATED:
 		break;
-	default:
-		_ggzdmod_error(ggzdmod,
-			       "Game requested incorrect state value");
 	}
-	
-	/* Is this right? has the gameover happened yet? */   
+
+	_ggzdmod_error(ggzdmod, "Game requested incorrect state value");
 }
 
 
@@ -1309,15 +1390,19 @@ void _ggzdmod_handle_launch_begin(GGZdMod * ggzdmod, int num_seats, int num_spec
 
 void _ggzdmod_handle_launch_seat(GGZdMod * ggzdmod, GGZSeat seat)
 {
+	int status = -1;
+
 	switch (seat.type) {
 	case GGZ_SEAT_OPEN:
 		ggzdmod_log(ggzdmod, "GGZDMOD: Seat %d is open", seat.num);
+		status = 0;
 		break;
 
 	case GGZ_SEAT_BOT:
 		/* Eventually ggzd may give us AI player names.  In the
 		   meantime there's just a placeholder. */
-		seat.name = "AI";
+		if(!seat.name)
+			seat.name = "AI";
 #if 0		
 		len = strlen(rand_bot_names[bots]) + 4;
 		seat.name = ggz_malloc(len);
@@ -1328,20 +1413,29 @@ void _ggzdmod_handle_launch_seat(GGZdMod * ggzdmod, GGZSeat seat)
 #endif
 		ggzdmod_log(ggzdmod, "GGZDMOD: Seat %d is a bot named %s",
 			    seat.num, seat.name);
+		status = 0;
 		break;
 
 	case GGZ_SEAT_RESERVED:
 		ggzdmod_log(ggzdmod, "GGZDMOD: Seat %d reserved for %s",
 			    seat.num, seat.name);
+		status = 0;
 		break;
-	default:
+	case GGZ_SEAT_NONE:
+	case GGZ_SEAT_PLAYER:
+	case GGZ_SEAT_ABANDONED:
+		break;
+	}
+
+
+	if (status < 0) {
 		_ggzdmod_error(ggzdmod,
 			       "Error: received unknown seat from GGZ");
 		ggzdmod_log(ggzdmod, "GGZDMOD: Unknown seat type %d",
 			    seat.type);
+		return;
 	}
 
-
 	if (_ggzdmod_set_seat(ggzdmod, &seat) < 0) {
 		_ggzdmod_error(ggzdmod, "Error setting seat");
 		ggzdmod_log(ggzdmod, "GGZDMOD: Error setting seat");
@@ -1405,9 +1499,9 @@ void _ggzdmod_handle_reseat(GGZdMod * ggzdmod,
 
 	/* Change the old seat, and dup the previous value */
 	if (was_spectator) {
-		GGZSpectator s = {num: old_seat,
-				  name: NULL,
-				  fd: -1};
+		GGZSpectator s = {.num = old_seat,
+				  .name = NULL,
+				  .fd = -1};
 		GGZSpectator *old;
 
 		/* We have to manually pull off the FD to prevent it
@@ -1423,10 +1517,10 @@ void _ggzdmod_handle_reseat(GGZdMod * ggzdmod,
 		_ggzdmod_set_spectator(ggzdmod, &s);
 		old_event = GGZDMOD_EVENT_SPECTATOR_LEAVE;
 	} else {
-		GGZSeat s = {num: old_seat,
-			     name: NULL,
-			     type: GGZ_SEAT_OPEN,
-			     fd: -1};
+		GGZSeat s = {.num = old_seat,
+			     .name = NULL,
+			     .type = GGZ_SEAT_OPEN,
+			     .fd = -1};
 		GGZSeat *old;
 
 		/* We have to manually pull of the FD to prevent it
@@ -1438,6 +1532,16 @@ void _ggzdmod_handle_reseat(GGZdMod * ggzdmod,
 
 		name = ggz_strdup(old->name);
 
+		if (ggzdmod->state == GGZDMOD_STATE_PLAYING) {
+			/* Mark the seat as abandoned rather than open.
+			   FIXME - rather than have ggzdmod-game deduce this
+			   on its own, ggzdmod-ggz should instead send the
+			   full value of the new seat.  See also the comment
+			   in ggzdmod_reseat. */
+			s.name = name;
+			s.type = GGZ_SEAT_ABANDONED;
+		}
+
 		old_old = seat_copy(old);
 		_ggzdmod_set_seat(ggzdmod, &s);
 		old_event = GGZDMOD_EVENT_LEAVE;
@@ -1445,19 +1549,19 @@ void _ggzdmod_handle_reseat(GGZdMod * ggzdmod,
 
 	/* Change the new seat, and dup the preevious value */
 	if (is_spectator) {
-		GGZSpectator s = {num: new_seat,
-				  name: name,
-				  fd: fd};
+		GGZSpectator s = {.num = new_seat,
+				  .name = name,
+				  .fd = fd};
 		GGZSpectator old = ggzdmod_get_spectator(ggzdmod, new_seat);
 
 		new_old = spectator_copy(&old);
 		_ggzdmod_set_spectator(ggzdmod, &s);
 		new_event = GGZDMOD_EVENT_SPECTATOR_JOIN;
 	} else {
-		GGZSeat s = {num: new_seat,
-			     name: name,
-			     type: GGZ_SEAT_PLAYER,
-			     fd: fd};
+		GGZSeat s = {.num = new_seat,
+			     .name = name,
+			     .type = GGZ_SEAT_PLAYER,
+			     .fd = fd};
 		GGZSeat old = ggzdmod_get_seat(ggzdmod, new_seat);
 
 		new_old = seat_copy(&old);
@@ -1627,29 +1731,67 @@ void ggzdmod_check(GGZdMod *ggzdmod)
 
 
 void ggzdmod_report_game(GGZdMod *ggzdmod,
-			 int *teams,
-			 GGZGameResult *results)
+			 int *the_teams,
+			 GGZGameResult *the_results, int *the_scores)
 {
-	if (ggzdmod) {
-		char *names[ggzdmod->num_seats];
+	if (ggzdmod && ggzdmod->type == GGZDMOD_GAME) {
 		GGZSeatType types[ggzdmod->num_seats];
-		int p;
+		const char *names[ggzdmod->num_seats];
+		int teams[ggzdmod->num_seats];
+		GGZGameResult results[ggzdmod->num_seats];
+		int scores[ggzdmod->num_seats];
+		int p, r = 0;
 
+		/* Currently the values reported correspond to the seats
+		 * currently at the table.  However this could be changed
+		 * to allow the game to give us this info. */
 		for (p = 0; p < ggzdmod->num_seats; p++) {
 			GGZSeat seat = ggzdmod_get_seat(ggzdmod, p);
-			types[p] = seat.type;
-			if (seat.type == GGZ_SEAT_PLAYER
-			    || seat.type == GGZ_SEAT_BOT)
-				names[p] = seat.name;
-			else {
-				_ggzdmod_error(ggzdmod, "Invalid player.");
-				return;
+			int usable = 0;
+
+			switch (seat.type) {
+			case GGZ_SEAT_PLAYER:
+			case GGZ_SEAT_BOT:
+			case GGZ_SEAT_ABANDONED:
+				usable = 1;
+				break;
+			case GGZ_SEAT_RESERVED:
+			case GGZ_SEAT_OPEN:
+			case GGZ_SEAT_NONE:
+				break;
 			}
+			/* An ugly switch is used so if a new type is added
+			 * we'll get a warning. */
+			if (!usable) {
+				continue;
+			}
+
+			types[r] = seat.type;
+			names[r] = seat.name;
+			if (seat.type == GGZ_SEAT_BOT
+			    && (!names[r] || names[r][0] == '\0')) {
+				/* Game may have already set the name; if not
+				   we do it here. */
+				names[r] = "AI";
+			}
+			teams[r] = the_teams ? the_teams[p] : r;
+			results[r] = the_results[p];
+			scores[r] = the_scores ? the_scores[p] : 0;
+			r++;
 		}
 
-		_io_send_game_report(ggzdmod->fd,
-				     ggzdmod->num_seats,
-				     names, types, teams, results);
+		_io_send_game_report(ggzdmod->fd, r,
+				     names, types, teams, results, scores);
+	}
+}
+
+
+void ggzdmod_report_savegame(GGZdMod *ggzdmod, const char *savegame)
+{
+	if (ggzdmod && ggzdmod->type == GGZDMOD_GAME) {
+		if (!savegame) return;
+
+		_io_send_savegame_report(ggzdmod->fd, savegame);
 	}
 }
 
@@ -1680,17 +1822,24 @@ void ggzdmod_request_open(GGZdMod * ggzdmod, int seat_num)
 
 void _ggzdmod_handle_report(GGZdMod * ggzdmod,
 			    int num_players, char **names, GGZSeatType *types,
-			    int *teams, GGZGameResult *results)
+			    int *teams, GGZGameResult *results, int *scores)
 {
-	GGZdModGameReportData data = {num_players: num_players,
-				      names: names,
-				      types: types,
-				      teams: teams,
-				      results: results};
+	GGZdModGameReportData data = {.num_players = num_players,
+				      .names = names,
+				      .types = types,
+				      .teams = teams,
+				      .results = results,
+				      .scores = scores};
 	call_handler(ggzdmod, GGZDMOD_EVENT_GAMEREPORT, &data);
 }
 
 
+void _ggzdmod_handle_savegame(GGZdMod * ggzdmod, char *savegame)
+{
+	call_handler(ggzdmod, GGZDMOD_EVENT_SAVEGAMEREPORT, savegame);
+}
+
+
 void _ggzdmod_handle_num_seats_request(GGZdMod * ggzdmod, int num_seats)
 {
 	call_handler(ggzdmod, GGZDMOD_EVENT_REQ_NUM_SEATS, &num_seats);
