@@ -4,7 +4,7 @@
  * Project: ggzdmod
  * Date: 10/14/01
  * Desc: GGZ game module functions
- * $Id: ggzdmod.h,v 1.66 2003/05/11 09:11:27 dr_maux Exp $
+ * $Id: ggzdmod.h 7910 2006-03-14 13:50:41Z josef $
  *
  * This file contains the main interface for the ggzdmod library.  This
  * library facilitates the communication between the GGZ server (ggzd)
@@ -13,19 +13,19 @@
  *
  * Copyright (C) 2001-2002 GGZ Development Team.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #include <ggz.h> /* libggz */
@@ -58,10 +58,12 @@
  *
  * @code
  *     // Game-defined handler functions for GGZ events; see below.
- *     void handle_state_change(GGZdMod* ggz, GGZdModEvent event, void* data);
- *     void handle_player_join(GGZdMod* ggz, GGZdModEvent event, void* data);
- *     void handle_player_leave(GGZdMod* ggz, GGZdModEvent event, void* data);
- *     void handle_player_data(GGZdMod* ggz, GGZdModEvent event, void* data);
+ *     void handle_state_change(GGZdMod* ggz, GGZdModEvent event,
+ *                              const void *data);
+ *     void handle_player_seat(GGZdMod* ggz, GGZdModEvent event,
+ *                             const void *data);
+ *     void handle_player_data(GGZdMod* ggz, GGZdModEvent event,
+ *                             const void *data);
  *
  *     // Other game-defined functions (not ggz-related).
  *     void game_init(GGZdMod *ggz); // initialize a game
@@ -77,9 +79,11 @@
  *         ggzdmod_set_handler(ggz, GGZDMOD_EVENT_STATE,
  *                             &handle_state_change);
  *         ggzdmod_set_handler(ggz, GGZDMOD_EVENT_JOIN,
- *                             &handle_player_join);
+ *                             &handle_player_seat);
  *         ggzdmod_set_handler(ggz, GGZDMOD_EVENT_LEAVE,
- *                             &handle_player_leave);
+ *                             &handle_player_seat);
+ *         ggzdmod_set_handler(ggz, GGZDMOD_EVENT_SEAT,
+ *                             &handle_player_seat);
  *         ggzdmod_set_handler(ggz, GGZDMOD_EVENT_PLAYER_DATA,
  *                             &handle_player_data);
  *
@@ -102,11 +106,12 @@
  *         ggzdmod_free(ggz);
  *     }
  *
- *     void handle_state_change(GGZdMod* ggz, GGZdModEvent event, void* data)
+ *     void handle_state_change(GGZdMod* ggz, GGZdModEvent event,
+ *                              const void *data)
  *     {
- *         GGZdModState old_state = *(GGZdModState*)data;
+ *         const GGZdModState *old_state = data;
  *         GGZdModState new_state = ggzdmod_get_state(ggz);
- *         if (old_state == GGZDMOD_STATE_CREATED)
+ *         if (*old_state == GGZDMOD_STATE_CREATED)
  *             // ggzdmod data isn't initialized until it connects with GGZ
  *             // during the game launch, so some initializations should wait
  *             // until here.
@@ -138,26 +143,25 @@
  *         }
  *     }
  *
- *     void handle_player_join(GGZdMod* ggz, GGZdModEvent event, void* data)
+ *     void handle_player_seat(GGZdMod* ggz, GGZdModEvent event,
+ *                             const void *data)
  *     {
- *         GGZSeat *old_seat = (GGZSeat*)data;
- *         GGZSeat new_seat = ggzdmod_get_seat(ggz, old_seat->num);
+ *       const GGZSeat *old_seat = data;
+ *       GGZSeat new_seat = ggzdmod_get_seat(ggz, old_seat->num);
  *
- *         // ... do other player initializations ...
+ *       if (new_seat.type == GGZ_SEAT_PLAYER
+ *           && old_seat->type != GGZ_SEAT_PLAYER) {
+ *         // join event ... do player initializations ...
  *
- *         if (ggzdmod_count_seats(ggz, GGZ_SEAT_OPEN) == 0)
- *             // this particular game will only play when all seats are full.
- *             // calling this function triggers the STATE event, so we'll end
- *             // up executing resume_playing() above.
- *             ggzdmod_set_state(ggz, GGZDMOD_STATE_PLAYING);
- *     }
- *
- *     void handle_player_leave(GGZdMod* ggz, GGZdModEvent event, void* data)
- *     {
- *         GGZSeat *old_seat = (GGZSeat*)data;
- *         GGZSeat new_seat = ggzdmod_get_seat(ggz, old_seat->num);
- *
- *         // ... do other player un-initializations ...
+ *         if (ggzdmod_count_seats(ggz, GGZ_SEAT_OPEN) == 0) {
+ *           // this particular game will only play when all seats are full.
+ *           // calling this function triggers the STATE event, so we'll end
+ *           // up executing resume_playing() above.
+ *           ggzdmod_set_state(ggz, GGZDMOD_STATE_PLAYING);
+ *         }
+ *       } else if (new_seat.type != GGZ_SEAT_PLAYER
+ *                  && old_seat->type == GGZ_SEAT_PLAYER) {
+ *         // leave event ... do de-initialization ...
  *
  *         if (ggzdmod_count_seats(ggz, GGZ_SEAT_PLAYER) == 0)
  *             // the game will exit when all human players are gone
@@ -167,12 +171,14 @@
  *             // calling this function triggers the STATE event, so we'll end
  *             // up executing stop_playing() above.
  *             ggzdmod_set_state(ggz, GGZDMOD_STATE_WAITING);
+ *       }
  *     }
  *
- *     void handle_player_data(GGZdMod* ggz, GGZdModEvent event, void* data)
+ *     void handle_player_data(GGZdMod* ggz, GGZdModEvent event,
+ *                             const void *data)
  *     {
- *         int player = *(int*)data;
- *         int socket_fd = ggzdmod_get_seat(ggz, player).fd;
+ *         const int *player = data;
+ *         int socket_fd = ggzdmod_get_seat(ggz, *player).fd;
  *
  *         // ... read a packet from the socket ...
  *     }
@@ -187,8 +193,8 @@
 
 #define GGZDMOD_VERSION_MAJOR 0
 #define GGZDMOD_VERSION_MINOR 0
-#define GGZDMOD_VERSION_MICRO 7
-#define GGZDMOD_VERSION_IFACE "2:0:0"
+#define GGZDMOD_VERSION_MICRO 13
+#define GGZDMOD_VERSION_IFACE "5:0:1"
 
 #ifdef __cplusplus
 extern "C" {
@@ -244,7 +250,8 @@ typedef enum {
 	 *  old seat (a GGZSeat*) is passed as the event's data.
 	 *  The seat information will be updated before the event
 	 *  is invoked.
-	 *  @note This may be dropped in favor of the SEAT event. */
+	 *  @note This event is deprecated.
+	 *  @see GGZDMOD_EVENT_SEAT. */
 	GGZDMOD_EVENT_JOIN,
 
 	/** @brief Player left
@@ -252,16 +259,21 @@ typedef enum {
 	 *  old seat (a GGZSeat*) is passed as the event's data.
 	 *  The seat information will be updated before the event
 	 *  is invoked.
-	 *  @note This may be dropped in favor of the SEAT event */
+	 *  @note This event is deprecated.
+	 *  @see GGZDMOD_EVENT_SEAT. */
 	GGZDMOD_EVENT_LEAVE,
 
 	/** @brief General seat change
 	 *  This event occurs when a seat change other than a player
-	 *  leave/join happens (which is currently impossible).  The
-	 *  old seat (a GGZSeat*) is passed as the event's data.  The
-	 *  seat information will be updated before the event is invoked.
-	 *  @note This is currently unused, but may eventually replace JOIN and LEAVE.
-	 */
+	 *  leave/join happens.  The old seat (a GGZSeat*) is passed as the
+	 *  event's data.  The seat information will be updated before the
+	 *  event is invoked. This event will replace the JOIN and LEAVE
+	 *  events.  Games are advised to register the same handler for all
+	 *  three and to check the seat event by comparing the new and old
+	 *  seats.  Possible operations include open|reserved->player,
+	 *  player->open, open->bot, bot->open, reserved->open,
+	 *  open->reserved, and bot->bot.  Name changes are allowed but
+	 *  there is no player->player (i.e., player swap) seat event. */
 	GGZDMOD_EVENT_SEAT,
 
 	/** @brief A spectator joins the game.
@@ -302,6 +314,9 @@ typedef enum {
 	/** @brief For GGZ only.  Reports the results of a game. */
 	GGZDMOD_EVENT_GAMEREPORT,
 
+	/** @brief For GGZ only.  Reports a savegame. */
+	GGZDMOD_EVENT_SAVEGAMEREPORT,
+
 	/* GGZ-side only.  These should be moved to ggzdmod-ggz.h. */
 	GGZDMOD_EVENT_REQ_NUM_SEATS,
 	GGZDMOD_EVENT_REQ_BOOT,
@@ -315,6 +330,8 @@ typedef enum {
 	 *  not guaranteed that the GGZ connection will continue to
 	 *  work after an error has happened. */
 	GGZDMOD_EVENT_ERROR
+
+	/* GGZDMOD_EVENT_ERROR must be last in the list! */
 } GGZdModEvent;
 
 /** @brief The "type" of ggzdmod.
@@ -339,7 +356,7 @@ typedef struct GGZdMod GGZdMod;
  *
  *  A function of this type will be called to handle a ggzdmod event.
  *  @param mod The ggzdmod state object.
- *  @param e The event that has occured.
+ *  @param event The event that has occured.
  *  @param data Pointer to additional data for the event.
  *  The additional data will be of the following form:
  *    - GGZDMOD_EVENT_STATE: The old state (GGZdModState*)
@@ -353,7 +370,8 @@ typedef struct GGZdMod GGZdMod;
  *    - GGZDMOD_EVENT_SPECTATOR_DATA: The spectator number (int*)
  *    - GGZDMOD_EVENT_ERROR: An error string (char*)
  */
-typedef void (*GGZdModHandler) (GGZdMod * mod, GGZdModEvent e, void *data);
+typedef void (*GGZdModHandler) (GGZdMod * mod, GGZdModEvent event,
+				const void *data);
 
 /** @brief A seat at a GGZ game table.
  *
@@ -362,7 +380,7 @@ typedef void (*GGZdModHandler) (GGZdMod * mod, GGZdModEvent e, void *data);
 typedef struct {
 	unsigned int num;	/**< Seat index; 0..(num_seats-1). */
 	GGZSeatType type;	/**< Type of seat. */
-	char *name;		/**< Name of player occupying seat. */
+	const char *name;	/**< Name of player occupying seat. */
 	int fd;			/**< fd to communicate with seat occupant. */
 } GGZSeat;
 
@@ -372,7 +390,7 @@ typedef struct {
  */
 typedef struct {
 	unsigned int num;	/**< Spectator seat index */
-	char *name;		/**< The spectator's name */
+	const char *name;	/**< The spectator's name */
 	int fd;			/**< File descriptor for communication */
 } GGZSpectator;
 
@@ -435,6 +453,13 @@ int ggzdmod_get_num_seats(GGZdMod * ggzdmod);
  */
 GGZSeat ggzdmod_get_seat(GGZdMod * ggzdmod, int seat);
 
+/** @brief Return class for named bot.
+ *  @param ggzdmod The GGZdMod object.
+ *  @param name Name of the bot.
+ *  @return The bot's class, or NULL for anonymous bots.
+ */
+char * ggzdmod_get_bot_class(GGZdMod * ggzdmod, const char * name);
+
 /** @brief Return gamedata pointer
  *
  *  Each GGZdMod object can be given a "gamedata" pointer that is returned
@@ -593,7 +618,7 @@ int ggzdmod_disconnect(GGZdMod * ggzdmod);
  *  @param fmt A printf-style format string.
  *  @return 0 on success, -1 on failure.
  */
-int ggzdmod_log(GGZdMod * ggzdmod, char *fmt, ...)
+int ggzdmod_log(GGZdMod * ggzdmod, const char *fmt, ...)
 		ggz__attribute((format(printf, 2, 3)));
 		
 		
@@ -631,10 +656,23 @@ typedef enum {
  *  @param ggzdmod The ggzdmod object.
  *  @param teams An array listing a team number for each player, or NULL.
  *  @param results An array listing the result of the game for each player.
+ *  @param scores The scores for all players (may be NULL)
  */
 void ggzdmod_report_game(GGZdMod *ggzdmod,
 			 int *teams,
-			 GGZGameResult *results);
+			 GGZGameResult *results, int *scores);
+
+/** @brief Report the savegame to GGZ.
+ *
+ *  If a game saves the game data to disk, the directory name, file name or
+ *  any other associated token can be reported to GGZ.
+ *  In the case of a continuous game log, the reporting should happen at the
+ *  beginning as to allow the continuation of the saved game.
+ *
+ *  @param ggzdmod The ggzdmod object.
+ *  @param savegame Name of the savegame file within the game's directory.
+ */
+void ggzdmod_report_savegame(GGZdMod *ggzdmod, const char *savegame);
 
 /** @brief Tell GGZ to change the number of seats at this table.
  *  @note This functionality is incomplete, and should not yet be used. */
